/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel Function table
NOTES  : 
AUTHOR : Niels Mache
DATE   : 25.06.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include "kr_typ.h"	    /*	Kernel Types and Constants  */
#include "kr_def.h"	    /*	Default Values	*/


/*  The function table contains 5 entries:
     - the XGUI-Identifier of this function (a pointer to a string constant),
       note that the identifier have to match [A-Za-z]^[|, ]* (i.e. the first
       character of the symbol must be a letter and all other characters must
       not match [|, ])

     - an integer constant that determine the class of the function.
       7 different classes are possible:
       OUT_FUNC, ACT_FUNC, ACT_DERIV_FUNC, SITE_FUNC, UPDATE_FUNC, LEARN_FUNC
       and INIT_FUNC

     - the number of input parameters for this function (only needed for Update,
       Init and Learning functions)

     - the number of output parameters returned by this function (only needed
       for Update and Learning functions)

     - a pointer to the (user defined) function.
*/



/*  Some of the following functions are default functions of the SNNS-Kernel.
    If you want to change the default presettings, modify the constants
    defined in "kr_const.h".
*/


/*  Declare all functions as external functions !
*/
/*  Output functions  */
extern FlintType  OUT_Threshold05();
extern FlintType  OUT_Clip_01();
extern FlintType  OUT_Clip_11();

/*  Activation functions  */
extern FlintType  ACT_Logistic();
extern FlintType  ACT_BSBFunction();
/* extern FlintType  ACT_TanHFunction(); */
extern FlintType  ACT_Perceptron();
extern FlintType  ACT_Signum();
extern FlintType  ACT_Signum0();
extern FlintType  ACT_StepFunction();
extern FlintType  ACT_BAMFunction();
extern FlintType  ACT_LogisticI();
extern FlintType  ACT_MinOutPlusWeight();
extern FlintType  ACT_Linear();
extern FlintType  ACT_LogisticTbl();

/*  Derivation functions of the activation functions
*/
extern FlintType  ACT_DERIV_Logistic();
extern FlintType  ACT_DERIV_BSBFunction();
extern FlintType  ACT_DERIV_Dummy();
extern FlintType  ACT_DERIV_Identity();
extern FlintType  ACT_DERIV_Logistic();

/*  Site functions  */
extern FlintType  SITE_WeightedSum();
extern FlintType  SITE_ProductA();
extern FlintType  SITE_Product();
extern FlintType  SITE_Max();
extern FlintType  SITE_Min();

/*  Update functions  */
extern krui_err  UPDATE_serialPropagate();
extern krui_err  UPDATE_syncPropagate();
extern krui_err  UPDATE_randomOrderPropagate();
extern krui_err  UPDATE_randomPermutPropagate();
extern krui_err  UPDATE_topologicalPropagate();
extern krui_err  UPDATE_CPNPropagate();

/*  Leaning Functions  */
extern krui_err  LEARN_backprop();
extern krui_err  LEARN_backpropMomentum();
extern krui_err  LEARN_perc();
extern krui_err  LEARN_quickprop();
extern krui_err  LEARN_CPN();

/*  Initialisation Functions  */
extern krui_err  INIT_randomizeWeights();
extern krui_err  INIT_Weights_CPN();


/*  *********************************************************************  */

/*  Here is the function table
*/
struct FuncTable  func_table[]	=
  {

/*  Output functions  */
/*  In 'kr_const.h' is defined that the NULL-pointer presents the
    identity function. So the "Out_Identity" function is represented
    by the NULL-pointer instead of the function pointer.
*/
"Out_Identity",     OUT_FUNC, 0, 0, (FunctionPtr) OUT_IDENTITY,

"Out_Threshold05",  OUT_FUNC, 0, 0, (FunctionPtr) OUT_Threshold05,
"Out_Clip_01",      OUT_FUNC, 0, 0, (FunctionPtr) OUT_Clip_01,
"Out_Clip_11",      OUT_FUNC, 0, 0, (FunctionPtr) OUT_Clip_11,

/*  Activation functions  */
"Act_Logistic",     ACT_FUNC, 0, 0, (FunctionPtr) ACT_Logistic,
"Act_BSB",          ACT_FUNC, 0, 0, (FunctionPtr) ACT_BSBFunction,
/* "Act_TanH",         ACT_FUNC, 0, 0, (FunctionPtr) ACT_TanHFunction, */
"Act_Perceptron",   ACT_FUNC, 0, 0, (FunctionPtr) ACT_Perceptron,
"Act_Signum",       ACT_FUNC, 0, 0, (FunctionPtr) ACT_Signum,
"Act_Signum0",      ACT_FUNC, 0, 0, (FunctionPtr) ACT_Signum0,
"Act_StepFunc",     ACT_FUNC, 0, 0, (FunctionPtr) ACT_StepFunction,
"Act_BAM",          ACT_FUNC, 0, 0, (FunctionPtr) ACT_BAMFunction,
"Logistic_notInhibit" , ACT_FUNC, 0, 0, (FunctionPtr) ACT_LogisticI,
"Act_MinOutPlusWeight", ACT_FUNC, 0, 0, (FunctionPtr) ACT_MinOutPlusWeight,
"Act_Identity",     ACT_FUNC, 0, 0, (FunctionPtr) ACT_Linear,
"Act_LogisticTbl",  ACT_FUNC, 0, 0, (FunctionPtr) ACT_LogisticTbl,

/*  Derivation functions of the activation functions
    NOTE: The derivative function of the activation function MUST have
	  the same name as the belonging activation function.
*/
"Act_Logistic",     ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Logistic,
"Act_BSB",          ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_BSBFunction,
/* "Act_TanH",         ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy, */
"Act_Perceptron",   ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Act_Signum",       ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Act_Signum0",      ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Act_StepFunc",     ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Act_BAM",          ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Logistic_notInhibit",  ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Act_MinOutPlusWeight", ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Dummy,
"Act_Identity",     ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Identity,
"Act_LogisticTbl",  ACT_DERIV_FUNC, 0, 0, (FunctionPtr) ACT_DERIV_Logistic,

/*  Site functions  */
"Site_WeightedSum", SITE_FUNC, 0, 0, (FunctionPtr) SITE_WeightedSum,
"Site_Pi",          SITE_FUNC, 0, 0, (FunctionPtr) SITE_ProductA,
"Site_Produkt",     SITE_FUNC, 0, 0, (FunctionPtr) SITE_Product,
"Site_Max",         SITE_FUNC, 0, 0, (FunctionPtr) SITE_Max,
"Site_Min",         SITE_FUNC, 0, 0, (FunctionPtr) SITE_Min,

/*  Update functions  */
"Serial_Order",        UPDATE_FUNC, 0, 0, (FunctionPtr) UPDATE_serialPropagate,
"Synchonous_Order",    UPDATE_FUNC, 0, 0, (FunctionPtr) UPDATE_syncPropagate,
"Random_Order",        UPDATE_FUNC, 0, 0, (FunctionPtr) UPDATE_randomOrderPropagate,
"Random_Permutation",  UPDATE_FUNC, 0, 0, (FunctionPtr) UPDATE_randomPermutPropagate,
"Topological_Order",   UPDATE_FUNC, 0, 0, (FunctionPtr) UPDATE_topologicalPropagate,
"CPN_Order",           UPDATE_FUNC, 0, 0, (FunctionPtr) UPDATE_CPNPropagate,

/*  Leaning Functions  */
"Std_Backpropagation", LEARN_FUNC, 2, 1, (FunctionPtr) LEARN_backprop,
"BackpropMomentum",    LEARN_FUNC, 4, 1, (FunctionPtr) LEARN_backpropMomentum,
"BackPercolation",     LEARN_FUNC, 2, 1, (FunctionPtr) LEARN_perc,
"Quickprop",           LEARN_FUNC, 4, 1, (FunctionPtr) LEARN_quickprop,
"Counterpropagation",  LEARN_FUNC, 3, 1, (FunctionPtr) LEARN_CPN,

/*  Initialisation Functions  */
"Randomize_Weights",   INIT_FUNC, 2, 0, (FunctionPtr) INIT_randomizeWeights,
"CPN_Weights",         INIT_FUNC, 2, 0, (FunctionPtr) INIT_Weights_CPN

  };


/*  Define the function table size
*/
#ifdef __STDC__
const int  NO_OF_FUNCS = (sizeof (func_table) / sizeof (struct FuncTable));
#else
int  NO_OF_FUNCS = (sizeof (func_table) / sizeof (struct FuncTable));
#endif

/*  **********************************************************************  */
