/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel Network Initialisation Functions
NOTES  : 
AUTHOR : Niels Mache
DATE   : 18.03.91
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <math.h>

#include "kr_typ.h"	 /*  Kernel Types and Constants  */
#include "kr_const.h"	 /*  Constant Declarators for SNNS-Kernel  */
#include "kr_def.h"	 /*  Default Values  */
#include "random.h"	 /*  Randomize Library Function Prototypes  */
#include "kr_mac.h"	 /*  Kernel Macros   */


#define  INIT_PARAM1( param )   param[ 0 ]  /*    contains the 1st initialisation parameter  */
#define  INIT_PARAM2( param )   param[ 1 ]  /*    contains the 2nd initialisation parameter  */
#define  INIT_PARAM3( param )   param[ 2 ]  /*    contains the 3rd initialisation parameter  */
#define  INIT_PARAM4( param )   param[ 3 ]  /*    contains the 4th initialisation parameter  */


/*#################################################

GROUP: Global Var's

#################################################*/


extern bool    NetModified,	 /*  TRUE, if the network topology was modified  */
               PatternsValid,    /*  TRUE, if In-Core patterns are valid  */
               NetInitialize;    /*  TRUE, if the network has been initialized  */

extern int  NoOfUnits,		 /*  no. of units in the network  */
	    MinUnitNo,		 /*  the first (lowest) used unit number in the network  */
	    MaxUnitNo,		 /*  the last (highest) used unit number in the network  */
	    NoOfPatternPairs,	 /*  no. of defined pattern pairs */
	    NoOfShuffledPatterns,  /*  no. of entries in the pattern_number array  */
	    NoOfInputUnits,	 /*  no. of input units  */
	    NoOfOutputUnits,	 /*  no. of output units  */
	    TopoSortID; 	 /*  topologic mode identifier	*/

extern UnitArray       unit_array;  /*	the unit array	*/

extern TopoPtrArray    topo_ptr_array;	/*  stores pointers to topological sorted units
					    used by kr_topoSort()  */

extern Patterns        in_patterns,	/*  the input pattern array  */
		       out_patterns;	/*  the output pattern array  */
extern PatternNumbers  pattern_numbers;   /*  contains shuffled pattern numbers
					      used by kr_shufflePatterns()  */

extern int     no_of_topo_units;	/*  no. of unit pointers in the topo_ptr_array	*/


/*  --------------  Initialisation  Functions  ----------------------    */



/*  Initializes connection weights with uniform distributed random values.
    Function calls drand48(). <min> must be less then <max>.
*/
krui_err  INIT_randomizeWeights( parameterArray, NoOfParams )
float parameterArray[];
int   NoOfParams;
{
  register unsigned short    flags;
  register struct Link   *link_ptr;
  register struct Site   *site_ptr;
  register struct Unit   *unit_ptr;
  register FlintType     range, min_weight;
  FlintType   max_weight;

  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return( KRERR_NO_UNITS );  /*  there is nothing to do  */
  /*
  if (NoOfParams != 2)
    return( KRERR_PARAMETERS );  */ /*  Not the same no. of input parameters  */

  min_weight = INIT_PARAM1( parameterArray );
  max_weight = INIT_PARAM2( parameterArray ); 
  range = max_weight - min_weight;

  if (range == 0.0)
    {
    FOR_ALL_UNITS( unit_ptr )
      {
      flags = unit_ptr->flags;

      if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
        {     /*  unit is in use  */
        unit_ptr->bias = (FlintType) 0;

        if ( (flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
          {       /*  unit has sites  */
	  FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
              link_ptr->weight = min_weight;
          }
        else
          {       /*  unit has no sites   */
          if ( (flags & UFLAG_INPUT_PAT) == UFLAG_DLINKS )
            {     /*  unit has direct links   */
	    FOR_ALL_LINKS( unit_ptr, link_ptr )
              link_ptr->weight = min_weight;
            }
          }
        }
      }
    }
  else
    {
    FOR_ALL_UNITS( unit_ptr )
      {
      flags = unit_ptr->flags;

      if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
        {     /*  unit is in use  */
        unit_ptr->bias = (FlintType) 0;

        if ( (flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
          {       /*  unit has sites  */
	  FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
            link_ptr->weight = drand48() * range + min_weight;
          }
        else
          {       /*  unit has no sites   */

          if ( (flags & UFLAG_INPUT_PAT) == UFLAG_DLINKS )
            {     /*  unit has direct links   */
	    FOR_ALL_LINKS( unit_ptr, link_ptr )
              link_ptr->weight = drand48() * range + min_weight;
            }
          }
        }
      }
    }

  return( KRERR_NO_ERROR );
}



krui_err INIT_Weights_CPN(parameterArray, NoOfParams )
float parameterArray[];
int   NoOfParams;
{
  register struct Unit *unit_ptr;
  register struct Site *site_ptr;
  register struct Link *link_ptr;
  register TopoPtrArray  topo_ptr;

  register FlintType  sum, amount, range;
  FlintType min, max;
  int  ret_code;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return( KRERR_NO_UNITS );  /*  there is nothing to do  */
  /*
  if (NoOfParams != 2)
    return( KRERR_PARAMETERS );  */ /*  Not the same no. of input parameters  */

  min = INIT_PARAM1( parameterArray );
  max = INIT_PARAM2( parameterArray );
  range = max - min;

  if (NetModified || (TopoSortID != TOPOLOGIC_TYPE))
    {  /*  networt was modified or topologic array isn't initialized  */
    ret_code = kr_topoSort( TOPOLOGIC_TYPE );
    if (ret_code != KRERR_NO_ERROR)
      return( ret_code );

    NetModified = FALSE;
    }

  topo_ptr = topo_ptr_array + (NoOfInputUnits + 1);
  sum = 0.0;

  /*  initialize weights of the hidden units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    { /*  this is a hidden unit  */
      /***********************************************************/
      /*     initialize the weights to the Kohonen Layer         */
      /***********************************************************/

    if ((unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
      { /* the unit has sites */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	  {
	  link_ptr->weight = drand48() * range + min;
	  sum += link_ptr->weight * link_ptr->weight;
	  }
      }
    else
      { /* the unit has direct links */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{
	link_ptr->weight = drand48() * range + min;
	sum += link_ptr->weight * link_ptr->weight;
	}
      }
    /* normalize the weightvector to the Kohonen Layer */

    amount = 1.0 / sqrt( sum );

    if ((unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
      /* the unit has sites */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	  link_ptr->weight = link_ptr->weight * amount;
    else
      /* the unit has direct links */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	link_ptr->weight = link_ptr->weight * amount;
  }


  while ((unit_ptr = *++topo_ptr) != NULL)
    { /* this is a output unit */
      /***********************************************************/
      /*     initialize the weights to the Grossberg Layer       */
      /***********************************************************/

    if ((unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
      { /* the unit has sites */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	  link_ptr->weight = drand48() * range + min;
      }
    else
      {  /* the unit has direct links */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	link_ptr->weight = drand48() * range + min;
      }
    }

  return( KRERR_NO_ERROR );
}
