/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS Kernel 
NOTES  : 
AUTHOR : Niels Mache
DATE   : 20.02.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <memory.h>
#include <math.h>

#include "kr_typ.h"	 /*  Kernel Types and Constants  */
#include "kr_const.h"	 /*  Constant Declarators for SNNS-Kernel  */
#include "kr_def.h"	 /*  Default Values  */
#include "random.h"	 /*  Randomize Library Function Prototypes  */

#include "kernel.h"	 /*  Function Prototypes  */
#include "kr_mem.h"	 /*  Function Prototypes  */
#include "kr_funcs.h"	 /*  Function Prototypes  */
#include "kr_mac.h"	 /*  Kernel Macros  */



/*#################################################

GROUP: Global Var's

#################################################*/

bool	NetModified = FALSE,	  /*  TRUE, if the network topology was modified  */
	PatternsValid = FALSE,	  /*  TRUE, if In-Core patterns are valid  */
        NetInitialize = TRUE,     /*  TRUE, if the network has been initialized  */
        LearnFuncHasChanged = TRUE;  /*  TRUE, if the learning function has been changed  */

FlagWord
     DefaultSType   = DEF_STYPE;  /*  default topological type	*/

int  NoOfUnits	    = 0,	/*  no. of units in the network  */
     MinUnitNo	    = 0,	/*  the first (lowest) used unit number in the network	*/
     MaxUnitNo	    = 0,	/*  the last (highest) used unit number in the network	*/
     NoOfPatternPairs = 0,	/*  no. of defined pattern pairs */
     CurrentPattern   = 0,	/*  current pattern no.  */
     NoOfShuffledPatterns = 0,	/*  no. of entries in the pattern_number array	*/
     NoOfInputUnits   = 0,	/*  no. of input units	*/
     NoOfOutputUnits  = 0,      /*  no. of output units  */
     NoOfHiddenUnits  = 0,	/*  no. of hidden units  */
     TopoSortID       = NOT_SORTED,  /*  topologic mode identifier  */

     DefaultSubnetNo  = DEF_SUBNET_NO,	 /*  default subnet no.  */
     DefaultLayerNo   = DEF_LAYER_NO,	 /*  default layer no.	*/
     DefaultPosX      = DEF_POS_X,	 /*  default x-position  */
     DefaultPosY      = DEF_POS_Y;	 /*  default y-position  */
#ifdef KERNEL3D
int  DefaultPosZ      = DEF_POS_Z;	 /*  default z-position  */
#endif

FlintType
     DefaultIAct      = DEF_I_ACT,	 /*  default initial activation  */
     DefaultBias      = DEF_BIAS;	 /*  default bias */

OutFuncPtr
     DefaultUFuncOut  = (OutFuncPtr) DEF_UFUNC_OUT;   /*  default output function  */
ActFuncPtr
     DefaultUFuncAct  = (ActFuncPtr) DEF_UFUNC_ACT;   /*  default activation function  */
ActDerivFuncPtr
     DefaultUFuncActDeriv = (ActDerivFuncPtr) DEF_UFUNC_ACT_DERIV;   /*  default derivation act. function  */
LearnFuncPtr
     CurrLearnFunc    = (LearnFuncPtr) DEF_LEARN_FUNC;  /*  default learning function  */
UpdateFuncPtr
     CurrUpdateFunc   = (UpdateFuncPtr) DEF_UPDATE_FUNC; /*  default update function  */
InitFuncPtr
     CurrInitFunc     = (InitFuncPtr) DEF_INIT_FUNC;     /*  default initialisation function  */


UnitArray	unit_array	= NULL;  /*  the unit array  */

TopoPtrArray	topo_ptr_array	= NULL;  /*  stores pointers to topological sorted units
					     used by kr_topoSort()  */

Patterns	in_patterns	= NULL,  /*  the input pattern array  */
		out_patterns	= NULL;  /*  the output pattern array  */
PatternNumbers	pattern_numbers = NULL;  /*  contains shuffled pattern numbers
					     used by kr_shufflePatterns()  */

int	no_of_topo_units = 0;		 /*  no. of unit pointers in the topo_ptr_array  */



/*  global variables for the krui_error( ... ) function needed to generate
    error messages
*/
int  lineno = 0,	   /*  file I/O: line no. in network file, if network loading failed  */
     topo_NoOfCycles,	   /*  topological sorting: no. of detected cycles in the network  */
     topo_NoOfDeadUnits,   /*  topological sorting: no. of dead units in the network  */
     topo_err_unit;	   /*  topological sorting: first cycle or first dead unit  */

krui_err
     io_err = 0;	   /*  file I/O: I/O error code, if network loading failed  */


/*#################################################

GROUP: Local Vars

#################################################*/

/*  topological sorting: global pointer to the topologic array
    (reduces memory consumption in the recursive depth search routine)	*/
static TopoPtrArray	global_topo_ptr;

static int max_topo_depth;

static krui_err  topo_err;  /*	contains the error code of the depth-search
				routine, if searching failed  */


/*#################################################

GROUP: Low-Level Kernel Functions

#################################################*/


/*  delete all inputs at the given unit
*/
void    kr_deleteAllInputs( unit_ptr )
struct Unit   *unit_ptr;
{
  register struct Site   *site_ptr;


  if (UNIT_HAS_SITES( unit_ptr ))
    {   /*  Unit has sites  */
    FOR_ALL_SITES( unit_ptr, site_ptr )
      /*  Release all links   */
      krm_releaseAllLinks( site_ptr->links );

    krm_releaseAllSites( unit_ptr->sites );
    }
  else
    {   /*  Unit don't has sites   */
    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      krm_releaseAllLinks( (struct Link *) unit_ptr->sites );
    }

  unit_ptr->sites = NULL;

  /*  The unit has no inputs now  */
  unit_ptr->flags &= (~UFLAG_INPUT_PAT);
}


/*  Deletes all output links at <source_unit>
    NOTE: This function is slow.
*/
void   kr_deleteAllOutputLinks( source_unit_ptr )
struct Unit  *source_unit_ptr;
{
  register struct Link   *link_ptr,
                         *pred_link_ptr;
  register struct Site   *site_ptr;
  register struct Unit   *unit_ptr;
  register FlagWord	 flags;



  FOR_ALL_UNITS( unit_ptr )
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {   /*  unit is in use  */
      if (flags & UFLAG_SITES)
        {  /*  unit has sites  */
	FOR_ALL_SITES( unit_ptr, site_ptr )
          for (link_ptr = site_ptr->links, pred_link_ptr = NULL;
               link_ptr != NULL;
               pred_link_ptr = link_ptr, link_ptr = link_ptr->next)

            if (link_ptr->to == source_unit_ptr)
              {     /*  Connection between unit and source_unit found   */
              if (pred_link_ptr == NULL)
                site_ptr->links = link_ptr->next;
              else
                pred_link_ptr->next = link_ptr->next;

              krm_releaseLink( link_ptr );

              break;    /*  next site/unit  */
              }
        }
      else
        {  /*  unit has no sites   */
	if (flags & UFLAG_DLINKS)
          /*  unit has direct links   */
          for (link_ptr = (struct Link *) unit_ptr->sites, pred_link_ptr = NULL;
               link_ptr != NULL;
               pred_link_ptr = link_ptr, link_ptr = link_ptr->next)

            if (link_ptr->to == source_unit_ptr)
              {     /*  Connection between unit and source_unit found   */
              if (pred_link_ptr == NULL)
                {
                unit_ptr->sites = (struct Site *) link_ptr->next;
                if (link_ptr->next == NULL)
                  /*  The unit has no inputs now  */
                  unit_ptr->flags &= (~UFLAG_INPUT_PAT);
                }
              else
                pred_link_ptr->next = link_ptr->next;

              krm_releaseLink( link_ptr );

              break;    /*  next unit  */
              }
        }
      } /*  if  */
    }   /*  for */

}



/*  Copies all input links. Link list must not be NULL !
    Returns NULL if memory allocation fails.
*/
struct Link     *kr_copyLinkList( link_list )
struct Link   *link_list;
{
  register struct Link   *org_link,
                         *new_link_list,
                         *prev_link,
                         *cpy_link;


  if ( (new_link_list = krm_getLink() ) == NULL)  return( NULL );

  new_link_list->to     = link_list->to;
  new_link_list->weight = link_list->weight;
  prev_link = new_link_list;

  for (org_link = link_list->next; org_link != NULL; org_link = org_link->next)
    {
    if ( (cpy_link = krm_getLink() ) == NULL)  return( NULL );

    prev_link->next = cpy_link;
    prev_link = cpy_link;

    cpy_link->to     = org_link->to;
    cpy_link->weight = org_link->weight;
    }

  prev_link->next = NULL;

  return( new_link_list );
}


/*  Copies all sites with input links. Site list must not be NULL !
    Returns NULL if memory allocation fails.
*/
struct Site     *kr_copySiteListAndLinks( site_list )
struct Site   *site_list;
{
  struct Site   *org_site,
                *new_site_list,
                *prev_site,
                *cpy_site;

  if ( (new_site_list = krm_getSite() ) == NULL)  return( NULL );

  if (site_list->links != NULL)
    {
    if ( (new_site_list->links = kr_copyLinkList( site_list->links ) ) == NULL)
      return( NULL );
    }
  else
    new_site_list->links = NULL;

  new_site_list->site_table  = site_list->site_table;
  prev_site = new_site_list;

  for (org_site = site_list->next; org_site != NULL; org_site = org_site->next)
    {
    if ( (cpy_site = krm_getSite() ) == NULL)  return( NULL );

    prev_site->next = cpy_site;
    prev_site = cpy_site;

    if (org_site->links != NULL)
      {
      if ( (cpy_site->links = kr_copyLinkList( org_site->links ) ) == NULL)
        return( NULL );
      }
    else
      cpy_site->links = NULL;

    cpy_site->site_table = org_site->site_table;
    }

  prev_site->next = NULL;

  return( new_site_list );
}


/*  Copies all sites. Site list must not be NULL !
    Returns NULL if memory allocation fails.
*/
struct Site     *kr_copySiteList( site_list )
struct Site   *site_list;
{
  struct Site   *org_site,
                *new_site_list,
                *prev_site,
                *cpy_site;

  if ( (new_site_list = krm_getSite() ) == NULL)  return( NULL );

  new_site_list->links = NULL;
  new_site_list->site_table = site_list->site_table;
  prev_site = new_site_list;

  for (org_site = site_list->next; org_site != NULL; org_site = org_site->next)
    {
    if ( (cpy_site = krm_getSite() ) == NULL)  return( NULL );

    prev_site->next = cpy_site;
    prev_site = cpy_site;

    cpy_site->links = NULL;
    cpy_site->site_table = org_site->site_table;
    }

  prev_site->next = NULL;

  return( new_site_list );
}



/*  Copies all output links at <source_unit> to <new_unit>.
    Returns error code if memory allocation fails.
NOTE: This function is slow.
*/
int             kr_copyOutputLinks( source_unit_ptr, new_unit_ptr )
struct Unit     *source_unit_ptr,
                *new_unit_ptr;
{
  register struct Link   *link_ptr,
                         *new_link_ptr;
  register struct Site   *site_ptr;
  register FlagWord	 flags;
  register struct Unit   *unit_ptr;


  FOR_ALL_UNITS( unit_ptr )
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {   /*  unit is in use  */
      if (flags & UFLAG_SITES)
        {  /*  unit has sites  */
	FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	  if (link_ptr->to == source_unit_ptr)
	    {	  /*  Connection between unit and source_unit found   */
	    if ( (new_link_ptr = krm_getLink() ) == NULL)  return( 1 );

	    new_link_ptr->next = site_ptr->links;
	    site_ptr->links = new_link_ptr;

	    new_link_ptr->to = new_unit_ptr;
	    new_link_ptr->weight = link_ptr->weight;

	    break;    /*  next site/unit  */
	    }
        }
      else
        {  /*  unit has no sites   */
	if (flags & UFLAG_DLINKS)
          /*  unit has direct links   */
	  FOR_ALL_LINKS( unit_ptr, link_ptr )
            if (link_ptr->to == source_unit_ptr)
              {     /*  Connection between unit and source_unit found   */
              if ( (new_link_ptr = krm_getLink() ) == NULL)  return( 1 );

              new_link_ptr->next = (struct Link *) unit_ptr->sites;
              unit_ptr->sites = (struct Site *) new_link_ptr;

              new_link_ptr->to = new_unit_ptr;
              new_link_ptr->weight = link_ptr->weight;

              break;    /*  next unit  */
              }
        }
      } /*  if  */
    }   /*  for */

  return( KRERR_NO_ERROR );
}



/*#################################################

GROUP: Site Name/Func functions

#################################################*/

/*  search for a site at a unit 
*/
struct Site   *kr_searchUnitSite( unit_ptr, stbl_ptr )
struct Unit   *unit_ptr;
struct SiteTable    *stbl_ptr;
{
  register struct Site   *site_ptr;

  FOR_ALL_SITES( unit_ptr, site_ptr )
    if (site_ptr->site_table == stbl_ptr)
      return( site_ptr );

  return( NULL );   /*  there is no site at this unit with this name    */
}


/*  search for a site at a unit and returns also the predecessor site ptr 
*/
void    kr_searchUnitSiteWithPred( unit_ptr, stbl_ptr, Rsite_ptr, Rsite_pred_ptr )
struct Unit   *unit_ptr;
struct SiteTable    *stbl_ptr;
struct Site   * *Rsite_ptr,  * *Rsite_pred_ptr;
{
  register struct Site   *site_ptr,
                         *site_pred_ptr;

  for (site_ptr = unit_ptr->sites, site_pred_ptr = NULL;
       site_ptr != NULL;
       site_pred_ptr = site_ptr, site_ptr = site_ptr->next)
    if (site_ptr->site_table == stbl_ptr)
      {
      *Rsite_ptr = site_ptr;
      *Rsite_pred_ptr = site_pred_ptr;
      return;
      }

  *Rsite_ptr      = NULL;    /*  there is no site at this unit with this name    */
  *Rsite_pred_ptr = NULL;
}



/*  searches for a site in the network  
*/
int     kr_searchNetSite( stbl_ptr )
struct SiteTable    *stbl_ptr;
{
  int   i;
  struct Unit   *unit_ptr;
  struct Site   *site_ptr;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return( KRERR_NO_ERROR );

  for (i = MinUnitNo, unit_ptr = unit_array + MinUnitNo;
       i <= MaxUnitNo;
       i++, unit_ptr++)
    {
    if ( ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) &&
	 (UNIT_HAS_SITES( unit_ptr )) )
      {     /*  unit is in use and unit has sites */
      FOR_ALL_SITES( unit_ptr, site_ptr )
        {
        if (site_ptr->site_table == stbl_ptr)
          return( i );  /*  return unit no. */
        }
      }
    }

  return( 0 );  /*  site isn't in use   */
}


/*#################################################

GROUP: Link Functions

#################################################*/

/*  Add random uniform distributed values to connection weights.
    <minus> must be less then <plus>.
*/
void    kr_jogWeights( minus, plus )
FlintTypeParam   minus, plus;
{
  FlagWord	flags;
  struct Unit   *unit_ptr;
  struct Site   *site_ptr;
  struct Link   *link_ptr;
  FlintTypeParam  range;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )  return;
  range = plus - minus;

  FOR_ALL_UNITS( unit_ptr )
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      /*  unit is in use  */
      if (flags & UFLAG_DLINKS)
	/*  unit has direct links   */
	FOR_ALL_LINKS( unit_ptr, link_ptr )
	  link_ptr->weight += (FlintType) (drand48() * range + minus);
      else
	if (flags & UFLAG_SITES)
	  /*  unit has sites  */
	  FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	    link_ptr->weight += (FlintType) (drand48() * range + minus);
  }
}



/*#################################################

GROUP: Site Functions

#################################################*/

/*  Creates a new site with default initialisation
*/
struct Site     *kr_createDefaultSite()
{
  struct Site   *site_ptr;


  if ( (site_ptr = krm_getSite() ) == NULL)   return( NULL );

  site_ptr->links = NULL;
  site_ptr->next  = NULL;

  return( site_ptr );
}



/*#################################################

GROUP: Unit Functions

#################################################*/

/*  returns the pointer to the given unit, returns NULL if unit doesn't
    exist   
*/
struct Unit   *kr_getUnitPtr( unit_no )
int     unit_no;
{
struct Unit   *unit_ptr;


  if ( (unit_no < MinUnitNo) || (unit_no > MaxUnitNo) || (NoOfUnits == 0) )
    return( NULL );    /*  invalid unit #  */

  unit_ptr = unit_array + unit_no;
  if ((unit_ptr->flags & UFLAG_IN_USE) != UFLAG_IN_USE)
    return( NULL );    /*  invalid unit #  */

  return( unit_ptr );
}



/*  Searches for a unit with the given symbol pointer.
    Returns the first unit no. if a unit with the given name was found,
    0 otherwise.
*/
int  kr_unitNameSearch( min_unit_no, unit_symbol_ptr )
int  min_unit_no;
char    *unit_symbol_ptr;
{
  register char   *symbol;
  register struct Unit   *unit_ptr;


  if ((symbol = unit_symbol_ptr) == NULL)
    return( 0 );

  /*  search for symbol pointer  */
  for (unit_ptr = unit_array + min_unit_no; unit_ptr <= unit_array + MaxUnitNo; unit_ptr++)
    if ( (unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      if (unit_ptr->unit_name == symbol)
        return( unit_ptr - unit_array );

  return( 0 );
}



/*  Initializes a unit with default values.
*/
void   kr_makeDefaultUnit( unit_ptr )
struct Unit  *unit_ptr;
{
  unit_ptr->Out.output = (FlintType) DEF_OUT;
  unit_ptr->Ftype_entry = NULL;
  unit_ptr->act      = DefaultIAct;
  unit_ptr->i_act    = DefaultIAct;
  unit_ptr->bias     = DefaultBias;

  unit_ptr->value_a = (FlintType) 0;          /*previous bias change*/
  unit_ptr->value_b = (FlintType) 0;          /*previous bias slope*/
  unit_ptr->value_c = (FlintType) 0;          /*actual bias slope*/

  unit_ptr->out_func = DefaultUFuncOut; 	/*  default output function  */
  unit_ptr->act_func = DefaultUFuncAct; 	/*  default activation function */
  unit_ptr->act_deriv_func = DefaultUFuncActDeriv; 	/*  default derivation act. function */
  unit_ptr->unit_name= NULL;			/*  default is no unit name */
  unit_ptr->subnet_no  = DefaultSubnetNo;
  unit_ptr->layer_no   = DefaultLayerNo;
  unit_ptr->unit_pos.x = DefaultPosX;
  unit_ptr->unit_pos.y = DefaultPosY;
#ifdef KERNEL3D
  unit_ptr->unit_pos.z = DefaultPosZ;
#endif

  unit_ptr->sites    = NULL;
  unit_ptr->flags    = UFLAG_INITIALIZED | DefaultSType;

}


/*  copy the source unit, but not the flags of the source unit
*/
void    kr_copyUnit( source_unit_ptr, new_unit_ptr )
struct  Unit    *source_unit_ptr,
                *new_unit_ptr;
{
  memcpy( (char *) new_unit_ptr, (char *) source_unit_ptr, UNIT_SIZE );
  /*  set unit-flags  */
  new_unit_ptr->flags = UFLAG_NOT_FREE;

  /*  delete site pointer  */
  new_unit_ptr->sites	    = NULL;
}



/*#################################################

GROUP: Ftype Unit Functions

#################################################*/

/*  changes all units in the network with the given functionality type
    to the new functions of the (new) functionality type
*/
void   kr_changeFtypeUnits( Ftype_entry )
struct  FtypeUnitStruct     *Ftype_entry;
{
  register struct Unit   *unit_ptr;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return;

  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {     /*  unit is in use  */
      if (unit_ptr->Ftype_entry == Ftype_entry)
	{  /*  unit with this type was found. Now change the transfer functions
	       of the unit to the modified functionality type */
        unit_ptr->act_func = Ftype_entry->act_func;
        unit_ptr->out_func = Ftype_entry->out_func;
        unit_ptr->act_deriv_func = Ftype_entry->act_deriv_func;
        }
      }
}


/*   delete the functionality type of the units with the given type
*/
void  kr_deleteUnitsFtype( ftype_ptr )
struct  FtypeUnitStruct   *ftype_ptr;
{
  register struct Unit   *unit_ptr;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return;

  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {     /*  unit is in use  */
      if (unit_ptr->Ftype_entry == ftype_ptr)
        unit_ptr->Ftype_entry = NULL;
      }
}


/*  create a new unit with the given functionality type
*/
struct  Unit   *kr_makeFtypeUnit( unit_ptr, ftype_ptr )
struct  Unit      *unit_ptr;
struct  FtypeUnitStruct   *ftype_ptr;
{
  register struct  Site  *ftype_site,
                *site_ptr;


  kr_makeDefaultUnit( unit_ptr );

  unit_ptr->Ftype_entry = ftype_ptr;
  unit_ptr->out_func    = ftype_ptr->out_func;
  unit_ptr->act_func    = ftype_ptr->act_func;
  unit_ptr->act_deriv_func = ftype_ptr->act_deriv_func;

  ftype_site = ftype_ptr->sites;

  while (ftype_site != NULL)
    {   /*  Ftype has sites */
    if ( (site_ptr = krm_getSite() ) == NULL)
      {
      krm_releaseAllSites( unit_ptr->sites );
      unit_ptr->sites = NULL;
      return( NULL );
      }

    site_ptr->next = unit_ptr->sites;
    unit_ptr->sites = site_ptr;

    site_ptr->site_table = ftype_site->site_table;
    ftype_site = ftype_site->next;
    }

  if (ftype_ptr->sites != NULL)
    unit_ptr->flags |= UFLAG_SITES;     /*  unit has now sites  */

  return( unit_ptr );
}


/*  returns TRUE, if there exists the given site at the given ftype entry
*/
bool    kr_FtypeSiteSearch( ftype_first_site, site_table_ptr)
struct  Site        *ftype_first_site;
struct  SiteTable   *site_table_ptr;
{
  register struct  Site      *site_ptr;


  for (site_ptr = ftype_first_site; site_ptr != NULL; site_ptr = site_ptr->next)
    if (site_ptr->site_table == site_table_ptr)
      return( TRUE );

  return( FALSE );
}


/*  change the properties of the given unit to the properties of the
    given F-Type
*/
void    kr_changeFtypeUnit( unit_ptr, ftype_ptr )
struct  Unit      *unit_ptr;
struct  FtypeUnitStruct   *ftype_ptr;
{
  FlagWord	flags;
  struct  Site  *site_ptr,
                *pred_site_ptr,
                *tmp_ptr,
                *ftype_site;


  unit_ptr->out_func = ftype_ptr->out_func;
  unit_ptr->act_func = ftype_ptr->act_func;
  unit_ptr->act_deriv_func = ftype_ptr->act_deriv_func;

  flags = unit_ptr->flags & UFLAG_INPUT_PAT;

  switch (flags)
    {
    case  UFLAG_NO_INP:
      /*  Unit has no inputs  */
      if (ftype_ptr->sites != NULL)
        /*    Ftype has sites, delete unit's Ftype  */
        unit_ptr->Ftype_entry = NULL;
      else
        /*    Ftype and unit don't have sites */
        unit_ptr->Ftype_entry = ftype_ptr;    /*  unit accept Ftype and inputs */

      return;     /*  done !  */

    case  UFLAG_SITES:
      /*  Unit has sites  */
      ftype_site = ftype_ptr->sites;
      if (ftype_site == NULL)
        {  /*    unit has sites, but Ftype don't has sites, delete unit's Ftype and all inputs  */
        unit_ptr->Ftype_entry = NULL;

        kr_deleteAllInputs( unit_ptr );
        unit_ptr->flags = UFLAG_INITIALIZED;  /*  unit has no inputs now !    */
        }
      else
        {     /*  both unit and Ftype have sites: check sites  */
        unit_ptr->Ftype_entry = ftype_ptr;

        site_ptr = unit_ptr->sites;
        pred_site_ptr = NULL;

        do
          {
          if ( ! kr_FtypeSiteSearch( ftype_site, site_ptr->site_table ))
            {  /*  Ftype and unit site definitions are not equivalent: remove site    */
            if (pred_site_ptr == NULL)
              {   /*  this is the first site at the unit  */
              unit_ptr->sites = site_ptr->next;

              if (site_ptr->next == NULL)
                /*  unit don't has any inputs   */
                unit_ptr->flags &= (~UFLAG_INPUT_PAT);
              }
            else
              {   /*  this site isn't the first site at the unit  */
              pred_site_ptr->next = site_ptr->next;
              pred_site_ptr = site_ptr;
              }

            tmp_ptr = site_ptr;           /*  work with temporary pointer and get */
            site_ptr = site_ptr->next;    /*  next site pointer BEFORE krm_releaseSite    */
            krm_releaseAllLinks( tmp_ptr->links );    /*  (important in a multiprocessor system   */
            krm_releaseSite( tmp_ptr );
            /*    delete unit's Ftype */
            unit_ptr->Ftype_entry = NULL;
            }
          else
            {
            pred_site_ptr = site_ptr;
            site_ptr = site_ptr->next;
            }
          }
        while (site_ptr != NULL);

        if (unit_ptr->sites == NULL)
          unit_ptr->flags = UFLAG_INITIALIZED;  /*  unit has no inputs now !    */
        }

      return;


    case  UFLAG_DLINKS:
    /*  Unit has direct links   */
      if (ftype_ptr->sites != NULL)
        {     /*  unit has direct links, but Ftype entry has sites: delete links  */
        unit_ptr->Ftype_entry = NULL;

        kr_deleteAllInputs( unit_ptr );
        unit_ptr->flags = UFLAG_INITIALIZED;  /*  unit has no inputs now !    */
        }
      else
        {    /*   unit has direct links and Ftype entry has no sites: use direct links  */
        unit_ptr->Ftype_entry = ftype_ptr;
        }

    }
}


/*  change a site at the F-Type
*/
void    kr_changeFtypeSites( Ftype_entry, old_site_table, new_site_table)
struct  FtypeUnitStruct     *Ftype_entry;
struct  SiteTable   *old_site_table,
                    *new_site_table;
{
  struct Unit   *unit_ptr;
  struct Site   *site_ptr;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return;

  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {     /*  unit is in use  */
      if (unit_ptr->Ftype_entry == Ftype_entry)
        {
	FOR_ALL_SITES( unit_ptr, site_ptr )
          {
          if (site_ptr->site_table == old_site_table)
            site_ptr->site_table = new_site_table;
          }
        }
      }
}


/*#################################################

GROUP: Miscellanous

#################################################*/

/*  spell checker  (check identifiers for matching [A-Za-z]^[|, ]*)
*/
bool    kr_symbolCheck( symbol )
char    *symbol;
{
  register  char  c;

  if (!isalpha( *symbol ))  return( FALSE );

  while ( (c = *(++symbol)) != '\0' )
    {
    if (!isgraph( c ))  return( FALSE );
    if ( c == '|' || c == ',')  return( FALSE );
    }

  return( TRUE );
}


/*  translate unit flags to the topological type of the unit
*/
int     kr_getTType( flags )
int     flags;
{
  switch (flags)
    {
    case UFLAG_TTYP_UNKN:  return( UNKNOWN );
    case UFLAG_TTYP_IN  :  return( INPUT );
    case UFLAG_TTYP_OUT :  return( OUTPUT );
    case UFLAG_TTYP_DUAL:  return( DUAL );
    case UFLAG_TTYP_HIDD:  return( HIDDEN );
    case UFLAG_TTYP_SPEC:  return( SPECIAL );

    default:  return( UNKNOWN );
    }
}


/*  translate the topological type to unit flags
*/
int     kr_setTType( ttype )
int     ttype;
{
  switch (ttype)
    {
    case UNKNOWN :  return( UFLAG_TTYP_UNKN );
    case INPUT   :  return( UFLAG_TTYP_IN );
    case OUTPUT  :  return( UFLAG_TTYP_OUT );
    case DUAL    :  return( UFLAG_TTYP_DUAL );
    case HIDDEN  :  return( UFLAG_TTYP_HIDD );
    case SPECIAL :  return( UFLAG_TTYP_SPEC );

    default :  return( -1 );
    }
}


/*  update the outputs of all units in the network
*/
void    kr_updateUnitOutputs()
{
  register struct Unit   *unit_ptr;


  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_INITIALIZED) == UFLAG_INITIALIZED)
      {     /*  unit is in use and enabled  */
      if (unit_ptr->out_func == NULL)
        /*  Identity Function   */
        unit_ptr->Out.output = unit_ptr->act;
      else
        unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
      }
}



/*  returns the no. of units of the specified topologic type
    (i.e. Input, Hidden, Output or Special units)
*/
int         kr_getNoOfUnits( UnitTType )
int  UnitTType;
{
  register struct Unit   *unit_ptr;
  register int   no_of_units;
  register FlagWord      ttyp_flg;
  int   flg;


  if ((NoOfUnits == 0) || ((flg = kr_setTType( UnitTType )) == -1))  
    return( 0 );  /*  no units or this topologic type doesn't exist  */

  ttyp_flg = (FlagWord) flg;
  no_of_units = 0;
  FOR_ALL_UNITS( unit_ptr )
    if ( ((unit_ptr->flags & UFLAG_TTYP_PAT) == ttyp_flg) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      no_of_units++;

  return( no_of_units );
}




/*  force unit array garbage collection
*/
void  kr_forceUnitGC()
{
  krm_unitArrayGC();
}


/*#################################################

GROUP: Functions default presettings

#################################################*/

/*  Returns information about the unit default settings.
*/
void	kr_getUnitDefaults( act, bias, ttflags, subnet_no, layer_no,
			    act_func_ptr, out_func_ptr )
FlintType   *act, *bias;
int	    *ttflags;
int	    *subnet_no, *layer_no;
ActFuncPtr  *act_func_ptr;
OutFuncPtr  *out_func_ptr;
{
  *act          = DefaultIAct;
  *bias         = DefaultBias;
  *ttflags	= (int) DefaultSType;
  *subnet_no    = DefaultSubnetNo;
  *layer_no     = DefaultLayerNo;
  *act_func_ptr = DefaultUFuncAct;
  *out_func_ptr = DefaultUFuncOut;
}


/*  Changes the unit default settings.
*/
void   kr_setUnitDefaults( act, bias, ttflags, subnet_no, layer_no,
			   act_func_ptr, act_deriv_func_ptr, out_func_ptr )
FlintTypeParam  act, bias;
int	    ttflags;
int         subnet_no, layer_no;
ActFuncPtr  act_func_ptr;
ActDerivFuncPtr  act_deriv_func_ptr;
OutFuncPtr  out_func_ptr;
{
  DefaultIAct       = (FlintType) act;
  DefaultBias       = (FlintType) bias;
  DefaultSType	    = (FlagWord) ttflags;
  DefaultPosX       = DEF_POS_X;
  DefaultPosY       = DEF_POS_Y;
#ifdef KERNEL3D
  DefaultPosZ       = DEF_POS_Z;
#endif
  DefaultSubnetNo   = subnet_no;
  DefaultLayerNo    = layer_no;
  DefaultUFuncOut   = out_func_ptr;
  DefaultUFuncAct   = act_func_ptr;
  DefaultUFuncActDeriv = act_deriv_func_ptr;
}


/*#################################################

GROUP: Topological Sorting Functions

#################################################*/


/*  Clears the 'touch' (refresh) flag of all units
*/
static void  clr_T_flags()
{
  register struct Unit   *unit_ptr;


  FOR_ALL_UNITS( unit_ptr )
    if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {
      unit_ptr->flags &= ~UFLAG_REFRESH;
      unit_ptr->Aux.int_no = 0;
      }
}


/*  Depth search routine for topological sorting
*/
static void  DepthFirst1( unit_ptr, depth )
struct Unit   *unit_ptr;
int     depth;
{
  struct Site   *site_ptr;
  struct Link   *link_ptr;


  if (unit_ptr->flags & UFLAG_REFRESH)
    {  /*  the 'touch' flag is set: don't continue search  */
    if (unit_ptr->Aux.int_no == 0)
      {  /*  Cycle found  */
      topo_NoOfCycles++;
      if (topo_err == KRERR_NO_ERROR)
	{  /*  remember the cycle unit	*/
	topo_err_unit = unit_ptr - unit_array;
	topo_err = KRERR_CYCLES;
      }
    }

    return;
  }
  else
    /*	set the 'touch' flag  */
    unit_ptr->flags |= UFLAG_REFRESH;

  switch (unit_ptr->flags & UFLAG_INPUT_PAT)
    {
    case  UFLAG_DLINKS:   /*  unit has direct links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	DepthFirst1( link_ptr->to, depth + 1 );  /*  increase depth  */

      break;

    case  UFLAG_SITES:	/*  unit has sites  */
	FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	  DepthFirst1( link_ptr->to, depth + 1 );  /*  increase depth  */

      break;
  }

  unit_ptr->Aux.int_no = depth;  /*  remember the depth (for cycle detection and statistics)  */
  *global_topo_ptr++ = unit_ptr;  /*  store sorted unit pointer  */
}


/*  Depth search routine for topology check function
*/
static void  DepthFirst2( unit_ptr, depth )
struct Unit   *unit_ptr;
int     depth;
{
  struct Site   *site_ptr;
  struct Link   *link_ptr;


  if (unit_ptr->flags & UFLAG_REFRESH)
    {  /*  the 'touch' flag is set: don't continue search  */
    if (unit_ptr->Aux.int_no == 0)
      {  /*  Cycle found  */
      topo_NoOfCycles++;
      if (topo_err == KRERR_NO_ERROR)
	{  /*  remember the cycle unit	*/
	topo_err_unit = unit_ptr - unit_array;
	topo_err = KRERR_CYCLES;
      }
    }

    return;
  }
  else
    /*	set the 'touch' flag  */
    unit_ptr->flags |= UFLAG_REFRESH;

  switch (unit_ptr->flags & UFLAG_INPUT_PAT)
    {
    case  UFLAG_DLINKS:   /*  unit has direct links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	DepthFirst2( link_ptr->to, depth + 1 );  /*  increase depth  */

      break;

    case  UFLAG_SITES:	/*  unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	DepthFirst2( link_ptr->to, depth + 1 );  /*  increase depth  */

      break;
  }

  /*  remember the depth (for cycle detection and statistics)  */
  unit_ptr->Aux.int_no = depth;
  /*  store highest layer no.  */
  if (depth > max_topo_depth)  max_topo_depth = depth;
}


/*  Depth search routine for topological sorting in feedforward networks
*/
static void  DepthFirst3( unit_ptr, depth )
struct Unit   *unit_ptr;
int     depth;
{
  struct Site   *site_ptr;
  struct Link   *link_ptr;


  if (unit_ptr->flags & UFLAG_REFRESH)
    {  /*  the 'touch' flag is set: don't continue search  */
    topo_err_unit = unit_ptr - unit_array; /*  store unit number  */

    if ((unit_ptr->flags & UFLAG_TTYP_PAT) == UFLAG_TTYP_OUT)
      {  /*  this output unit has a output connection to another unit  */
      if (topo_err == KRERR_NO_ERROR)
	topo_err = KRERR_O_UNITS_CONNECT;
    }
    else
      if (unit_ptr->Aux.int_no == 0)
	{  /*  Cycle found  */
	topo_NoOfCycles++;
	if (topo_err == KRERR_NO_ERROR)
	  topo_err = KRERR_CYCLES;
      }

    return;
  }
  else
    /*	set the 'touch' flag  */
    unit_ptr->flags |= UFLAG_REFRESH;

  switch (unit_ptr->flags & UFLAG_INPUT_PAT)
    {
    case  UFLAG_DLINKS:   /*  unit has direct links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	DepthFirst3( link_ptr->to, depth + 1 );  /*  increase depth  */

      break;

    case  UFLAG_SITES:	/*  unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	DepthFirst3( link_ptr->to, depth + 1 );  /*  increase depth  */

      break;
  }

  /*  remember the depth (for cycle detection and statistics)  */
  unit_ptr->Aux.int_no = depth;

  /*  store only hidden units  */
  if (unit_ptr->flags & UFLAG_TTYP_HIDD)
    *global_topo_ptr++ = unit_ptr;  /*	store sorted unit pointer  */
}




/*  Sort units topological (general version) and stores the
    pointers to this units in the topologic array.

NOTE:
    Units are not sorted by their topologic type (that's not possible in
    general case).
*/
static krui_err  kr_topoSortT()
{
  register struct Unit	 *unit_ptr;
  bool	o_units;


  topo_NoOfCycles = topo_NoOfDeadUnits = topo_err_unit = 0;
  topo_err = KRERR_NO_ERROR;

  clr_T_flags();    /*	reset units 'touch' flags  */
  global_topo_ptr = topo_ptr_array;  /*  initialize global pointer */

  /*  limit left side of the topological array with NULL pointer  */
  *global_topo_ptr++ = NULL;

  /*  begin depth search at the first output unit  */
  o_units = FALSE;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_OUT)  &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      o_units = TRUE;  /*  there is a output unit  */
      DepthFirst1( unit_ptr, 1 );  /*  sort the units topological (using depth search
				       algorithm, starting at this output unit */
      if (topo_err != KRERR_NO_ERROR)
	return( topo_err );  /*  stop if an error occured  */
      }

  if (!o_units)  return( KRERR_NO_OUTPUT_UNITS );  /*  no output units	*/

  /*  limit right side of the topologic array with NULL pointer  */
  *global_topo_ptr++ = NULL;

  /*  calc. no. of sorted units  */
  no_of_topo_units = (global_topo_ptr - topo_ptr_array) - 2;

  /*  search for dead units i.e. units without inputs  */
  FOR_ALL_UNITS( unit_ptr )
    if ( !(unit_ptr->flags & UFLAG_REFRESH) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      topo_NoOfDeadUnits++;
      if (topo_err_unit == 0)
	topo_err_unit = unit_ptr - unit_array;	/*  store the unit no.	*/
      }

  if (topo_NoOfDeadUnits != 0)	return( KRERR_DEAD_UNITS );

  return( KRERR_NO_ERROR );
}




/*  Sorts unit topological in feed-forward networks and stores the
    pointers to this units in the topologic array in the following order:
     - input,
     - hidden and
     - output units

    This function make following assumtions (like all learning functions for
    feed-forward networks):
     a) input units doesn't have input connections to other units and
     b) output units doesn't have outputs connections to other units.
*/
static krui_err  kr_topoSortFF()
{
  register struct Unit	 *unit_ptr;
  bool	io_units;


  topo_NoOfCycles = topo_NoOfDeadUnits = topo_err_unit = 0;
  topo_err = KRERR_NO_ERROR;

  clr_T_flags();    /*	reset units 'touch' flags  */
  global_topo_ptr = topo_ptr_array;  /*  initialize global pointer */

  /*  limit left side of the topological array with NULL pointer  */
  *global_topo_ptr++ = NULL;

  /*  put all input units in the topologic array  */
  io_units = FALSE;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_IN)  &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      if ((unit_ptr->flags & UFLAG_INPUT_PAT) != UFLAG_NO_INP)
	{  /*  this input unit has a connection to another unit  */
	topo_err_unit = unit_ptr - unit_array;	/*  store the unit no.	*/
	return( KRERR_I_UNITS_CONNECT );  /*  input unit has input  */
      }

      io_units = TRUE;	/*  there is a input unit  */
      *global_topo_ptr++ = unit_ptr;  /*  save input unit  */
    }

  if (!io_units)  return( KRERR_NO_INPUT_UNITS );  /*  no input units */

  /*  limit input units in the topological array with NULL pointer  */
  *global_topo_ptr++ = NULL;

  /*  begin depth search at the first output unit  */
  io_units = FALSE;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_OUT) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      io_units = TRUE;	/*  there is a output unit  */
      DepthFirst3( unit_ptr, 1 );  /*  sort the units topological (using depth search
				       algorithm, starting at this output unit */
      if (topo_err != KRERR_NO_ERROR)
	return( topo_err );  /*  stop if an error occured  */
    }

  if (!io_units)  return( KRERR_NO_OUTPUT_UNITS );  /*	no output units */

  /*  limit hidden units in the topological array with NULL pointer  */
  *global_topo_ptr++ = NULL;

  /*  put all output units in the topological array  */
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_OUT)  &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      *global_topo_ptr++ = unit_ptr;  /*  save output unit  */

  /*  limit right side of the topologic array with NULL pointer  */
  *global_topo_ptr++ = NULL;

  /*  calc. no. of sorted units  */
  no_of_topo_units = (global_topo_ptr - topo_ptr_array) - 4;

  /*  search for dead units i.e. units without inputs  */
  FOR_ALL_UNITS( unit_ptr )
    if ( !(unit_ptr->flags & UFLAG_REFRESH) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      topo_NoOfDeadUnits++;
      if (topo_err_unit == 0)
	topo_err_unit = unit_ptr - unit_array;	/*  store the unit no.	*/
    }

  if (topo_NoOfDeadUnits != 0)	return( KRERR_DEAD_UNITS );

  return( KRERR_NO_ERROR );
}



/*  Sort units by their topologic type, i.e. Input, Hidden, Output units and
    stores the pointers to this units in the topologic array.
*/
static krui_err  kr_topoSortIHO()
{
  TopoPtrArray     topo_ptr;
  register struct Unit   *unit_ptr;
  int  no_of_units;


  topo_ptr = topo_ptr_array;

  /*  limit left side of the topological array with NULL pointer  */
  *topo_ptr++ = NULL;

  /*  get input units  */
  no_of_units = 0;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_IN) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      *topo_ptr++ = unit_ptr;
      no_of_units++;
      }

  if (no_of_units != NoOfInputUnits)
    if (no_of_units == 0)
      return( KRERR_NO_INPUT_UNITS );
    else
      return( KRERR_CHANGED_I_UNITS );

  /*  limit input units in the topological array with NULL pointer  */
  *topo_ptr++ = NULL;

  /*  get hidden units  */
  no_of_units = 0;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_HIDD) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      *topo_ptr++ = unit_ptr;
      no_of_units++;
      }

  if ((NoOfHiddenUnits = no_of_units) == 0)  return( KRERR_NO_HIDDEN_UNITS );

  /*  limit hidden units in the topological array with NULL pointer  */
  *topo_ptr++ = NULL;

  /*  get output units  */
  no_of_units = 0;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_OUT) &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      *topo_ptr++ = unit_ptr;
      no_of_units++;
      }

  if (no_of_units != NoOfOutputUnits)
    if (no_of_units == 0)
      return( KRERR_NO_OUTPUT_UNITS );
    else
      return( KRERR_CHANGED_O_UNITS );

  /*  limit right side of the topologic array with NULL pointer  */
  *topo_ptr++ = NULL;

  /*  calc. no. of sorted units  */
  no_of_topo_units = (topo_ptr - topo_ptr_array) - 4;

  return( KRERR_NO_ERROR );
}


/*  Sort units according to the given mode:
    TOPOLOGICAL:
      Sort units topological (general version) and stores the
      pointers to this units in the topologic array.
      NOTE: Units are not sorted by their topologic type (that's not
	    possible in general case).

    TOPOLOGICAL_FF:
      Sorts unit topological in feed-forward networks and stores the
      pointers to this units in the topologic array in the following order:
       - input,
       - hidden and
       - output units

      This function make following assumtions (like all learning functions for
      feed-forward networks):
       a) input units doesn't have input connections to other units and
       b) output units doesn't have outputs connections to other units.

    TOPOLOGIC_TYPE:
      Sort units by their topologic type, i.e. Input, Hidden, Output units and
      stores the pointers to this units in the topologic array.
*/
krui_err  kr_topoSort( topo_sorting_mode )
int  topo_sorting_mode;
{
  krui_err  ret_code;


  TopoSortID = NOT_SORTED;
  if (NoOfUnits == 0)  return( KRERR_NO_UNITS );  /*  no units defined	*/

  if ( krm_allocUnitTopoArray( NoOfUnits + 4) != 0)
    return( KRERR_INSUFFICIENT_MEM );   /*  memory fault  */

  switch (topo_sorting_mode)
    {
    case  TOPOLOGICAL:
	ret_code = kr_topoSortT();
	break;
    case  TOPOLOGICAL_FF:
	ret_code = kr_topoSortFF();
	break;
    case  TOPOLOGIC_TYPE:
	ret_code = kr_topoSortIHO();
	break;

    default:
	ret_code = KRERR_TOPOMODE;
    }

  if ((ret_code != KRERR_NO_ERROR) && (ret_code != KRERR_DEAD_UNITS))
    {  /*  an serios error has occured, release topo array  */
    TopoSortID = NOT_SORTED;
    krm_releaseUnitTopoArray();
    }
  else
    {
    TopoSortID = topo_sorting_mode;
    }

  return( ret_code );
}



/*  *************************************************
*/

/*  Checks the topology of the network:
     a) counts the number of layers of the network and
     b) determines if the network has cycles.

    Returns the no. of layers of the network.
*/
int  kr_topoCheck()
{
  struct Unit   *unit_ptr;
  bool      o_units;


  topo_NoOfCycles = topo_NoOfDeadUnits = topo_err_unit = max_topo_depth = 0;
  topo_err = KRERR_NO_ERROR;

  if (NoOfUnits == 0)  return( KRERR_NO_UNITS );  /*  no units defined	*/

  clr_T_flags();    /*	reset units 'touch' flags  */

  /*  begin depth search at the first output unit  */
  o_units = FALSE;
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_TTYP_OUT)  &&
         ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {
      o_units = TRUE;
      DepthFirst2( unit_ptr, 1 );
      if (topo_err != KRERR_NO_ERROR)
	return( topo_err );  /*  stop if an error occured  */
      }

  if (!o_units)  return( KRERR_NO_OUTPUT_UNITS );  /*  no output units */

  /*  return the no. of layers of the network  */
  return( max_topo_depth );
}


/*  Count the no. of input and output units and return an error code
    if the no. do not fit to the loaded patterns.
*/
krui_err  kr_IOCheck()
{
  register struct Unit   *unit_ptr;
  register int  no_of_i_units, no_of_o_units;


  /*  get no. of input and output units  */
  no_of_i_units = 0;
  no_of_o_units = 0;
  FOR_ALL_UNITS( unit_ptr )
    if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE )
      if (unit_ptr->flags & UFLAG_TTYP_IN)
        no_of_i_units++;
      else 
	if (unit_ptr->flags & UFLAG_TTYP_OUT)
          no_of_o_units++;
      
  if (no_of_i_units != NoOfInputUnits)
    if (no_of_i_units == 0)
      return( KRERR_NO_INPUT_UNITS );
    else
      return( KRERR_CHANGED_I_UNITS );

  if (no_of_o_units != NoOfOutputUnits)
    if (no_of_o_units == 0)
      return( KRERR_NO_OUTPUT_UNITS );
    else
      return( KRERR_CHANGED_O_UNITS );

  return( KRERR_NO_ERROR );
}


krui_err  kr_makeUnitPermutation()
{
  register struct Unit   *unit_ptr;
  register int	   no_of_units, i;
  TopoPtrArray     topo_ptr,  t_ptr1,  t_ptr2;


  TopoSortID = NOT_SORTED;
  if (NoOfUnits == 0)  return( KRERR_NO_UNITS );  /*  no units defined	*/

  if ( krm_allocUnitTopoArray( NoOfUnits + 2) != 0)
    return( KRERR_INSUFFICIENT_MEM );   /*  memory fault  */

  topo_ptr = topo_ptr_array;

  /*  limit left side of the topological array with NULL pointer  */
  *topo_ptr++ = NULL;

  /*  initialize permutation array  */
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_INITIALIZED) == UFLAG_INITIALIZED)
      /*  unit is in use and enabled  */
      *topo_ptr++ = unit_ptr;

  no_of_topo_units = topo_ptr - topo_ptr_array;  /*   calc no. of sorted units	*/
  no_of_units = no_of_topo_units;

  topo_ptr = topo_ptr_array;
  /*  permutate unit order  */
  for (i = 0; i < no_of_units; i++)
    {
    t_ptr1 = topo_ptr + (lrand48() % no_of_units);
    t_ptr2 = topo_ptr + (lrand48() % no_of_units);

    unit_ptr = *t_ptr1;
    *t_ptr1 = *t_ptr2;
    *t_ptr2 = unit_ptr;
    }

  /*  limit right side of the topologic array with NULL pointer  */
  *topo_ptr++ = NULL;

  TopoSortID = PERMUTATION;
  NetModified = FALSE;

  return( KRERR_NO_ERROR );
}


/*#################################################

GROUP: Functions for pattern management

#################################################*/

/*  set the current pattern
*/
int  kr_setCurrentPattern( pattern_no  )
int  pattern_no;
{
  if ((pattern_no < 1) || (pattern_no > NoOfPatternPairs) )
    return( -1 );  /*	Invalid pattern number	*/

  CurrentPattern = pattern_no;
  return( 0 );
}

/*  returns the current pattern
*/
int  kr_getCurrentPattern()
{
  return( CurrentPattern );
}


/*  delete the current pattern pair
*/
int   kr_deletePattern()
{
  if ((CurrentPattern < 1) || (CurrentPattern > NoOfPatternPairs) )
    return( -2 );  /*	Invalid pattern */

  if (krm_deletePatternPair( CurrentPattern ) != 0) 
    return( -1 );  /*  mem. alloc failed  */
  else
    return( 0 );
}


/*  According to the mode kr_showPattern stores the current
    Pattern into the units activation (and/or output) values.
    The modes are:
      - OUTPUT_NOTHING
         store input pattern into input units activations
      - OUTPUT_ACT
         store input pattern into input units activations and
         store output pattern into output units activations
      - OUTPUT_OUT
         store input pattern into input units activations,
         store output pattern into output units activations and
         update output units output

NOTE: See include file glob_typ.h for mode constants.
*/
void  kr_showPattern( pattern_no, mode )
int  pattern_no, mode;
{
  register struct Unit   *unit_ptr;
  register int   i;
  Patterns  in_pat,  out_pat;


  /*  calc. startaddress of pattern arrays  */
  in_pat = in_patterns + pattern_no * NoOfInputUnits;
  out_pat = out_patterns + pattern_no * NoOfOutputUnits;

  switch (mode)
    {
    case  OUTPUT_NOTHING:
      FOR_ALL_UNITS( unit_ptr )
	if ( (unit_ptr->flags & UFLAG_TTYP_IN) &&
             ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
          unit_ptr->act = *in_pat++;

      break;
    case  OUTPUT_ACT:
      FOR_ALL_UNITS( unit_ptr )
        if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
          {
	  if (unit_ptr->flags & UFLAG_TTYP_IN)
            unit_ptr->act = *in_pat++;
          else
	    if (unit_ptr->flags & UFLAG_TTYP_OUT)
              unit_ptr->act = *out_pat++;
          }

      break;
    case  OUTPUT_OUT:
      FOR_ALL_UNITS( unit_ptr )
        if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
          {
	  if (unit_ptr->flags & UFLAG_TTYP_IN)
            unit_ptr->act = *in_pat++;
          else
	    if (unit_ptr->flags & UFLAG_TTYP_OUT)
              {
              unit_ptr->act = *out_pat++;
              if (unit_ptr->out_func == NULL)
                /*  Identity Function   */
                unit_ptr->Out.output = unit_ptr->act;
              else
                unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
              }
          }

      break;
    }

}



/*  determines the network error for the given pattern and counts the output units
    with a higher error then the given delta max.

*/
krui_err  kr_checkUnitsOutput( pattern_no,
			       parameterInArray, NoOfInParams,
			       parameterOutArray, NoOfOutParams )
int  pattern_no;
float   parameterInArray[], * *parameterOutArray;
int     NoOfInParams, *NoOfOutParams;
{
  static  float  OutParameter[2];   /*	OutParameter[0] stores the learning error,
                                        OutParameter[1] stores the no. of error units  */
  register struct Unit   *unit_ptr;
  register int   i, error_units;
  register float  devit, sum_error, delta_max;
  Patterns  out_pat;


  /*  calc. startaddress of pattern array  */
  out_pat = out_patterns + pattern_no * NoOfOutputUnits;
  sum_error = 0.0;  /*  reset network error  */
  error_units = 0;  /*  reset no. of error units */
  delta_max = parameterInArray[0];

  FOR_ALL_UNITS( unit_ptr )
    if ((unit_ptr->flags & UFLAG_TTYP_OUT) &&
	((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE))
      {
      devit = *out_pat++ - unit_ptr->Out.output;  /*  calc. devitation	*/
      if ( fabs( devit ) > delta_max )
	{  /*  calc. error for output units   */
	sum_error += devit * devit;  /*  sum up the error of the network  */
	error_units++;	/*  increase no. of error units  */
      }
    }

  OutParameter[0] = sum_error * 0.5;
  OutParameter[1] = (float) error_units;
  *NoOfOutParams = 2;
  *parameterOutArray = OutParameter;   /*  set the output parameter reference  */

  return( KRERR_NO_ERROR );
}
  





/*  modify the current pattern pair
*/
int  kr_modifyPattern()
{
  register struct Unit   *unit_ptr;
  Patterns  in_pat,  out_pat;


  if ((CurrentPattern < 1) || (CurrentPattern > NoOfPatternPairs) )
    return( -1 );  /*	Invalid pattern */

  /*  calc. startaddress of pattern entries  */
  in_pat = in_patterns + (CurrentPattern - 1) * NoOfInputUnits;
  out_pat = out_patterns + (CurrentPattern - 1) * NoOfOutputUnits;

  /*  store current activation value of input/output units in pattern array  */
  FOR_ALL_UNITS( unit_ptr )
    if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {
      if (unit_ptr->flags & UFLAG_TTYP_IN)
        *in_pat++ = unit_ptr->act;
      else
	if (unit_ptr->flags & UFLAG_TTYP_OUT)
          *out_pat++ = unit_ptr->act;
      }

  return( KRERR_NO_ERROR );
}



/*  create new pattern pair
*/
int     kr_createPattern()
{
  register struct Unit   *unit_ptr;
  Patterns  in_pat,  out_pat;


  if (krm_newPattern() == 0)
    return( -1 );  /*  memory fault  */

  /*  calc. startaddress of pattern entries  */
  in_pat = in_patterns + (NoOfPatternPairs - 1) * NoOfInputUnits;
  out_pat = out_patterns + (NoOfPatternPairs - 1) * NoOfOutputUnits;

  /*  store current activation value of input/output units in pattern array  */
  FOR_ALL_UNITS( unit_ptr )
    if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {
      if (unit_ptr->flags & UFLAG_TTYP_IN)
        *in_pat++ = unit_ptr->act;
      else
	if (unit_ptr->flags & UFLAG_TTYP_OUT)
          *out_pat++ = unit_ptr->act;
      }

  return( 0 );
}


/*  permutate the order of pattern pairs
*/
int     kr_shufflePatterns()
{
  PatternNumbers  pat_numbers, ptr1, ptr2;
  int   i,  n;


  /*  create permutation array	*/
  if (krm_allocPatternNoArray() != 0)
    return( -1 );  /*  memory fault  */

  /*  initialize permutation array  */
  for (i = 0, pat_numbers = pattern_numbers; i < NoOfPatternPairs; i++)
    *pat_numbers++ = i;

  /*  permutate the numbers of the pattern pairs  */
  for (i = 0; i < (NoOfPatternPairs / 2) + 1; i++)
    {
    ptr1 = pattern_numbers + (lrand48() % NoOfPatternPairs);
    ptr2 = pattern_numbers + (lrand48() % NoOfPatternPairs);

    n = *ptr1;
    *ptr1 = *ptr2;
    *ptr2 = n;
    }

  NoOfShuffledPatterns = NoOfPatternPairs;
  return( 0 );
}



/*#################################################

GROUP: Functions for Handeling Network Propagation Functions

#################################################*/

/*  returns a pointer to the current update function
*/
UpdateFuncPtr  kr_getUpdateFunc()
{
  return( CurrUpdateFunc );
}


/*  sets the update function
*/
void  kr_setUpdateFunc( update_func_ptr )
UpdateFuncPtr  update_func_ptr;
{
  if (update_func_ptr != CurrUpdateFunc)
    NetInitialize = TRUE;   /*  learning function has changed, initialize the network  */

  CurrUpdateFunc = update_func_ptr;
}

/*  calls update function
*/
krui_err  kr_updateNet(parameterArray, NoOfParams  )
float   parameterArray[];
int  NoOfParams;
{
  int  ret_code;


  ret_code = (*kr_getUpdateFunc()) ( parameterArray, NoOfParams );
  
  return( ret_code );
}



/*#################################################

GROUP: Functions for Handeling Learning Functions

#################################################*/

/*  returns a pointer to the current learning function
*/
LearnFuncPtr  kr_getLearnFunc()
{
  return( CurrLearnFunc );
}


/*  sets the learning function
*/
void  kr_setLearnFunc( learn_func_ptr )
LearnFuncPtr  learn_func_ptr;
{
  if (CurrLearnFunc != learn_func_ptr)
    LearnFuncHasChanged = TRUE;   /*  Learning function has changed  */

  CurrLearnFunc = learn_func_ptr;
}


/*  calls learning function
*/
krui_err  kr_learnNet( start_pattern, end_pattern, parameterInArray, NoOfInParams, parameterOutArray, NoOfOutParams )
int  start_pattern, end_pattern;
float   parameterInArray[],  * *parameterOutArray;
int  NoOfInParams, *NoOfOutParams;
{
  int ret_code;

  /*  call current learning function  */
  ret_code = (*kr_getLearnFunc()) ( start_pattern, end_pattern, 
                                    parameterInArray, NoOfInParams, 
                                    parameterOutArray, NoOfOutParams );

  if (ret_code == KRERR_NO_ERROR)  
    {
    NetInitialize = FALSE;  /*  learning function has initialized the network  */
    LearnFuncHasChanged = FALSE; 
    }

  return( ret_code );
}



/*#################################################

GROUP: Functions for Handeling Network Initialisation Functions

#################################################*/

/*  returns a pointer to the current initialisation function
*/
InitFuncPtr  kr_getInitialisationFunc()
{
  return( CurrInitFunc );
}


/*  sets the initialisation function
*/
void  kr_setInitialisationFunc( init_func_ptr )
InitFuncPtr  init_func_ptr;
{
  CurrInitFunc = init_func_ptr;
}


/*  initializes the network
*/
krui_err   kr_initializeNet( parameterArray, NoOfParams )
float   parameterArray[];
int  NoOfParams;
{
  int  ret_code;

  NetInitialize = TRUE;
  ret_code = (*kr_getInitialisationFunc()) ( parameterArray, NoOfParams );
  
  return( ret_code );
}
