/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS Kernel Function Prototypes
NOTES  : 
AUTHOR : Niels Mache
DATE   : 20.02.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#ifdef __STDC__

/*  delete all inputs at the given unit
*/
void kr_deleteAllInputs( struct Unit  *unit_ptr );

/*  Deletes all output links at <source_unit>
NOTE: This function is slow.
*/
void kr_deleteAllOutputLinks( struct Unit  *source_unit_ptr );

/*  Copies all input links. Link list must not be NULL !
    Returns NULL if memory allocation fails.
*/
struct Link *kr_copyLinkList( struct Link   *link_list );

/*  Copies all sites with input links. Site list must not be NULL !
    Returns NULL if memory allocation fails.
*/
struct Site *kr_copySiteListAndLinks( struct Site   *site_list );

/*  Copies all sites. Site list must not be NULL !
    Returns NULL if memory allocation fails.
*/
struct Site *kr_copySiteList( struct Site   *site_list );

/*  Copies all output links at <source_unit> to <new_unit>.
    Returns error code if memory allocation fails.
NOTE: This function is slow.
*/
int kr_copyOutputLinks( struct Unit  *source_unit_ptr, struct Unit  *new_unit_ptr );

/*  search for a site at a unit 
*/
struct Site *kr_searchUnitSite( struct Unit  *unit_ptr, struct SiteTable  *stbl_ptr );

/*  search for a site at a unit and returns also the predecessor site ptr 
*/
void kr_searchUnitSiteWithPred( struct Unit  *unit_ptr, struct SiteTable  *stbl_ptr,
                                struct Site  * *Rsite_ptr, struct Site  * *Rsite_pred_ptr );

/*  searches for a site in the network  
*/
int kr_searchNetSite( struct SiteTable  *stbl_ptr );

/*  Creates a new site with default initialisation
*/
struct Site *kr_createDefaultSite( void );

/*  returns the pointer to the given unit, returns NULL if unit doesn't
    exist   
*/
struct Unit *kr_getUnitPtr( int  unit_no );

/*  Searches for a unit with the given symbol pointer.
    Returns the first unit no. if a unit with the given name was found,
    0 otherwise.
*/
int kr_unitNameSearch( int  min_unit_no, char  *unit_symbol_ptr );

/*  Initializes a unit with default values.
*/
void kr_makeDefaultUnit( struct Unit  *unit_ptr );

/*  copy the source unit, but not the flags of the source unit
*/
void kr_copyUnit( struct  Unit  *source_unit_ptr, struct  Unit  *new_unit_ptr );

/*  changes all units in the network with the given functionality type
    to the new functions of the (new) functionality type
*/
void kr_changeFtypeUnits( struct  FtypeUnitStruct  *Ftype_entry );

/*   delete the functionality type of the units with the given type
*/
void kr_deleteUnitsFtype( struct  FtypeUnitStruct  *ftype_ptr );

/*  create a new unit with the given functionality type
*/
struct Unit *kr_makeFtypeUnit(struct Unit  *unit_ptr, struct  FtypeUnitStruct  *ftype_ptr );

/*  returns TRUE, if there exists the given site at the given ftype entry
*/
int kr_FtypeSiteSearch( struct  Site  *ftype_first_site, struct  SiteTable  *site_table_ptr );

/*  change the properties of the given unit to the properties of the
    given F-Type
*/
void kr_changeFtypeUnit( struct  Unit  *unit_ptr, struct  FtypeUnitStruct  *ftype_ptr );

/*  change a site at the F-Type
*/
void kr_changeFtypeSites( struct  FtypeUnitStruct  *Ftype_entry, 
                          struct  SiteTable  *old_site_table,
                          struct  SiteTable  *new_site_table );

/*  spell checker  (check identifiers for matching [A-Za-z]^[|, ]*)
*/
int kr_symbolCheck( char    *symbol );

/*  translate unit flags to the topological type of the unit
*/
int kr_getTType( int  flags );

/*  translate the topological type to unit flags
*/
int kr_setTType( int  ttype );

/*  update the outputs of all units in the network
*/
void kr_updateUnitOutputs( void );

/*  returns the no. of units of the specified topologic type
    (i.e. Input, Hidden, Output or Special units)
*/
int kr_getNoOfUnits( int  UnitTType );

/*  force unit array garbage collection
*/
void  kr_forceUnitGC( void );


/*#################################################

GROUP: Topological Sorting Functions

#################################################*/


/*  Sort units according to the given mode:
    TOPOLOGICAL:
      Sort units topological (general version) and stores the
      pointers to this units in the topologic array.
      NOTE: Units are not sorted by their topologic type (that's not
	    possible in general case).

    TOPOLOGICAL_FF:
      Sorts unit topological in feed-forward networks and stores the
      pointers to this units in the topologic array in the following order:
       - input,
       - hidden and
       - output units

      This function make following assumtions (like all learning functions for
      feed-forward networks):
       a) input units doesn't have input connections to other units and
       b) output units doesn't have outputs connections to other units.

    TOPOLOGIC_TYPE:
      Sort units by their topologic type, i.e. Input, Hidden, Output units and
      stores the pointers to this units in the topologic array.
*/
krui_err kr_topoSort( int sort_mode );


/*  Checks the topology of the network:
     a) counts the number of layers of the network and
     b) determines if the network has cycles.

    Returns the no. of layers of the network.
*/
int kr_topoCheck( void );

/*  Count the no. of input and output units and return an error code
    if the no. do not fit to the loaded patterns.
*/
krui_err  kr_IOCheck( void );


krui_err  kr_makeUnitPermutation( void );

/*#################################################

GROUP: Functions for pattern management

#################################################*/

/*  set the current pattern
*/
int kr_setCurrentPattern( int  pattern_no );

/*  returns the current pattern
*/
int kr_getCurrentPattern( void );

/*  delete the current pattern pair
*/
int kr_deletePattern( void );

/*  modify the current pattern pair
*/
int kr_modifyPattern( void );

/*  create new pattern pair
*/
int kr_createPattern( void );

/*  permutate the order of pattern pairs
*/
int kr_shufflePatterns( void );

/*  According to the mode kr_showPattern stores the current
    Pattern into the units activation (and/or output) values.
    The modes are:
      - OUTPUT_NOTHING
         store input pattern into input units activations
      - OUTPUT_ACT
         store input pattern into input units activations and
         store output pattern into output units activations
      - OUTPUT_OUT
         store input pattern into input units activations,
         store output pattern into output units activations and
         update output units output

NOTE: See include file glob_typ.h for mode constants.
*/
void kr_showPattern( int  pattern_no, int mode );

/*  determines the network error for the given pattern and counts the output units
    with a higher error then the given delta max.

*/
krui_err kr_checkUnitsOutput( int  pattern_no,
			      float  parameterInArray[], int  NoOfInParams,
			      float * *parameterOutArray, int  *NoOfOutParams );


/*  Add random uniform distributed values to connection weights.
    <minus> must be less then <plus>.
*/
void kr_jogWeights( FlintTypeParam  minus, FlintTypeParam  plus );

/*  returns a pointer to the current learning function
*/
LearnFuncPtr kr_getLearnFunc( void );

/*  returns a pointer to the current update function
*/
UpdateFuncPtr kr_getUpdateFunc( void );

/*  returns a pointer to the current initialisation function
*/
InitFuncPtr kr_getInitialisationFunc( void );

/*  sets the learning function
*/
void kr_setLearnFunc( LearnFuncPtr  learn_func_ptr );

/*  sets the update function
*/
void kr_setUpdateFunc( UpdateFuncPtr  update_func_ptr );

/*  sets the initialisation function
*/
void kr_setInitialisationFunc( InitFuncPtr  init_func_ptr );

/*  initializes the network
*/
int kr_initializeNet( float   parameterArray[], int  NoOfParams );

/*  calls learning function
*/
int kr_learnNet( int  start_pattern, int end_pattern, 
                 float  parameterInArray[], int  NoOfInParams,
                 float  * *parameterOutArray, int *NoOfOutParams );

/*  calls update function
*/
int kr_updateNet( float   parameterArray[], int  NoOfParams );

/*  Returns information about the unit default settings.
*/
void kr_getUnitDefaults( FlintType  *act, FlintType  *bias, int	 *ttflags, int  *subnet_no,
                         int  *layer_no, ActFuncPtr  *act_func_ptr, OutFuncPtr  *out_func_ptr );

/*  Changes the unit default settings.
*/
void kr_setUnitDefaults( FlintTypeParam  act, FlintTypeParam  bias, int  ttflags, int  subnet_no,
                         int  layer_no, ActFuncPtr  act_func_ptr, ActDerivFuncPtr  act_deriv_func_ptr,
                         OutFuncPtr  out_func_ptr );


#else

void kr_deleteAllInputs();
void kr_deleteAllOutputLinks();
struct Link *kr_copyLinkList();
struct Site *kr_copySiteListAndLinks();
struct Site *kr_copySiteList();
int kr_copyOutputLinks();
struct Site *kr_searchUnitSite();
void kr_searchUnitSiteWithPred();
int kr_searchNetSite();
struct Site *kr_createDefaultSite();
struct Unit *kr_getUnitPtr();
int kr_unitNameSearch();
void kr_makeDefaultUnit();
void kr_copyUnit();
void kr_changeFtypeUnits();
void kr_deleteUnitsFtype();
struct Unit *kr_makeFtypeUnit();
int kr_FtypeSiteSearch();
void kr_changeFtypeUnit();
void kr_changeFtypeSites();
int kr_symbolCheck();
int kr_getTType();
int kr_setTType();
void kr_updateUnitOutputs();
int kr_getNoOfUnits();
void  kr_forceUnitGC();

krui_err  kr_topoSort();
int kr_topoCheck();
krui_err  kr_IOCheck();
krui_err  kr_makeUnitPermutation();

int kr_setCurrentPattern();
int kr_getCurrentPattern();
void kr_showPattern();
krui_err kr_checkUnitsOutput();
int kr_deletePattern();
int kr_modifyPattern();
int kr_createPattern();
int kr_shufflePatterns();

void kr_jogWeights();
LearnFuncPtr kr_getLearnFunc();
UpdateFuncPtr kr_getUpdateFunc();
InitFuncPtr kr_getInitialisationFunc();
void kr_setLearnFunc();
void kr_setUpdateFunc();
void kr_setInitialisationFunc();

int kr_initializeNet();
int kr_learnNet();
int kr_updateNet();

void kr_setUnitDefaults();
void kr_getUnitDefaults();

#endif
