/*     %W%  %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel User Interface File I/O
NOTES  : 
AUTHOR : Niels Mache
DATE   : 28.05.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include <varargs.h>

#include "kr_typ.h"	 /*  Kernel Types and Constants  */
#include "kr_const.h"	 /*  Constant Declarators for SNNS-Kernel  */
#include "kr_def.h"	 /*  Default Values  */
#include "kr_ui.h"	 /*  Interface Functions  */
#include "kernel.h"      /*  Function Prototypes */
#include "kr_mem.h"      /*  Function Prototypes */
#include "kr_io.h"	 /*  Function Prototypes */
#include "kr_mac.h"	 /*  Kernel Macros  */
#include "version.h"	 /*  Version and Patchlevel  */


#ifdef ultrix
#define retchk( ret_code )  if (ret_code == EOF)  return( -1 )
#define RETCHKGTO( ret_code )  if (ret_code == EOF)  goto end 
#else
#define retchk( ret_code )  if (ret_code < 0)  return( ret_code )
#define RETCHKGTO( ret_code )  if (ret_code < 0)  goto end
#endif



/*#################################################

GROUP: Global Var's (as declared by the kernel)

#################################################*/

extern bool  PatternsValid;	  /*  TRUE, if In-Core patterns are valid  */

extern	int  NoOfUnits, 	  /*  no. of units in the network  */
	     MinUnitNo, 	  /*  the first (lowest) used unit number in the network  */
	     MaxUnitNo, 	  /*  the last (highest) used unit number in the network  */
	     NoOfPatternPairs,	  /*  no. of defined pattern pairs */
	     NoOfInputUnits,	  /*  no. of input units  */
	     NoOfOutputUnits;	  /*  no. of output units  */

extern	UnitArray    unit_array;  /*  the unit array  */

extern	Patterns	in_patterns,   /*  the input pattern array  */
			out_patterns;  /*  the output pattern array  */

extern int  lineno,    /*  file I/O: line no. in network file, if network loading failed  */
	    io_err;    /*  file I/O: operatin system I/O error value, if network loading failed  */



/*#################################################

GROUP: Local Var's

#################################################*/

#define LIN_MAX      (MAX_LIN_LEN + 1)

#define SITE_DEF     1
#define TYPE_DEF     2
#define DEFAULT_DEF  3
#define UNIT_DEF     4
#define CONNECT_DEF  5
#define SUBNET_DEF   6
#define LAYER_DEF    7


#define EOS	'\0'	/*  End Of String  */


typedef  char   SymbolType[LIN_MAX];

#ifdef KERNEL3D
static struct TransTable  xyTransTable[ TRANS_TABLE_SIZE ];
#endif


static FILE            *io_stream;

static char            * *site_name_ptrs = NULL;
static SymbolType      *site_names       = NULL;


static bool  is_subnet_info,
	     is_layer_info,
	     units_have_sites;

static int   site_name_len,
	     unit_name_len,
	     type_name_len,
	     site_func_len,
	     act_func_len,
	     out_func_len,
	     def_act_func_len,
	     def_out_func_len,
             unit_no_len,
             max_connects_per_line,
             max_subnets_per_line,
             max_layers_per_line,
             pos_no_len,
             subnet_no_len,
             layer_no_len,
             NoOfLinks,
             NoOfSiteTypes,
             NoOfUnitTypes;



static char    fmt_shape1 [ LIN_MAX ],
               fmt_shape2 [ LIN_MAX ],
               fmt_shape3 [ LIN_MAX ],
               fmt_hdr1   [ LIN_MAX ],
               fmt_hdr2   [ LIN_MAX ],
               fmt_blank  [ LIN_MAX ];


static char    *title[] = {
    "SNNS network definition file",
    "generated at",
    "network name",
    "no. of units",
    "no. of connections",
    "no. of unit types",
    "no. of site types",
    "learning function",
    "site definition section",
    "type definition section",
    "unit definition section",
    "connection definition section",
    "subnet definition section",
    "unit default section",
    "source files",
    "layer definition section",
    "update function",
    "3D translation section"
    };   /*  17  */

static int  NoOfTitles = (sizeof title) / (sizeof title[0]);


static char    *patHeader[] = {
    "SNNS pattern definition file %s\n",
    "generated at",
    "No. of patterns     : %d\n",
    "No. of input units  : %d\n",
    "No. of output units : %d\n"
    };


static char     *headers[] = {
    " site name | site function",
    " name | act func | out func | sites",
    " act | bias | st | subnet | layer | act func | out func",
    " no. | typeName | unitName | act | bias | st | position | act func | out func | sites",
    " target | site | source:weight",
    " subnet | unitNo.",
    " layer | unitNo.",
    " delta x | delta y | z"
    };



/*  some useful functions
*/

/*  mstrcat is called
     mstrcat( dest, src1, src2, ... , (char *) 0 )
*/
static void  mstrcat( va_alist )
va_dcl
{
  va_list  args;
  char	  *dest, *src;


  va_start( args );
  if ((dest = va_arg( args, char * )) == (char *) 0)  return;

  while ((src = va_arg( args, char * )) != (char *) 0)
    strcat( dest, src );

  va_end( args );
}


/*  mstrcat is called
     mstrcat( dest, src1, src2, ... , (char *) 0 )
*/
static void  mstrcpy( va_alist )
va_dcl
{
  va_list  args;
  char	  *dest, *src;


  va_start( args );
  if ((dest = va_arg( args, char * )) == (char *) 0)  return;
  if ((src  = va_arg( args, char * )) == (char *) 0)  return;
  while( *src != EOS )
    *dest++ = *src++;

  while ((src = va_arg( args, char * )) != (char *) 0)
    {
    while( *src != EOS )
      *dest++ = *src++;
  }

  va_end( args );

  *dest = EOS;
}



/*  returns the current file I/O version number
*/
static char  *krio_getIOVersion()
{
  static char	io_version[128];

  strcpy( io_version, NETFILE_VERSION );
#ifdef KERNEL3D
  strcat( io_version, KERNEL3D_NETFILE_VERSION );
#endif

  return( io_version );
}


/*#################################################

GROUP: Kernel File Output Functions

#################################################*/

static krui_err  krio_writeHeader( stream, version, net_name, learning_func, update_func )
FILE    *stream;
char    *net_name,
        *version,
        *learning_func,
        *update_func;
{
  long  clock;
  int   err;
  int   no_of_sites,
        no_of_STable_entries,
        no_of_FTable_entries;

  krui_getNetInfo( &no_of_sites, &NoOfLinks, &no_of_STable_entries,
		   &no_of_FTable_entries );
  units_have_sites = no_of_sites > 0;

  clock = 1;
  (void) time( &clock);

  err = fprintf( stream, "%s %s\n%s %s\n%s : ",
                title[0], version, title[1], ctime( &clock), title[2] );
  retchk( err );

  if (net_name == NULL)
    err = fprintf( stream, "UNTITLED\n" );
  else
    err = fprintf( stream, "%s\n", net_name );

  retchk( err );
  err = fprintf( stream, "%s :\n%s : %u\n%s : %u\n%s : %u\n%s : %u\n",
                title[14],
                title[3], NoOfUnits,
                title[4], NoOfLinks,
                title[5], no_of_FTable_entries,
                title[6], no_of_STable_entries );

  retchk( err );
  err = fprintf( stream, "\n\n%s : %s\n",
                title[7], learning_func );
  retchk( err );
  err = fprintf( stream, "%s   : %s\n",
                title[16], update_func );
  retchk( err );

  return( 0 );
}

/*  cut all trailing '0' 
*/
static void krio_cutTrailingZeros( string )
char *string;
{
    int j;


    if (*string == EOS)  return;
    for (j = strlen( string ) - 1; ((string[j] == '0') && (j > 0)); j--) {}

    if (string[j] == '.')
      string[j] = EOS;
    else
      string[j + 1] = EOS;
}


static  char    *krio_repchar( c, N )
char    c;
int     N;
{
  int   i;
  char  str[180];

  if ( (N - 1) >= sizeof (str))   N = sizeof (str) - 1;

  for (i = 0; i < N; i++)
    str[ i ] = c;

  str[N] = 0;
  return( str );
}



static void    krio_stringLimits()
{
  char  *symbol, *name;
  int   sym_type,
        unit_no,
	len, i1, i2, i3,
	subnet_no, layer_no;
  FlintType  act, bias;
  int	st;
  char	*def_act_func,	*def_out_func;

  struct  PosType   pos;
  bool	  neg1, neg2, neg3;


  unit_name_len = site_name_len = type_name_len = 0;
  out_func_len = act_func_len = site_func_len = 0;
  def_out_func_len = def_act_func_len = 0;

  /*  get default subnet and layer numbers  */
  krui_getUnitDefaults( &act, &bias, &st, &subnet_no, &layer_no,
			&def_act_func, &def_out_func );

  if ( krui_getFirstSymbolTableEntry( &symbol, &sym_type) )
    {
    do
      {
      len = strlen( symbol );

      switch (sym_type)
        {
        case  UNIT_SYM:
            unit_name_len = Max( unit_name_len, len );
            break;
        case  SITE_SYM:
            site_name_len = Max( site_name_len, len );
            break;
        case  FTYPE_UNIT_SYM:
            type_name_len = Max( type_name_len, len );
      }
    }
    while ( krui_getNextSymbolTableEntry( &symbol, &sym_type ) );
  }

  unit_no = krui_getFirstUnit();
  do
    {
    if ((symbol = krui_getUnitOutFuncName( unit_no )) != def_out_func)
      out_func_len = Max( out_func_len, strlen( symbol ));

    if ((symbol = krui_getUnitActFuncName( unit_no )) != def_act_func)
      act_func_len = Max( act_func_len, strlen( symbol ));
  }
  while ((unit_no = krui_getNextUnit() ) != 0);

  def_out_func_len = strlen( def_out_func );
  def_act_func_len = strlen( def_act_func );

  if (krui_getFirstSiteTableEntry( &name, &symbol ))
    {  do
	 site_func_len = Max( site_func_len, strlen( symbol ));
      while (krui_getNextSiteTableEntry( &name, &symbol ));
    }

  unit_no = krui_getFirstUnit();
  neg1 = FALSE;
  neg2 = FALSE;
  is_subnet_info = FALSE;
  neg3 = FALSE;
  is_layer_info = FALSE;
  i1 = 0;  i2 = 0;  i3 = 0;

  do
    {
    krui_getUnitPosition( unit_no, &pos );
    i1 = Max( i1, abs( pos.x ) );
    i1 = Max( i1, abs( pos.y ) );
    if (pos.x < 0) neg1 = TRUE;
    if (pos.y < 0) neg1 = TRUE;

    i2 = krui_getUnitSubnetNo( unit_no );
    if (i2 < 0) neg2 = TRUE;
    i2 = Max( i2, abs( krui_getUnitSubnetNo( unit_no )) );
    if (i2 != subnet_no)
      is_subnet_info = TRUE;

    i3 = (int) krui_getUnitLayerNo( unit_no );
    if (i3 < 0) neg2 = TRUE;
    i3 = Max( i3, abs( (int) krui_getUnitLayerNo( unit_no )) );
    if (i3 != layer_no)
      is_layer_info = TRUE;

  }
  while ( (unit_no = krui_getNextUnit() ) > 0);

  pos_no_len = 1;
  while ( (i1 /= 10) != 0)
    ++pos_no_len;
  if (neg1)  pos_no_len++;

  subnet_no_len = 1;
  while ( (i2 /= 10) != 0)
    ++subnet_no_len;
  if (neg2)  subnet_no_len++;

  layer_no_len = 1;
  while ( (i3 /= 10) != 0)
    ++layer_no_len;
  if (neg3)  layer_no_len++;


  len = NoOfUnits;
  unit_no_len = 1;
  while ( (len /= 10) != 0)
    ++unit_no_len;

}



static krui_err    krio_fmtShapeing( choose_me )
int     choose_me;
{
  static char   *formats[] = {
                /*  0   */   "site name",
                /*  1   */   "sites",
                /*  2   */   "unitName",
                /*  3   */   "typeName",
                /*  4   */   "name",
                /*  5   */   "site function",
                /*  6   */   "act func",
                /*  7   */   "out func",
                /*  8   */   "no.",
                /*  9   */   "target",
                /*  10  */   "act",
                /*  11  */   "bias",
                /*  12  */   "position",
                /*  13  */   "site",
                /*  14  */   " source:weight",
                /*  15  */   "subnet",
                /*  16  */   "layer",
                /*  17  */   " unitNo."
                             };

  char  fmt [ MAX_LIN_LEN ],
        aux [ MAX_LIN_LEN ];

  int	len, len1, act_len, out_len;


  switch( choose_me )
    {
    case  SITE_DEF:
        if (site_name_len + site_func_len + 5 > MAX_LIN_LEN)
          return( KRERR_SAVE_LINE_LEN );
        len = Max( site_name_len, strlen( formats[ 0 ] ) );
        sprintf( aux, "%d", len );
	mstrcpy( fmt_shape1, " %-", aux, "s | %s\n", (char *) 0 );

	mstrcpy( fmt, " %-", aux, "s | ", (char *) 0 );
	sprintf( fmt_hdr1, fmt, formats[ 0 ] );
	mstrcat( fmt_hdr1, formats[ 5 ], "\n", (char *) 0 );

	mstrcpy( fmt_hdr2 ,krio_repchar( '-', len + 2 ), "|", (char *) 0 );

	len = Max( site_func_len, strlen( formats[ 6 ] ));
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "\n", (char *) 0 );

        break;

    case  TYPE_DEF:
	if (act_func_len == 0)	act_len = def_act_func_len;
	else  act_len = act_func_len;
	if (out_func_len == 0)	out_len = def_out_func_len;
	else  out_len = out_func_len;

	len = act_len + out_len + site_name_len + type_name_len + 11;
	if (len > MAX_LIN_LEN)	return( KRERR_SAVE_LINE_LEN );

	len = Max( type_name_len, strlen( formats[ 4 ] ));
	len1 = len + 2;
	sprintf( aux, "%d", len );
	mstrcpy( fmt_shape1, "%-", aux, "s |", (char *) 0 );
	sprintf( fmt_hdr1, fmt_shape1, formats[ 4 ] );
	mstrcpy( fmt_hdr2, krio_repchar( '-', len + 1 ), "|", (char *) 0 );


	len = Max( act_len, strlen( formats[ 6 ] ) );
        len1 = len1 + len + 3;
        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
        strcat( fmt_shape1, fmt );
        sprintf( aux, fmt, formats[ 6 ] );
	strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );

	len = Max( out_len, strlen( formats[ 7 ] ) );
        len1 = len1 + len + 3;
        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
	strcat( fmt_shape1, fmt );

        sprintf( aux, fmt, formats[ 7 ] );
	mstrcat( fmt_hdr1, aux, formats[1], "\n", (char *) 0 );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );
	mstrcpy( fmt_blank, ",\n", krio_repchar( ' ', len1 ), (char *) 0 );

        len = Max( site_name_len, strlen( formats[ 1 ] ) );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "\n", (char *) 0 );

        break;

    case  DEFAULT_DEF:
        len = Max( 8, strlen( formats[ 10 ] ) );
        len1 = len + 2;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        strcpy( fmt_shape1, "%8.5f |" );

        sprintf( aux, "%d", len );
	mstrcpy( fmt, "%-", aux, "s |", (char *) 0 );
	sprintf( fmt_hdr1, fmt, formats[ 10 ] );

	mstrcpy( fmt_hdr2, krio_repchar( '-', len + 1 ), "|", (char *) 0 );

        len = Max( 8, strlen( formats[ 11 ] ) );
        len1 = len1 + len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        strcat( fmt_shape1, " %8.5f |" );

        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
        sprintf( aux, fmt, formats[ 11 ] );
	strcat( fmt_hdr1, aux );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );

        len1 = len1 + 5;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        strcat( fmt_shape1, " %s  |" );
        strcat( fmt_hdr1, " st |" );
	strcat( fmt_hdr2, "----|" );

        len = Max( subnet_no_len, strlen( formats[ 15 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );
	mstrcat( fmt_shape1, " %", aux, "d |", (char *) 0 );

	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
        sprintf( aux, fmt, formats[ 15 ] );
	strcat( fmt_hdr1, aux );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );

        len = Max( layer_no_len, strlen( formats[ 16 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );
	mstrcat( fmt_shape1, " %", aux, "d |", (char *) 0 );

	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
        sprintf( aux, fmt, formats[ 16 ] );
	strcat( fmt_hdr1, aux );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );

	len = Max( def_act_func_len, strlen( formats[ 6 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s | ", (char *) 0 );
        strcat( fmt_shape1, fmt );
        sprintf( aux, fmt, formats[ 6 ] );
	mstrcat( fmt_hdr1, aux, formats[7], "\n", (char *) 0 );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


	len = Max( def_out_func_len, strlen( formats[ 7 ] ) );
	len1 += len + 1;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );
	mstrcat( fmt_shape1, "%-", aux, "s \n", (char *) 0 );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 1 ), "\n", (char *) 0 );

        break;

    case  UNIT_DEF:
        len = Max( unit_no_len, strlen( formats[ 8 ] ) );
	len1 = len + 2;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );

	sprintf( aux, "%d", len );
	mstrcpy( fmt_shape1, "%", aux, "d |", (char *) 0 );

	mstrcpy( fmt, "%", aux, "s |", (char *) 0 );
	sprintf( fmt_hdr1, fmt, formats[ 8 ] );

	mstrcpy( fmt_hdr2, krio_repchar( '-', len + 1 ), "|", (char *) 0);


        len = Max( type_name_len, strlen( formats[ 3 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );

	sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
	strcat( fmt_shape1, fmt );

	sprintf( aux, fmt, formats[ 3 ] );
        strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


        len = Max( unit_name_len, strlen( formats[ 2 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );

	sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
	strcat( fmt_shape1, fmt );

	sprintf( aux, fmt, formats[ 2 ] );
	strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


        len = Max( 8, strlen( formats[ 10 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );

        strcat( fmt_shape1, " %8.5f |" );
        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );

        sprintf( aux, fmt, formats[ 10 ] );
        strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


        len = Max( 8, strlen( formats[ 11 ] ) );
        len1 = len1 + len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        strcat( fmt_shape1, " %8.5f |" );
        sprintf( aux, "%d", len );

	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );

        sprintf( aux, fmt, formats[ 11 ] );
        strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


	len1 += 5;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
	strcat( fmt_shape1, " %s  |" );
	strcat( fmt_hdr1, " st |" );
	strcat( fmt_hdr2, "----|" );

#ifndef KERNEL3D
        len = Max( pos_no_len * 2 + 1, strlen( formats[ 12 ] ) );
#else
        len = Max( pos_no_len * 3 + 2, strlen( formats[ 12 ] ) );
#endif
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", pos_no_len );
	mstrcpy( fmt, " %", aux, "d,", (char *) 0 );
        strcat( fmt_shape1, fmt );

        sprintf( aux, "%d", pos_no_len );
	mstrcpy( fmt, "%", aux, "d", (char *) 0 );

#ifdef KERNEL3D
        strcat( fmt_shape1, fmt );
        sprintf( aux, "%d", pos_no_len );
	mstrcpy( fmt, ",%", aux, "d", (char *) 0 );
#endif
#ifndef KERNEL3D
        if ((pos_no_len * 2 + 1) < len)
          strcat( fmt, krio_repchar( ' ', len - (pos_no_len * 2 + 1)) );
#else
        if ((pos_no_len * 3 + 2) < len)
          strcat( fmt, krio_repchar( ' ', len - (pos_no_len * 3 + 2)) );
#endif

        strcat( fmt, " |" );
        strcat( fmt_shape1, fmt );

        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );

        sprintf( aux, fmt, formats[ 12 ] );
        strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


        len = Max( act_func_len, strlen( formats[ 6 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
	strcpy( fmt_shape2, fmt_shape1 );
	strcat( fmt_shape2, "||" );
        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
	strcat( fmt_shape1, fmt );

        sprintf( aux, fmt, formats[ 6 ] );
        strcat( fmt_hdr1, aux );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );


        len = Max( out_func_len, strlen( formats[ 7 ] ) );
	len1 += len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );
	mstrcpy( fmt, " %-", aux, "s | ", (char *) 0 );
	strcat( fmt_shape1, fmt );

        sprintf( aux, fmt, formats[ 7 ] );
	mstrcat( fmt_hdr1, aux, formats[1], "\n", (char *) 0 );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );
	mstrcpy( fmt_blank, ",\n", krio_repchar( ' ', len1 ), (char *) 0 );


        len = Max( site_name_len, strlen( formats[ 1 ] ) );
	len1 += len + 2;
	if (len1 + len > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "\n", (char *) 0 );

        break;

    case  CONNECT_DEF:
        sprintf( aux, "%d", unit_no_len );
	mstrcpy( fmt_shape3, " %", aux, "d:%8.5f", (char *) 0 );

        len = Max( unit_no_len, strlen( formats[ 9 ] ) );
        len1 = len + 2;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );
	mstrcpy( fmt_shape1, "%", aux, "d |", (char *) 0 );

	mstrcpy( fmt, "%", aux, "s |", (char *) 0 );
	sprintf( fmt_hdr1, fmt, formats[ 9 ] );
	mstrcpy( fmt_hdr2, krio_repchar( '-', len + 1 ), "|", (char *) 0 );
        strcpy( fmt_shape2, krio_repchar( ' ', len1 ) );

        len = Max( site_name_len, strlen( formats[ 13 ] ) );
        len1 = len1 + len + 3;
	if (len1 > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        sprintf( aux, "%d", len );


	mstrcpy( fmt, " %-", aux, "s |", (char *) 0 );
        strcat( fmt_shape1, fmt );
        strcat( fmt_shape2, fmt );
        sprintf( aux, fmt, formats[ 13 ] );
	mstrcat( fmt_hdr1, aux, formats[14], "\n", (char *) 0 );

	mstrcat( fmt_hdr2, krio_repchar( '-', len + 2 ), "|", (char *) 0 );
	mstrcpy( fmt_blank, "\n", krio_repchar( ' ', len1 ), (char *) 0 );

	if ((len1 + unit_no_len + 14) > MAX_LIN_LEN)  return( KRERR_SAVE_LINE_LEN );
        max_connects_per_line = (BEST_LINE_LEN - len1) / (unit_no_len + 11);
	mstrcat( fmt_hdr2,
		 krio_repchar( '-', (unit_no_len + 11) * max_connects_per_line ),
		 "\n", (char *) 0 );

        break;

      case  SUBNET_DEF:
        len = Max( subnet_no_len, strlen( formats[ 15 ] ) );
	len1 = len + 1;
        sprintf( aux, "%d", len );
	mstrcpy( fmt_shape1, "\n%", aux, "d |", (char *) 0 );

        sprintf( aux, "%d", len );
	mstrcpy( fmt, "%", aux, "s |", (char *) 0 );
        sprintf( fmt_hdr1, fmt, formats[ 15 ] );
        strcat( fmt_hdr1, formats[17] );
	mstrcpy( fmt_hdr2, "\n", krio_repchar( '-', len1 ), "|", (char *) 0 );
	mstrcpy( fmt_blank, "\n", krio_repchar( ' ', len1 + 1 ), (char *) 0 );

        sprintf( aux, "%d", unit_no_len );
	mstrcat( fmt_shape1, " %", aux, "d", (char *) 0 );
	mstrcpy( fmt_shape2, " %", aux, "d", (char *) 0 );

        max_subnets_per_line = (BEST_LINE_LEN - len1 - 1) / (unit_no_len + 2);
        strcat( fmt_hdr2, krio_repchar( '-', (unit_no_len + 2) * max_subnets_per_line ) );

        break;

      case  LAYER_DEF:
        len = Max( layer_no_len, strlen( formats[ 16 ] ) );
	len1 = len + 1;
        sprintf( aux, "%d", len );
	mstrcpy( fmt_shape1, "\n%", aux, "d |", (char *) 0 );

	mstrcpy( fmt, "%", aux, "s |", (char *) 0 );
        sprintf( fmt_hdr1, fmt, formats[ 16 ] );
        strcat( fmt_hdr1, formats[17] );

	mstrcpy( fmt_hdr2, "\n", krio_repchar( '-', len1 ), "|", (char *) 0 );
	mstrcpy( fmt_blank, "\n", krio_repchar( ' ', len1 + 1 ), (char *) 0 );

        sprintf( aux, "%d", unit_no_len );
	mstrcat( fmt_shape1, " %", aux, "d", (char *) 0 );
	mstrcpy( fmt_shape2, " %", aux, "d", (char *) 0 );

        max_layers_per_line = (BEST_LINE_LEN - len1 - 1) / (unit_no_len + 2);
        strcat( fmt_hdr2, krio_repchar( '-', (unit_no_len + 2) * max_layers_per_line ) );

        break;
  }

  return( 0 );
}




static krui_err    krio_writeSiteDefinitions( stream )
FILE    *stream;
{
  int   err;
  char  *site_name,
        *site_func;


  if ( !krui_getFirstSiteTableEntry( &site_name, &site_func ) )
    return( 0 );

  err = krio_fmtShapeing( SITE_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[8] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  do
    {
    err = fprintf( stream, fmt_shape1,
                   site_name, site_func );
    retchk( err );
  }
  while (krui_getNextSiteTableEntry( &site_name, &site_func ) );

  err = fprintf( stream, fmt_hdr2 );
  retchk( err );
  return( 0 );
}




static krui_err    krio_writeTypeDefinitions( stream )
FILE    *stream;
{
  int   err;
  bool  second;


  if ( !krui_setFirstFTypeEntry() )
    return( 0 );

  err = krio_fmtShapeing( TYPE_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[9] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  do
    {
    err = fprintf( stream, fmt_shape1,
                   krui_getFTypeName(),
                   krui_getFTypeActFuncName(), krui_getFTypeOutFuncName() );
    retchk( err );

    if ( krui_setFirstFTypeSite() )
      {
      second = FALSE;
      do
        {
        if (second)
          {
          err = fprintf( stream, fmt_blank );
          retchk( err );
	}

        err = fprintf( stream, " %-s", krui_getFTypeSiteName() );
        retchk( err );
        second = TRUE;
      }
      while ( krui_setNextFTypeSite() );
    }

    err = fprintf( stream, "\n" );
    retchk( err );
  }
  while (krui_setNextFTypeEntry() );

  err = fprintf( stream, fmt_hdr2 );
  if (err <= 0)  return( err );

  return( 0 );
}


static  char    *getTType( st )
int     st;
{
  static char   *ttype[] = { "-", "i", "o", "d", "h", "s" };

  switch ( st )
    {
    case  INPUT:
      return( ttype[1] );
    case  OUTPUT:
      return( ttype[2] );
    case  DUAL:
      return( ttype[3] );
    case  HIDDEN:
      return( ttype[4] );
    case  SPECIAL:
      return( ttype[5] );
    default:
      return( ttype[0] );
  }
}



static krui_err    krio_writeDefaultDefinitions( stream )
FILE    *stream;
{
  int   err;
  FlintType  act, bias;
  int   st, subnet_no, layer_no;
  char  *act_func,  *out_func;


  err = krio_fmtShapeing( DEFAULT_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[13] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  krui_getUnitDefaults( &act, &bias, &st, &subnet_no, &layer_no,
                        &act_func, &out_func );

  err = fprintf( stream, fmt_shape1, act, bias, getTType( st ),
                 subnet_no, layer_no, act_func, out_func );
  retchk( err );

  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  return( 0 );
}



static krui_err    krio_writeUnitDefinitions( stream )
FILE    *stream;
{
  static char   *blank = "\0";

  struct  PosType   pos;

  char  *u_type,
        *u_name,
        *act_func,
        *act_func_def,
        *out_func_def,
        *out_func;

  bool  no_Ftype,
        second;
  int   err,
        unit_no,
        u_no,
        st_def, subnet_no, layer_no;

  FlintType  act_def, bias_def;



  if (NoOfUnits <= 0)
    return( 0 );

  err = krio_fmtShapeing( UNIT_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[10] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  krui_getUnitDefaults( &act_def, &bias_def, &st_def, &subnet_no, &layer_no,
                        &act_func_def, &out_func_def );

  unit_no = krui_getFirstUnit();
  u_no = 1;

  do
    {
    krui_getUnitPosition( unit_no, &pos );

    if ( (u_name = krui_getUnitName( unit_no )) == NULL)
      u_name = blank;

    if ( (u_type = krui_getUnitFTypeName( unit_no )) == NULL)
      {  /*  no ftype  */
      no_Ftype = TRUE;

      act_func = krui_getUnitActFuncName( unit_no );
      out_func = krui_getUnitOutFuncName( unit_no );

      if ((act_func == act_func_def) && (out_func == out_func_def))
	{
#ifndef KERNEL3D
    err = fprintf( stream, fmt_shape2,
		   u_no, blank, u_name,
                   krui_getUnitActivation( unit_no ),
                   krui_getUnitBias( unit_no ),
                   getTType( krui_getUnitTType( unit_no ) ),
		   pos.x, pos.y );
#else
    err = fprintf( stream, fmt_shape2,
		   u_no, blank, u_name,
                   krui_getUnitActivation( unit_no ),
                   krui_getUnitBias( unit_no ),
                   getTType( krui_getUnitTType( unit_no ) ),
		   pos.x, pos.y, pos.z );
#endif
	}
      else
	{
	if (act_func == act_func_def)  act_func = blank;
	if (out_func == out_func_def)  out_func = blank;

#ifndef KERNEL3D
    err = fprintf( stream, fmt_shape1,
		   u_no, blank, u_name,
                   krui_getUnitActivation( unit_no ),
                   krui_getUnitBias( unit_no ),
                   getTType( krui_getUnitTType( unit_no ) ),
                   pos.x, pos.y, act_func, out_func );
#else
    err = fprintf( stream, fmt_shape1,
		   u_no, blank, u_name,
                   krui_getUnitActivation( unit_no ),
                   krui_getUnitBias( unit_no ),
                   getTType( krui_getUnitTType( unit_no ) ),
                   pos.x, pos.y, pos.z, act_func, out_func );
#endif
      }
    }
    else
      {
#ifndef KERNEL3D
    err = fprintf( stream, fmt_shape2,
                   u_no, u_type, u_name,
                   krui_getUnitActivation( unit_no ),
                   krui_getUnitBias( unit_no ),
                   getTType( krui_getUnitTType( unit_no ) ),
		   pos.x, pos.y );
#else
    err = fprintf( stream, fmt_shape2,
                   u_no, u_type, u_name,
                   krui_getUnitActivation( unit_no ),
                   krui_getUnitBias( unit_no ),
                   getTType( krui_getUnitTType( unit_no ) ),
		   pos.x, pos.y, pos.z );
#endif

      no_Ftype = FALSE;
    }


    retchk( err );

    if ( no_Ftype )
      {
      if ( krui_setFirstSite() )
        {
        second = FALSE;
        do
          {
          if (second)
            {
            err = fprintf( stream, fmt_blank );
	    retchk( err );
	  }

          err = fprintf( stream, " %-s", krui_getSiteName() );
          retchk( err );
          second = TRUE;
	}
        while ( krui_setNextSite() );
      }
    }

    err = fprintf( stream, "\n" );
    retchk( err );
    u_no++;
  }
  while ( (unit_no = krui_getNextUnit() ) > 0);

  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  return( 0 );
}


/*  write link weights 
*/
static krui_err     krio_writeSourcesAndWeights( stream )
FILE    *stream;
{
  bool  second;
  int   i, err,
        source_unit;

  FlintType     weight;


  i = 0;
  source_unit = krui_getFirstPredUnit( &weight );
  second = FALSE;

  do
    {
    if (second)
      {
      err = fprintf( stream, "," );
      retchk( err );
    }

    if (++i > max_connects_per_line)
      {
      i = 1;
      err = fprintf( stream, fmt_blank );
      retchk( err );
    }

    err = fprintf( stream, fmt_shape3, source_unit, weight );
    retchk( err );

    second = TRUE;
  }
  while ( (source_unit = krui_getNextPredUnit( &weight )) > 0);

  err = fprintf( stream, "\n" );
  retchk( err );

  return( 0 );
}


static krui_err    krio_writeConnectionDefs( stream )
FILE    *stream;
{
  bool  second;
  int   target_unit, unit_no,
        err;

  FlintType     weight;


  if (NoOfUnits <= 0 || NoOfLinks == 0)
    return( 0 );

  err = krio_fmtShapeing( CONNECT_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[11] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  unit_no = krui_getFirstUnit();
  target_unit = 1;

  do
    {
    switch (krui_getUnitInputType( unit_no ))
      {
      case  DIRECT_LINKS:
        err = fprintf( stream, fmt_shape1, target_unit, " ");
        retchk( err );
        err = krio_writeSourcesAndWeights( stream );
        retchk( err );

        break;

      case  SITES:
        krui_setFirstSite();
        second = FALSE;

        do
          {
          if (krui_getFirstPredUnit( &weight ) > 0)
            {
            if (second)
              {
              err = fprintf( stream, fmt_shape2, krui_getSiteName() );
              retchk( err );
	    }
            else
              {
              err = fprintf( stream, fmt_shape1, target_unit, krui_getSiteName() );
              retchk( err );
	    }

            err = krio_writeSourcesAndWeights( stream );
            retchk( err );

            second = TRUE;
	  }
	}
        while ( krui_setNextSite() );

        break;
    }

    target_unit++;
  }
  while ( (unit_no = krui_getNextUnit()) > 0 );

  err = fprintf( stream, fmt_hdr2 );
  retchk( err );
  return( 0 );
}


/*  clear unit's refresh flags (for unit sorting)
*/
static void     clr_R_flags()
{
  register struct Unit   *unit_ptr;
  register struct Unit   *last_unit_ptr;

  last_unit_ptr = unit_array + MaxUnitNo;
  for (unit_ptr = unit_array + MinUnitNo; unit_ptr <= last_unit_ptr; unit_ptr++)
      unit_ptr->flags &= ~UFLAG_REFRESH;

}


static krui_err    krio_writeSubnetDefs( stream )
FILE    *stream;
{
  struct Unit   *unit_ptr,
                *unit_ptr2;
  int	i,  k, elem_no, err, dummy2, def_subnet_no;
  short  subnet_no;
  FlintType  dummy1;
  char	     *dummy3;


  if (!is_subnet_info)  return( 0 );

  err = krio_fmtShapeing( SUBNET_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[12] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  /*  get default subnet number  */
  krui_getUnitDefaults( &dummy1, &dummy1, &dummy2, &def_subnet_no, &dummy2,
			&dummy3, &dummy3 );

  /*  clear refresh flags  */
  clr_R_flags();

  for (i = MinUnitNo, unit_ptr = unit_array + MinUnitNo; i <= MaxUnitNo; i++, unit_ptr++)
    if ( !(unit_ptr->flags & UFLAG_REFRESH) &&
          ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {     /*  unit is in use and 'fresh' */
      subnet_no = unit_ptr->subnet_no;
      if (subnet_no == def_subnet_no)
        continue;

      err = fprintf( stream, fmt_shape1, subnet_no, i);
      retchk( err );
      elem_no = 0;

      for (k = i + 1, unit_ptr2 = unit_ptr + 1;
           k <= MaxUnitNo;
           k++, unit_ptr2++ )
        {
        if ( !(unit_ptr2->flags & UFLAG_REFRESH) &&
             ((unit_ptr2->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
          {     /*  unit is in use and 'fresh' */
          if (unit_ptr2->subnet_no == subnet_no)
            {
            unit_ptr2->flags |= UFLAG_REFRESH;

            err = fprintf( stream, "," );
            retchk( err );

            if ( (++elem_no % max_subnets_per_line) == 0)
              {
              err = fprintf( stream, fmt_blank );
              retchk( err );
	    }

            err = fprintf( stream, fmt_shape2,  k);
            retchk( err );
	  }
	}
      }
    }

  err = fprintf( stream, fmt_hdr2 );
  retchk( err );
  err = fprintf( stream, "\n" );
  retchk( err );

  return( 0 );
}


static krui_err    krio_writeLayerDefs( stream )
FILE    *stream;
{
  struct Unit   *unit_ptr,
                *unit_ptr2;

  int	i,  k, elem_no, err, dummy2, def_layer_no;
  short  layer_no;
  FlintType  dummy1;
  char	    *dummy3;


  if (!is_layer_info)  return( 0 );

  err = krio_fmtShapeing( LAYER_DEF );
  retchk( err );

  err = fprintf( stream, "\n\n%s :\n\n", title[15] );
  retchk( err );
  err = fprintf( stream, fmt_hdr1 );
  retchk( err );
  err = fprintf( stream, fmt_hdr2 );
  retchk( err );

  /*  get default layer number	*/
  krui_getUnitDefaults( &dummy1, &dummy1, &dummy2, &dummy2, &def_layer_no,
			&dummy3, &dummy3 );

  /*  clear refresh flags  */
  clr_R_flags();

  for (i = MinUnitNo, unit_ptr = unit_array + MinUnitNo; i <= MaxUnitNo; i++, unit_ptr++)
    if ( !(unit_ptr->flags & UFLAG_REFRESH) &&
          ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
      {     /*  unit is in use and 'fresh' */
      layer_no = (int) unit_ptr->layer_no;
      if (layer_no == def_layer_no)
        continue;

      err = fprintf( stream, fmt_shape1, layer_no, i);
      retchk( err );
      elem_no = 0;

      for (k = i + 1, unit_ptr2 = unit_ptr + 1;
           k <= MaxUnitNo;
           k++, unit_ptr2++ )
        {
        if ( !(unit_ptr2->flags & UFLAG_REFRESH) &&
             ((unit_ptr2->flags & UFLAG_IN_USE) == UFLAG_IN_USE) )
          {     /*  unit is in use and 'fresh' */
          if ((int) unit_ptr2->layer_no == layer_no)
            {
            unit_ptr2->flags |= UFLAG_REFRESH;

            err = fprintf( stream, "," );
            retchk( err );

            if ( (++elem_no % max_layers_per_line) == 0)
              {
              err = fprintf( stream, fmt_blank );
              retchk( err );
              }

            err = fprintf( stream, fmt_shape2, k);
            retchk( err );
            }
          }
        }
      }

  err = fprintf( stream, fmt_hdr2 );
  retchk( err );
  err = fprintf( stream, "\n" );
  retchk( err );

  return( 0 );
}


#ifdef KERNEL3D

/*  Return the base address of the XY-Translation Table
*/
struct TransTable *krio_getXYTransTablePtr()
{
  return(  xyTransTable );
}


/*  writes the XY-Translation Table to output file
*/
static krui_err  writeXYTransTable( stream )

FILE    *stream;
{
  int err;
  int z_index;
  bool  do_it;  


  do_it = FALSE;
  for (z_index = 0; z_index < TRANS_TABLE_SIZE; z_index++)
    if ((xyTransTable[z_index].x != 0) || (xyTransTable[z_index].y != 0))
      {
      do_it = TRUE;
      break;
      }

  if (do_it)
    {
    err = fprintf( stream, "\n\n%s :\n\n", title[17]);
    retchk( err );
    err = fprintf (stream, " delta x | delta y |    z    \n");
    retchk( err );
    err = fprintf (stream, "---------|---------|---------\n");
    retchk( err );
    for (z_index = 0; z_index < TRANS_TABLE_SIZE; z_index++)
      {
      if ((xyTransTable[z_index].x != 0) || 
               (xyTransTable[z_index].y != 0))
        {
        err = fprintf (stream, "%8d |", xyTransTable[z_index].x);
        retchk( err );
        err = fprintf (stream, "%8d |", xyTransTable[z_index].y);
        retchk( err );
        err = fprintf (stream, "%8d \n", z_index - TRANS_TABLE_OFFSET);
        retchk( err );
        }
      }

    err = fprintf (stream, "---------|---------|---------\n");
    retchk( err );
    err = fprintf( stream, "\n" );
    retchk( err );
    }

  return( KRERR_NO_ERROR );
}


#endif



/*  write complete network to disk
*/
krui_err  krio_saveNet( filename, netname )
char    *filename,
        *netname;
{
  FILE      *file_out;
  int	    err;


  if (NoOfUnits <= 0)
    return( KRERR_NO_UNITS );  /*  no units defined  */

  if ( (file_out = fopen( filename, "w" )) == NULL)
    {
    err = KRERR_IO;  /*  file open error  */
    goto  end;
  }

  kr_forceUnitGC();  /*  do unit garbage-collection  */

  /*  write file header  */
  err = krio_writeHeader( file_out, krio_getIOVersion(),
                          netname, krui_getLearnFunc(), krui_getUpdateFunc() );
  /*  calc. maximum string length of all identifiers  */
  krio_stringLimits();
  if (err < 0)	goto end;
  err = krio_writeSiteDefinitions( file_out );
  if (err < 0)	goto end;
  err = krio_writeTypeDefinitions( file_out );
  if (err < 0)	goto end;
  err = krio_writeDefaultDefinitions( file_out );
  if (err < 0)	goto end;
  err = krio_writeUnitDefinitions( file_out );
  if (err < 0)	goto end;
  err = krio_writeConnectionDefs( file_out );
  if (err < 0)	goto end;
  err = krio_writeSubnetDefs( file_out );
  if (err < 0)	goto end;
  err = krio_writeLayerDefs( file_out );
  if (err < 0)	goto end;
#ifdef KERNEL3D
  err = writeXYTransTable( file_out );
  if (err < 0)	goto end;
#endif 
  err = fclose( file_out );

end:
  if (err == -2000) return( KRERR_SAVE_LINE_LEN );  /*  line length limit exceeded  */
  if (err != 0) return( KRERR_IO );  /*  Physical I/O error  */

  return( KRERR_NO_ERROR );
}



/*#################################################

GROUP: Kernel I/O Input Functions

#################################################*/

static  bool    skipComments()
{
  register int   c;
  register  FILE  *stream;


  stream = io_stream;

  while (TRUE)
    {
    do
      {
      c = getc( stream );
      if (c == '\n')  lineno++;
    }
    while ( isspace( c ) );

    if (c != '#')  break;

    do
      c = getc( stream );
    while (c != EOF && c != '\n');

    if (c == '\n')  lineno++;
  }

  if (c == EOF)  return( FALSE );

  ungetc( c, stream );
  return( TRUE );
}


static bool     skipSpace()
{
  register int  c;

  do
    {
    c = getc( io_stream );
    if (c == '\n')  lineno++;
  }
  while ( isspace( c ) );

  if (c == EOF)  {  io_err = KRERR_EOF;  return( FALSE ); }
  ungetc( c, io_stream );

  return( TRUE );
}


static bool     comma()
{
  register int  c;
  register  FILE  *stream;


  stream = io_stream;
  do
    {
    c = getc( stream );
    if (c == '\n')  lineno++;
  }
  while ( isspace( c ) );

  if (c == EOF)  {  io_err = KRERR_EOF;  return( FALSE ); }
  if (c != ',')
    {
    ungetc( c, stream );
    return( FALSE );
  }

  do
    {
    c = getc( stream );
    if (c == '\n')  lineno++;
  }
  while ( isspace( c ) );

  ungetc( c, stream );
  return( TRUE );
}



static bool     get_nl()
{
  register int  c;

  do
    {
    c = getc( io_stream );
    if (c == '\n')
      {
      lineno++;
      return( TRUE );
    }
  }
  while ( isspace( c ) );

  if (c == EOF)  {  io_err = KRERR_EOF;  return( FALSE ); }

  ungetc( c, io_stream );
  return( FALSE );
}


static bool     get_separ()
{
  register int  c;

  do
    {
    c = getc( io_stream );
    if (c == '\n')
      {
      lineno++;
      return( FALSE );
    }
  }
  while ( isspace( c ) );

  if (c == EOF)  {  io_err = KRERR_EOF;  return( FALSE ); }
  if (c == '|')  return( TRUE );

  ungetc( c, io_stream );
  return( FALSE );
}



static char     *getSection( line, title_no )
char    *line;
int     *title_no;
{
  int   i;
  char  *s1_ptr, *s2_ptr;


  if ( !skipComments() )
    {  /* EOF  */
    *title_no = -2;
    return( NULL );
  }

  if (fgets( line, LIN_MAX, io_stream ) == NULL)
    {  /*  I/O error  */
    *title_no = -3;
    return( NULL );
  }

  for (i = 0; i < NoOfTitles; i++)
    {
    s1_ptr = line; 
    s2_ptr = title[ i ];
    while (*s2_ptr != EOS)
      {  /* compare strings, don't examine white space  */
      if (*s1_ptr == ' ') s1_ptr++;
      if (*s2_ptr == ' ') s2_ptr++;
      if (*s1_ptr++ != *s2_ptr++)  break;

      if (*s2_ptr == EOS)
	{   /*	strings are equal   */
	*title_no = i;
	return( s1_ptr );
      }
    }
  }


  *title_no = -1;  /*  title not found  */
  return( NULL );
}


static  bool    matchHead2( N )
int     N;
{
  register int  c;
  register  FILE  *stream;
  int   i;


  if (!skipSpace())  return( FALSE );
  stream = io_stream;

  c = getc( stream );
  if (c != '-')
    {
    ungetc( c, stream );
    return( FALSE );
  }

  for (i = 0; i < N; i++)
    {
    c = getc( stream );
    if (c != '-')
      return( FALSE );

    do
      c = getc( stream );
    while ( c == '-' );

    if (c != '|')  {  io_err = KRERR_FILE_SYNTAX;  return( FALSE );  }
  }

  c = getc( stream );
  if (c != '-')  {  io_err = KRERR_FILE_SYNTAX;  return( FALSE );  }

  do
    c = getc( stream );
  while ( c == '-' );

  if (c == EOF)  {  io_err = KRERR_EOF;  return( FALSE ); }
  ungetc( c, stream );

  if ( !get_nl() )  return( FALSE );

  return( TRUE );
}



/*  Find the first occurrence of find in s.
    (slow version)
*/
static char  *my_strstr(s, find)
char *s, *find;
{
  register char c, sc;
  register size_t len;

  if ((c = *find++) != 0)
    {
    len = strlen(find);
    do 
      {
      do 
        {
	if ((sc = *s++) == 0)	return (NULL);
      }
      while (sc != c);
    }
    while (strncmp(s, find, len) != 0);
    s--;
  }

  return ((char *) s);
}



static void  krio_readHeader( stream, netfile_version, net_name, learn_func,
			      update_func, no_of_units, no_of_connect,
			      no_of_unitTypes, no_of_siteTypes )
FILE    *stream;
char    *net_name,
        *netfile_version,
        *learn_func,
        *update_func;
int     *no_of_units, *no_of_connect, *no_of_unitTypes, *no_of_siteTypes;
{
  char  *cursor;
  int   err,  i, no_of_scan_params,
        title_no;


  io_stream = stream;
  err = 1;


  /*  get file version  */
  if ( (cursor = getSection( fmt_shape1, &title_no )) == NULL)  return;
  if (title_no != 0)  { io_err = KRERR_FILE_FORMAT;  return;  }   /*  "SNNS network definition file" not found  */
  err = sscanf( cursor, "%s", netfile_version );
  if (err != 1)  { io_err = KRERR_FILE_FORMAT;  return;  }   /*  version string not found */
  if (my_strstr( netfile_version, NETFILE_VERSION ) != NULL)
    /*  current netfile version  */
    no_of_scan_params = 9;
  else
    no_of_scan_params = 8;

  for (i = 0; i < no_of_scan_params; i++)
    {
    if ( (cursor = getSection( fmt_shape1, &title_no )) == NULL)  return;

    switch (title_no)
      {
      case  -3:  /*  Physical I/O error  */
        io_err = KRERR_IO;
        return;
      case  -2:  /*  unexpected EOF  */ 
        io_err = KRERR_EOF;
        return;
      case  -1:  /*  not found  */ 
        io_err = KRERR_FILE_FORMAT;  /*  Incompatible file format  */
        return;
      case  1:  /*  "generated at"  */
        break;
      case  2:  /*  network name"  */
        err = sscanf( cursor, " :%s", net_name );
        if (err == 0) {
          *net_name = EOS;
          err = 1;
        }
        break;
      case  3:  /*  "no. of units"  */
        err = sscanf( cursor, " :%d", no_of_units );
        break; 
      case  4:  /*  "no. of connections"  */
        err = sscanf( cursor, " :%d", no_of_connect );
        break;
      case  5:  /*  "no. of unit types"  */
        err = sscanf( cursor, " :%d", no_of_unitTypes );
        break;
      case  6:  /*  "no. of site types"  */
        err = sscanf( cursor, " :%d", no_of_siteTypes );
        break;
      case  7:  /*  "learning function"  */
        err = sscanf( cursor, " :%s\n", learn_func );
        if (err != 1)  {
          *learn_func = '\0';
          err = 1;
        }
        break;
      case 14:  /*  "source files"  */
        break;
      case  16:  /*  "update function"  */
        err = sscanf( cursor, " :%s\n", update_func );
        if (err != 1)  {
          *update_func = '\0';
          err = 1;
        }
        break;
      default:
        io_err = KRERR_FILE_FORMAT;  /*  Incompatible file format  */
        return;
     }

   switch (err)
     {
     case 1:    /*  one argument matched  */
       break;
     case EOF:  /*  unexpected EOF  */
       io_err = KRERR_EOF;
       return;
     default:   /*  Syntax error at line  */
       io_err = KRERR_FILE_SYNTAX;
       return;
     }

   lineno++;
   }

}


static int      str_to_Ttype( str )
char    *str;
{
  switch (*str)
    {
    case 'i':
        return( INPUT );
    case 'o':
        return( OUTPUT );
    case 'h':
        return( HIDDEN );
    case 'd':
        return( DUAL );
    case 's':
        return( SPECIAL );
    default:
        return( 0 );
    }
}



static void     krio_readSiteDefinitions( stream )
FILE    *stream;
{
  int   ret;


  io_stream = stream;
  NoOfSiteTypes = 0;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[0] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 1 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  while (TRUE)
    {
    if (matchHead2( 1 )) return;
    if ( !skipComments() )  return;
    ret = fscanf( stream, "%s | %s\n", fmt_shape1, fmt_shape2 );

    if (ret != 2)
      {  io_err = KRERR_FILE_SYNTAX;  return;  }

    lineno++;
    io_err = krui_createSiteTableEntry( fmt_shape1, fmt_shape2 );
    if (io_err)  return;
    NoOfSiteTypes++;
    }
}


static void     krio_readTypeDefinitions( stream )
FILE    *stream;
{
  int   no_sites,  ret,  i;


  NoOfUnitTypes = 0;
  io_stream = stream;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[1] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 3 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  if (NoOfSiteTypes > 0)
    {
    site_names = (SymbolType *) calloc( NoOfSiteTypes, sizeof (SymbolType) );
    site_name_ptrs = (char * *) calloc(  NoOfSiteTypes, sizeof (site_name_ptrs) );
    if (site_names == NULL || site_name_ptrs == NULL)
      {
      io_err = KRERR_INSUFFICIENT_MEM;
      return;
      }
    }

  while (TRUE)
    {
    if (io_err)  return;

    if ( matchHead2( 3 ) )  return;
    if ( !skipComments() )  return;
    NoOfUnitTypes++;

    ret = fscanf( stream, "%s | %s | %s |",
                  fmt_shape1, fmt_shape2,  fmt_shape3 );
    if (ret != 3)
      {  io_err = KRERR_FILE_SYNTAX;  return;  }

    if ( get_nl() )
      {
      io_err = krui_createFTypeEntry( fmt_shape1, fmt_shape2,  fmt_shape3,
                                      0, site_name_ptrs );
      continue;
      }

    no_sites = 0;
    while (no_sites < NoOfSiteTypes)
      {
      ret = fscanf( stream, " %[^,\n ]", site_names[ no_sites ] );
      if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      no_sites++;

      if ( !comma() )  break;
      }

    if (io_err)  return;
    for (i = 0; i < no_sites; i++)
      site_name_ptrs[ i ] = site_names[ i ];

    io_err = krui_createFTypeEntry( fmt_shape1, fmt_shape2,  fmt_shape3,
                                    no_sites, site_name_ptrs );
    }

}



static void     krio_readDefaultDefinitions( stream )
FILE    *stream;
{
  FlintType     act, bias;
  int           ret, st, subnet_no, layer_no;


  io_stream = stream;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[2] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 6 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  ret = fscanf( stream, "%f | %f | %s | %d | %d | %s | %s",
                &act, &bias, fmt_shape1, &subnet_no, &layer_no, fmt_shape2, fmt_shape3 );
  if (ret != 7)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  if ( (st = str_to_Ttype( fmt_shape1 )) == 0)
     {  io_err = KRERR_TTYPE; return;  }

  io_err = krui_setUnitDefaults( act, bias, st, subnet_no, layer_no,
                                 fmt_shape2, fmt_shape3 );
  if (io_err)  return;

  if ( !matchHead2( 6 ) )
    io_err = KRERR_FILE_SYNTAX;

  return;
}


static void     krio_readUnitDefinitions( stream, netfile_version_str)
FILE    *stream;
char    *netfile_version_str;
{
  struct  PosType   pos;
  FlintType     fp;
  int   no,  no_unit,  st,  ret,  x,  y, z;
  bool  is_Ftype;


  io_stream = stream;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[3] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 9 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  no_unit = 0;
  while (TRUE)
    {
    if (io_err)  return;

    if ( matchHead2( 9 ) )  return;
    if ( !skipComments() )  return;
    no_unit++;

    ret = fscanf( stream, "%d", &no );
    if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if (no != no_unit)  {  io_err = KRERR_OUTFUNC;  return;  }

    ret = fscanf( stream, " | %[^| ]", fmt_shape1 );
    if (ret != 0 && ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    if (ret == 1)
      {
      ret = krui_createFTypeUnit( fmt_shape1 );
      is_Ftype = TRUE;
      }
    else
      {
      ret = krui_createDefaultUnit();
      is_Ftype = FALSE;
      }
    if (ret < 0)    {  io_err = ret;  return;  }
    if (ret != no)  {  io_err = KRERR_MALLOC1;  return;  }

    ret = fscanf( stream, " | %[^| ]", fmt_shape1 );
    if (ret != 0 && ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if (ret == 1)
      {
      io_err = krui_setUnitName( no_unit, fmt_shape1 );
      if (io_err)  return;
      }

    ret = fscanf( stream, " | %f", &fp );
    if (ret == 1)
      {
      krui_setUnitInitialActivation( no_unit, fp );
      krui_setUnitActivation( no_unit, fp );
      }

    ret = fscanf( stream, " | %f", &fp );
    if (ret == 1)
      krui_setUnitBias( no_unit, fp );

    ret = fscanf( stream, " | %[^| ]", fmt_shape1 );
    if (ret != 0 && ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if (ret == 1)
      {
      if ( (st = str_to_Ttype( fmt_shape1 )) == 0)
         {  io_err = KRERR_TTYPE; return;  }

      io_err = krui_setUnitTType( no_unit, st );
      if (io_err)  return;
      }

    if (my_strstr( netfile_version_str, KERNEL3D_NETFILE_VERSION ) != NULL)
      {
      ret = fscanf( stream, " | %d , %d, %d", &x, &y, &z );
      if (ret != 0 && ret != 3)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      if (ret == 3)
        {
        pos.x = (short) x;
        pos.y = (short) y;
#ifdef KERNEL3D
        pos.z = (short) z;
#endif
        krui_setUnitPosition( no_unit, &pos );
        }
      }
    else
      {
      ret = fscanf( stream, " | %d , %d", &x, &y );
      if (ret != 0 && ret != 2)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      if (ret == 2)
        {
        pos.x = (short) x;
        pos.y = (short) y;
#ifdef KERNEL3D
        pos.z = (short) 0;
#endif
        krui_setUnitPosition( no_unit, &pos );
        }
      }

    if (is_Ftype)
      {
      ret = fscanf( stream, " | | |" );
      if ( ret != 0 || !get_nl() )  {  io_err = KRERR_FILE_SYNTAX;  return;  }

      continue;
      }

    ret = fscanf( stream, " | %[^| ]", fmt_shape1 );
    if (ret == 1)
      {
      io_err = krui_setUnitActFunc( no_unit, fmt_shape1 );
      if (io_err)  return;
      }

    ret = fscanf( stream, " | %[^| ]", fmt_shape1 );
    if (ret == 1)
      {
      io_err = krui_setUnitOutFunc( no_unit, fmt_shape1 );
      if (io_err)  return;
      }

    if ( !get_separ() )  {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if ( get_nl() )  continue;

    io_err = krui_setCurrentUnit( no_unit );
    if (io_err)  return;

    while (TRUE)
      {
      ret = fscanf( stream, " %[^,\n ]", fmt_shape1 );
      if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

      io_err = krui_addSite( fmt_shape1 );
      if (io_err)  return;

      if ( !comma() )  break;
      }
    }
}

static void     krio_readConnectionDefs( stream )
FILE    *stream;
{
  FlintType  weight;
  int   no, ret, src_unit;
  bool  new_unit, has_sites;


  io_stream = stream;
  has_sites = FALSE;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[4] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 2 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  new_unit = TRUE;
  while (TRUE)
    {
    if (io_err)  return;

    if ( matchHead2( 2 ) )  return;
    if ( !skipComments() )  return;

    ret = fscanf( stream, "%d", &no );
    if (ret != 0 && ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if (new_unit && ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    if (ret == 1)
      {
      io_err = krui_setCurrentUnit( no );
      if (io_err)  return;
      if ( !get_separ() )  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      new_unit = FALSE;
      has_sites = FALSE;
      }

    ret = fscanf( stream, " %[^| ] |", fmt_shape1 );
    if (ret != 0 && ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if (has_sites && ret == 0)   {  io_err = KRERR_FILE_SYNTAX;  return;  }
    if (ret == 1)
      {
      io_err = krui_setSite( fmt_shape1 );
      if (io_err)  return;
      has_sites = TRUE;
      }
    else
      {
      if ( !get_separ() )  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      new_unit = TRUE;
      }

    while (TRUE)
      {
      ret = fscanf( stream, "%d : %f", &src_unit, &weight );
      if (ret != 2)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

      io_err = krui_createLink( src_unit, weight );
      if (io_err)  return;

      if ( !comma() )  break;
      }
    }
}


static bool     checkUnitNo( unit_no )
int     unit_no;
{
  if ( (unit_no < MinUnitNo) || (unit_no > MaxUnitNo) )
    return( FALSE );    /*  invalid unit #  */

  if ((unit_array[ unit_no ].flags & UFLAG_IN_USE) != UFLAG_IN_USE)
    return( FALSE );    /*  invalid unit #  */

  return( TRUE );
}


static void     krio_readSubnetDefs( stream )
FILE    *stream;
{
  int   unit_no, subnet_no, ret;


  io_stream = stream;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[5] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 1 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }


  while (TRUE)
    {
    if (matchHead2( 1 )) return;
    if ( !skipComments() )  return;
    ret = fscanf( stream, "%d", &subnet_no );
    if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    if ( !get_separ() )  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    while (TRUE)
      {
      ret = fscanf( stream, "%d", &unit_no );
      if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      if (!checkUnitNo( unit_no ))
        {  io_err = KRERR_UNIT_NO;  return;  }

      krui_setUnitSubnetNo( unit_no, subnet_no );

      if ( !comma() )  break;
      }
    }
}


static void     krio_readLayerDefs( stream )
FILE    *stream;
{
  int   unit_no, layer_no, ret;


  io_stream = stream;

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[6] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 1 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  while (TRUE)
    {
    if (matchHead2( 1 )) return;
    if ( !skipComments() )  return;
    ret = fscanf( stream, "%d", &layer_no );
    if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    if ( !get_separ() )  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    while (TRUE)
      {
      ret = fscanf( stream, "%d", &unit_no );
      if (ret != 1)  {  io_err = KRERR_FILE_SYNTAX;  return;  }
      if (!checkUnitNo( unit_no ))
        {  io_err = KRERR_UNIT_NO;  return;  }

      krui_setUnitLayerNo( unit_no, (unsigned short) layer_no );

      if ( !comma() )  break;
      }
    }
}



#ifdef KERNEL3D


/*  read the XY-Tranlation Table from input file
*/
void  readXYTransTable( stream )
FILE    *stream;
{
  int x, y, z, ret;


  memset( (char *) xyTransTable, 0, sizeof( xyTransTable ) );

  if ( !skipComments() )  return;
  if (fscanf( stream, headers[7] ) != 0)
    {  io_err = KRERR_FILE_SYNTAX;  return;  }
  if ( !matchHead2( 2 ) )
    {  io_err = KRERR_FILE_SYNTAX;  return;  }

  while (TRUE)
    {
    if (matchHead2( 2 )) return;
    if ( !skipComments() )  return;

    ret = fscanf( stream, "%d | %d | %d", &x, &y, &z );
    if (ret != 3)  {  io_err = KRERR_FILE_SYNTAX;  return;  }

    xyTransTable[TRANS_TABLE_OFFSET + z].x = (short) x;        
    xyTransTable[TRANS_TABLE_OFFSET + z].y = (short) y;        
    }

}


#endif



/*  load network form disk
*/
krui_err    krio_loadNet( filename, netname, netfile_version )
char    *filename,
        **netname,
        **netfile_version;
{
  static char  netname_str[81],
               netfile_version_str[81],
               learn_func[81],
               update_func[81];

  FILE      *file_in;
  int       no_units, no_connect, no_unitT, no_siteT;
  int       title_no, status;


  io_err = lineno = status = 0;

  file_in = fopen( filename, "r" );
  if (file_in == NULL)
    {
    io_err = KRERR_FILE_OPEN;  /*  Can't open file  */
    goto end;
    }

  lineno = 1;
  if (unit_array != NULL)  krui_deleteNet();

  krio_readHeader( file_in, netfile_version_str, netname_str, learn_func, update_func, 
                   &no_units, &no_connect, &no_unitT, &no_siteT );
  if (io_err)  goto end;

  if( *learn_func != '\0') 
    {
    io_err = krui_setLearnFunc( learn_func );
    if (io_err)  goto end;
    }

  if( *update_func != '\0') 
    {
    io_err = krui_setUpdateFunc( update_func );
    if (io_err)  goto end;
    }


  *netname = netname_str;
  *netfile_version = netfile_version_str;

  while ( getSection( fmt_shape1, &title_no ) != NULL )
    {
    if (title_no >= 0)  lineno++;  /* increment line number */

    switch (title_no)
      {
      case  -3:  /*  Physical I/O error  */
        io_err = KRERR_IO;
        return( io_err );
      case  -2:  /*   EOF  */
        if (status < 2)  io_err = KRERR_EOF;  /*  unexpected EOF  */
        return( io_err );
      case  -1:  /*  not found  */
        io_err = KRERR_FILE_SYNTAX;  /*  Syntax error at line  */
        break;
      case  8:  /*  "site definition section"  */
        krio_readSiteDefinitions( file_in );
        break;
      case  9:  /*  "type definition section"  */
        krio_readTypeDefinitions( file_in );
        if (site_names != NULL)
          {
          free( (char *) site_names );
          free( (char *) site_name_ptrs );
          }
        break;
      case  10:  /*  "unit definition section"  */
        if (no_units > UNIT_BLOCK)
          {
          io_err = krui_allocateUnits( no_units );
          if (io_err)  goto end;
          }

        krio_readUnitDefinitions( file_in, netfile_version_str );
        status++;
        break;
      case  11:  /*  "connection definition section"  */
        krio_readConnectionDefs( file_in );
        status++;
        break;
      case  12:  /*  "subnet definition section"  */
        krio_readSubnetDefs( file_in );
        break;
      case  13:  /*  "unit default section"  */
        krio_readDefaultDefinitions( file_in );
        break;
      case  15:  /*  "layer definition section"  */
        krio_readLayerDefs( file_in );
        break;
#ifdef KERNEL3D
      case  17:  /*  "3D translation section"  */
        readXYTransTable( file_in );
        break;
#endif
      default:
        io_err = KRERR_FILE_FORMAT;  /*  Incompatible file format  */
        break;
      }

    if (io_err)  break;
    }

  if (io_err == 0)  lineno = 0;

end:
  if (file_in != NULL)
    fclose( file_in );

  return( io_err );
}


/*#################################################

GROUP: Functions for Pattern Management

#################################################*/

/*  loading pattern file from disk
*/
krui_err    krio_loadPatterns( filename )
char    *filename;
{
  FILE      *file_in;
  int       io_err,  ret,  i,  j;
  int       no_of_patterns,  no_of_input_units,  no_of_output_units;
  Patterns  in_pat,  out_pat;


  lineno = 1;
  io_err = 0;
  PatternsValid = FALSE;

  file_in = fopen( filename, "r" );
  if (file_in == NULL)
    return( KRERR_FILE_OPEN );  /*  Can't open file  */

  io_stream = file_in;

  if (!skipComments())
    {  io_err =  KRERR_EOF;  goto end;  }  /*  Unexpected EOF  */

/*  read header and version number  */
  if ( (ret = fscanf( file_in, patHeader[0], fmt_shape1 )) == 1)
    {  /* new pattern file format  */
    lineno++;

/*  read date  */
    if ( (ret = fscanf( file_in, patHeader[1] )) != 0)
      {  io_err = KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format        */
    if ( fgets( fmt_shape2, LIN_MAX, file_in  ) == NULL)
      {  io_err = KRERR_EOF; goto end;  }  /*  Incompatible file format  */
    lineno++;

    if (!skipComments())
      {  io_err =  KRERR_EOF;  goto end;  }  /*  Unexpected EOF  */

    /*  read no. of patterns  */
    if ( (ret = fscanf( file_in, patHeader[2], &no_of_patterns )) != 1)
      {  io_err =  KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format        */
    lineno++;
    /*  read no. of input units  */
    if ( (ret = fscanf( file_in, patHeader[3], &no_of_input_units )) != 1)
      {  io_err =  KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format        */
    lineno++;
    /*  read no. of output units  */
    if ( (ret = fscanf( file_in, patHeader[4], &no_of_output_units )) != 1)
      {  io_err =  KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format        */
    lineno++;
    }
  else
    {  /*  old pattern format */
    ret = fscanf( file_in, "%d%d%d", &no_of_patterns, &no_of_input_units,
                                     &no_of_output_units );
    if (ret != 3)
      {  io_err =  KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format        */
    lineno++;
    }

  if (no_of_input_units != krui_getNoOfInputUnits() )
    {  /*  Pattern file contains not the same no. of input units as the network  */
    io_err = KRERR_INPUT_PATTERNS;
    goto end;
    }
  if (no_of_output_units != krui_getNoOfOutputUnits() )
    {  /*  Pattern file contains not the same no. of output units as the network  */
    io_err = KRERR_OUTPUT_PATTERNS;
    goto end;
    }

  NoOfInputUnits = no_of_input_units;
  NoOfOutputUnits = no_of_output_units;

  krui_shufflePatterns( FALSE );  /*  free array for permutation of patterns  */

  krm_releasePatternArrays();

  if (krm_allocPatternArrays( (no_of_patterns / PATTERN_BLOCK + 1) * PATTERN_BLOCK ) != 0)
    {  io_err = KRERR_INSUFFICIENT_MEM;  goto end;  }  /*  memory fault  */

  in_pat = in_patterns;
  out_pat = out_patterns;
  for (i = 0; i < no_of_patterns; i++)
    {
    if (!skipComments())
      {  io_err = KRERR_EOF;  goto end;  }  /*  Unexpected EOF  */

    for (j = 0; j < no_of_input_units; j++)
      {
      ret = fscanf( file_in, "%f", in_pat );
      if (ret != 1)
        {  io_err = KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format  */
      lineno++;
      in_pat++;
      }

    if (!skipComments())
      {  io_err = KRERR_EOF;  goto end;  }  /*  Unexpected EOF  */

    for (j = 0; j < no_of_output_units; j++)
      {
      if (!skipComments())
        {  io_err = KRERR_EOF;  goto end;  }  /*  Unexpected EOF  */

      ret = fscanf( file_in, "%f", out_pat );
      if (ret != 1)
        {  io_err = KRERR_FILE_FORMAT;  goto end;  }  /*  Incompatible file format  */
      out_pat++;
      }

    }

  NoOfPatternPairs = no_of_patterns;

end:
  if (io_err == KRERR_NO_ERROR)
    {  PatternsValid = TRUE;
       lineno = 0;
       if (fclose( file_in ) != 0)
	 io_err = KRERR_IO;  /*  Physical I/O error  */
    }

  if (io_err != KRERR_NO_ERROR)
    {  krm_releasePatternArrays();
       PatternsValid = FALSE;
       (void) fclose( file_in );
    }

  return( io_err );
}



/*  write patterns to disk
*/
krui_err    krio_savePatterns( filename )
char    *filename;
{
  FILE      *file_out;
  char      work_str[ LIN_MAX ];
  int       io_err,  ret,  i,  j;
  long      clock;
  Patterns  in_pat,  out_pat;


  lineno = 1;
  io_err = 0;

  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );  /*  No patterns defined  */
  if (NoOfInputUnits == 0)
    return( KRERR_NO_INPUT_UNITS );  /*  No input units defined  */
  if (NoOfOutputUnits == 0)
    return( KRERR_NO_OUTPUT_UNITS );  /*  No output units defined  */

  file_out = fopen( filename, "w" );
  if (file_out == NULL)
    return( KRERR_FILE_OPEN );  /*  Can't open file  */

/*  write header and version number  */
  ret = fprintf( file_out, patHeader[0], krio_getIOVersion() );
  RETCHKGTO( ret );
  lineno++;
/*  write date  */
  clock = 1;
  (void) time( &clock);

  strcpy( work_str, patHeader[1] );
  strcat( work_str, " %s\n" );
  ret = fprintf( file_out, work_str, ctime( &clock) );
  RETCHKGTO( ret );
  lineno++;

/*  write no. of patterns, input units, output units  */
  strcpy( work_str, patHeader[2] );
  strcat( work_str, patHeader[3] );
  strcat( work_str, patHeader[4] );
  ret = fprintf( file_out, work_str, NoOfPatternPairs, NoOfInputUnits, NoOfOutputUnits );
  RETCHKGTO( ret );
  lineno += 3;

  in_pat = in_patterns;
  out_pat = out_patterns;
  for (i = 1; i <= NoOfPatternPairs; i++)
    {
    ret = fprintf( file_out, "\n#%d\n", i );
    RETCHKGTO( ret );

    for (j = 1; j <= NoOfInputUnits; j++)
      {
      sprintf( work_str, "%.5f", *in_pat );
      krio_cutTrailingZeros( work_str );

      if (j < NoOfInputUnits)
        if (j % 10 == 0)  strcat( work_str, "\n" );
        else  strcat( work_str, " " );
   
      if (fputs( work_str, file_out ) == EOF)  
        {
        io_err = KRERR_IO;    /*  Physical I/O error  */
	goto  write_err;
        }

      lineno++;
      in_pat++;
      }

    fputs( "\n", file_out );

    for (j = 1; j <= NoOfOutputUnits; j++)
      {
      sprintf( work_str, "%.5f", *out_pat );
      krio_cutTrailingZeros( work_str );

      if (j < NoOfOutputUnits)
        if (j % 10 == 0)  strcat( work_str, "\n" );
        else  strcat( work_str, " " );

      if (fputs( work_str, file_out ) == EOF)  
        {
        io_err = KRERR_IO;    /*  Physical I/O error  */
	goto  write_err;
        }

      out_pat++;
      }
    }

end:
#ifdef ultrix

  if (ret == EOF)
    io_err = KRERR_IO;	  /*  Physical I/O error  */
  else
    io_err = KRERR_NO_ERROR;

#else

  if (ret < 0)
    io_err = KRERR_IO;	  /*  Physical I/O error  */
  else
    io_err = KRERR_NO_ERROR;

#endif

  if (io_err == KRERR_NO_ERROR)
    {  lineno = 0;
       if (fclose( file_out ) != 0)
	 io_err = KRERR_IO;  /*  Physical I/O error  */
    }

write_err:
  if (io_err != KRERR_NO_ERROR)  (void) fclose( file_out );
 
  return( io_err );
}
