/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel: Datatypes and Constants for Internal Useage
NOTES  : 
AUTHOR : Niels Mache
DATE   : 12.02.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include "glob_typ.h"	  /*	Global Datatypes and Constants	*/

#ifndef KR_TYPES
#define KR_TYPES


/*#################################################

GROUP: Type Definitions

#################################################*/


/*  type definition of input/output pattern array
*/
typedef  FlintType     *Patterns;

/*  type definition of integer array for shuffeling input/output patterns
*/
typedef  int    *PatternNumbers;



/*
typedef  FlintType   (* SiteFuncPtr) ( struct Unit * );
typedef  FlintType   (* ActFuncPtr)  ( struct Unit * );
typedef  FlintType   (* OutFuncPtr)  ( FlintType );
*/

typedef  FlintType   (* SiteFuncPtr) ();
typedef  FlintType   (* ActFuncPtr)  ();
typedef  FlintType   (* ActDerivFuncPtr)  ();
typedef  FlintType   (* OutFuncPtr)  ();

typedef  krui_err    (* LearnFuncPtr) ();
typedef  krui_err    (* UpdateFuncPtr) ();
typedef  krui_err    (* InitFuncPtr) ();

/*  typedef  FlintType	 (* FunctionPtr) ();  */

typedef  void	(* FunctionPtr) ();

/*#################################################

GROUP: Type Definition of Name Table

#################################################*/

struct   NameTable
  {
  union
    {
    char               *symbol;             /*  stores symbol name  */
    struct  NameTable  *next;               /*  stores name-table block ptr
                                                and free name-table entry   */
    }  Entry;

  unsigned short       sym_type,            /*  stores type of entry   */
                       ref_count;           /*  stores no. of references to
                                                this symbol (If more then MAXSHORT
                                                references are detected the symbol will
                                                released only by krm_releaseMem()
                                            */
  };



/*#################################################

GROUP: Type Definition of Site-Table

#################################################*/

struct   SiteTable
  {
  union
    {
    struct NameTable	*site_name;  /*  stores the site symbol (it's not a direkt
					 pointer to the symbol, but a pointer to the
					 name-table entry */
    struct SiteTable	*next;	     /*  stores next site-table block ptr
					 and free site-table entries */
    }  Entry;

  SiteFuncPtr   site_func;      /*  stores site function    */
  };


/*#################################################

GROUP: type definition of unit functionality table

#################################################*/

struct   FtypeUnitStruct
  {
  struct  NameTable     *Ftype_symbol;

  OutFuncPtr      out_func;
  ActFuncPtr      act_func;
  ActDerivFuncPtr  act_deriv_func;

  struct Site              *sites;

  struct FtypeUnitStruct   *next;
  struct FtypeUnitStruct   *prev;
  };



/*#################################################

GROUP: Unit/Site/Link type definitions

#################################################*/


struct   Link
  {
  struct Unit   *to;
  FlintType     weight;
					   /*  for learning the weights:  */
  FlintType	value_a,		   /*  previous weight change  */
		value_b,		   /*  previous weight slope  */
		value_c;       		   /*  actual weight slope  */

  struct Link   *next;
  };



struct   Site
  {
  struct Link   *links;
  struct SiteTable      *site_table;
  struct Site   *next;
  };


typedef  unsigned short  FlagWord;

struct   Unit
  {
  /*  output MUST be the first element in unit structure !
  (access is faster if functions can assume that output is the first element  */
  union
    {
    FlintType     output;
    int           nextFreeUnit;
    } Out;


  FlagWord	  flags;

  struct  FtypeUnitStruct   *Ftype_entry;

  union
    {
    FlintType     flint_no;     /*  used by backpropagation: holds errors from successors   */
    int           int_no;       /*  used by layer sorting: holds the layer no of the unit   */
    char         *ptr;          /*  used for garbage collection  */
    }  Aux;   /*  general purpose use  */


  FlintType	  act,		/*  current activation	*/
		  i_act,	/*  initial activation	*/
		  bias; 	/*  bias  */

  FlintType       value_a,
                  value_b,
                  value_c;

  OutFuncPtr	  out_func;	    /*	output function  */
  ActFuncPtr	  act_func;	    /*	activation function */
  ActDerivFuncPtr  act_deriv_func;  /*	derivation act. function */

  char *	  unit_name;	    /*	unit name */

  short 	  subnet_no;	    /*	subnet no.  */
  unsigned short  layer_no;	    /*	layer no.  */

  struct PosType  unit_pos;	    /*	unit's position  */

#ifdef KERNEL3D
  /*  unit's center position  */
  struct PositionVector unit_center_pos[ NO_OF_UNIT_CENTER_POS ];
#endif

  struct Site     *sites;
  };



/*#################################################

GROUP: Type Definition of Function Table

#################################################*/

struct  FuncTable
  {
  char    *func_name;                 /*  name of the function  */
  short    func_type,                 /*  type of the function  */
           no_of_input_parameters,    /*  no. of input parameters needed for this function  */
                                      /*  (only for Update, Init and Learning functions)  */
           no_of_output_parameters;   /*  no. of ouput parameters returned by this function  */
                                      /*  (only for Update and Learning functions)  */
  FunctionPtr  function;              /*  pointer to the function  */
  };


/*#################################################

GROUP: Link/Site/Unit and Unit Ptrs Array Type Definitions

#################################################*/

/*  definition of link array
*/
typedef  struct Link    *LinkArray;

/*  definition of site array
*/
typedef  struct Site    *SiteArray;

/*  definition of pointer array for topological sorting
*/
typedef  struct Unit *  *TopoPtrArray;

/*  definition of unit array
*/
typedef  struct Unit    *UnitArray;

/*  Definition of Name Table Array
*/
typedef  struct NameTable  *NTableArray;

/*  Definition of Site Name Table Array
*/
typedef  struct SiteTable  *STableArray;


/*#################################################

GROUP: Structure Size Constants

#################################################*/

#define   UNIT_SIZE             sizeof (struct Unit)
#define   SITE_SIZE             sizeof (struct Site)
#define   LINK_SIZE             sizeof (struct Link)

#define   PATTERN_SIZE          sizeof (FlintType)
#define   PATTERN_NO_SIZE       sizeof (int)

#define   TOPO_PTR_SIZE         sizeof (struct Unit *)

#define   NTABLE_SIZE           sizeof (struct NameTable)

#define   STABLE_SIZE           sizeof (struct SiteTable)

#define   FTYPE_UNIT_SIZE       sizeof (struct FtypeUnitStruct)


#endif
