/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel User Interface
NOTES  : 
AUTHOR : Niels Mache
DATE   : 27.02.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <memory.h>


#include "kr_typ.h"	 /*  Kernel Types and Constants  */
#include "kr_const.h"	 /*  Constant Declarators for SNNS-Kernel  */
#include "kr_io.h"	 /*  Kernel I/O-Routines */
#include "random.h"	 /*  Randomize Library Function Prototypes  */

#include "kr_ui.h"	 /*  Interface function prototypes  */
#include "kr_mac.h"	 /*  Kernel Macros  */

#include "version.h"	 /*  Version and Patchlevel  */


/*#################################################

GROUP: Global Var's (as declared by the kernel)

#################################################*/


extern	bool	NetModified,	  /*  TRUE, if the network topology was modified  */
		PatternsValid;	  /*  TRUE, if In-Core patterns are valid  */

extern	int  NoOfUnits, 	  /*  no. of units in the network  */
	     MinUnitNo, 	  /*  the first (lowest) used unit number in the network  */
	     MaxUnitNo, 	  /*  the last (highest) used unit number in the network  */
	     NoOfPatternPairs,	  /*  no. of defined pattern pairs */
	     NoOfInputUnits,	  /*  no. of input units  */
	     NoOfOutputUnits;	  /*  no. of output units  */



extern	UnitArray	unit_array;	/*  the unit array  */
extern	Patterns	in_patterns,	/*  the input pattern array  */
			out_patterns;	/*  the output pattern array  */

/*  global variables for the krui_error( ... ) function needed to generate
    error messages
*/
extern	int  lineno,		  /*  file I/O: line no. in network file, if network loading failed  */
	     io_err,		  /*  file I/O: operating system I/O error value, if network loading failed  */
	     topo_NoOfCycles,	  /*  topological sorting: no. of detected cycles in the network  */
	     topo_NoOfDeadUnits,  /*  topological sorting: no. of dead units in the network  */
	     topo_err_unit;	  /*  topological sorting: first cycle or first dead unit  */




/*#################################################

GROUP: Local Var's

#################################################*/





static int  UICurrentUnitNo     = 0,
            UICurrentLinkUnitNo = 0,
            UICurrentNameSearchUnitNo = 0;

static struct Unit     *UICurrentUnitPtr       = NULL,
                       *UICurrentSuccUnitPtrFS = NULL,
                       *UISourceUnitPtrFS      = NULL;

static struct Site     *UICurrentSitePtr       = NULL,
                       *UICurrentSitePtrFS     = NULL,
                       *UICurrentPredSitePtr   = NULL,
                       *UICurrentPredSitePtrFS = NULL,
                       *UICurrentFtypeSite     = NULL;

static struct Link     *UICurrentLinkPtr       = NULL,
                       *UICurrentPredLinkPtr   = NULL;

static struct FtypeUnitStruct  *UICurrentFtypeEntry = NULL;

static char  *UICurrentNameSearchUnitSymbolPtr = NULL;


/*#################################################

GROUP: External Functions

#################################################*/

/*  Functions defined in kr_funcs.c */
#include "kr_funcs.h"

/*  Functions defined in kr_mem.c   */
#include "kr_mem.h"

/*  Functions defined in kernel.c   */
#include "kernel.h"


/*#################################################

GROUP: Unit Functions

#################################################*/


/*  Returns the number of units used by the network.
*/
int     krui_getNoOfUnits()
{
  return( NoOfUnits );
}


/*  Initializes the first available unit for access. If the unit has sites, the
    first site will be set to the current site.
    Returns the unit no. of the first unit or 0 if no units available.
*/
int     krui_getFirstUnit()
{
  if (NoOfUnits == 0)  return( 0 );
  UICurrentUnitNo  = MinUnitNo;
  UICurrentUnitPtr = unit_array + MinUnitNo;

  if ( (UICurrentUnitPtr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
    {   /*  Initialize current site pointer to the first available site */
    UICurrentPredSitePtr = NULL;
    UICurrentSitePtr = UICurrentUnitPtr->sites;
    }
  else
    {   /*  No sites available  */
    UICurrentPredSitePtr = NULL;
    UICurrentSitePtr     = NULL;
    }

  return( MinUnitNo );
}

/*  Initializes the next available unit for access. If the unit has sites, the
    first site will be set to the current site.
    Returns the unit no. of the next unit or 0 if no more units available.
*/
int     krui_getNextUnit()
{
  register int   i;
  register struct Unit   *unit_ptr;


  if (NoOfUnits == 0)  return( 0 );

  for (i = UICurrentUnitNo + 1, unit_ptr = UICurrentUnitPtr + 1;
       i <= MaxUnitNo;
       i++, unit_ptr++)

    if ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {
      UICurrentUnitNo  = i;
      UICurrentUnitPtr = unit_ptr;

      if ( (unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
        {   /*  Initialize current site pointer to the first available site */
        UICurrentPredSitePtr = NULL;
        UICurrentSitePtr = unit_ptr->sites;
        }
      else
        {   /*  No sites available  */
        UICurrentPredSitePtr = NULL;
        UICurrentSitePtr     = NULL;
        }

      return( i );
      }

  return( 0 );
}


/*  Returns the no. of the current unit or 0 if no units available.
*/
int     krui_getCurrentUnit()
{
  return( UICurrentUnitNo );
}


/*  Initializes a unit for access. If the unit has sites, the first site
    will be set to the current site.
    Returns error code if the given unit doesn't exist, 0 otherwise.
*/
krui_err  krui_setCurrentUnit( unit_no )
int     unit_no;
{
  struct Unit   *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  UICurrentUnitNo  = unit_no;
  UICurrentUnitPtr = unit_ptr;

  if ( (unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
    {   /*  Initialize current site pointer to the first available site */
    UICurrentPredSitePtr = NULL;
    UICurrentSitePtr = unit_ptr->sites;
    }
  else
    {   /*  No sites available  */
    UICurrentPredSitePtr = NULL;
    UICurrentSitePtr     = NULL;
    }

  return( KRERR_NO_ERROR );
}


/*  Returns the name of the unit. (NULL if not available)
*/
char    *krui_getUnitName( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( NULL );    /*  invalid unit #  */

  return( unit_ptr->unit_name );
}


/*  Sets the name of the unit <UnitNo>.
    If the unit_name pointer is NULL, the unit's symbol will be deleted.
    Function has no effect on the current unit.
    Returns error code if memory allocation fails, 0 otherwise.
*/
krui_err  krui_setUnitName( unit_no, unit_name )
int     unit_no;
char    *unit_name;
{
  struct Unit   *unit_ptr;
  char    *str_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if (unit_name == NULL)
    {
    krm_NTableCheckSymbol( unit_ptr->unit_name, UNIT_SYM );
    unit_ptr->unit_name = NULL;
    return( KRERR_NO_ERROR );
    }

  if (!kr_symbolCheck( unit_name ))
    return( KRERR_SYMBOL );  /*  symbol pattern invalid  */

  if ( (str_ptr = krm_NTableInsertSymbol( unit_name, UNIT_SYM ) ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

  unit_ptr->unit_name = str_ptr;

  return( KRERR_NO_ERROR );
}


/*  Searches for a unit with the given name.
    Returns the first unit no. if a unit with the given name was found,
    0 otherwise.

    Returns error code if no units defined.
*/
int  krui_searchUnitName( unit_name )
char    *unit_name;
{
  struct  NameTable  *n_ptr;


  if (NoOfUnits <= 0)
    {
    UICurrentNameSearchUnitNo = 0;
    return( KRERR_NO_UNITS );  /*  No Units defined  */
    }

  if ((n_ptr = krm_NTableSymbolSearch( unit_name, UNIT_SYM )) == NULL)
    {
    UICurrentNameSearchUnitNo = 0;
    return( 0 );
    }

  UICurrentNameSearchUnitSymbolPtr = n_ptr->Entry.symbol;
  UICurrentNameSearchUnitNo = kr_unitNameSearch( MinUnitNo, UICurrentNameSearchUnitSymbolPtr );

  return( UICurrentNameSearchUnitNo );
}



/*  Searches for the next unit with the given name.
    Returns the next unit no. if a unit with the given name was found,
    0 otherwise.

NOTE: Call krui_searchUnitName( unit_name ) before calling krui_searchNextUnitName().

    Returns error code if no units defined.
*/
int  krui_searchNextUnitName()
{
  if (NoOfUnits <= 0)
    {
    UICurrentNameSearchUnitNo = 0;
    return( KRERR_NO_UNITS );  /*  No Units defined  */
    }

  if (UICurrentNameSearchUnitNo != 0)
    {
    UICurrentNameSearchUnitNo = kr_unitNameSearch( UICurrentNameSearchUnitNo + 1, UICurrentNameSearchUnitSymbolPtr );
    return( UICurrentNameSearchUnitNo );
    }
  else
    return( 0 );
}


/*  Returns the output function name of the unit.
*/
char    *krui_getUnitOutFuncName( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( NULL );    /*  invalid unit #  */

  return( krf_getFuncName( (FunctionPtr) unit_ptr->out_func ) );
}


/*  The unit's FType will be set to 0, i.e. the
    unit's functionality type will be deleted.
    Function has no effect on the current unit.
NOTE: returns 0, if the function is a valid output function,
      error code otherwise.
*/
int     krui_setUnitOutFunc( UnitNo, unitOutFuncName )
int     UnitNo;
char    *unitOutFuncName;
{
  struct Unit   *unit_ptr;
  FunctionPtr func_ptr;


  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if ( !krf_funcSearch( unitOutFuncName, OUT_FUNC, &func_ptr ))
    return( KRERR_OUTFUNC );  /*  Invalid unit output function  */

  unit_ptr->out_func = (OutFuncPtr) func_ptr;
  unit_ptr->Ftype_entry = NULL;

  return( KRERR_NO_ERROR );
}


/*  Returns the activation function name of the unit.
*/
char    *krui_getUnitActFuncName( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( NULL );    /*  invalid unit #  */

  return( krf_getFuncName( (FunctionPtr) unit_ptr->act_func ) );
}


/*  The unit's FType will be set to 0, i.e. the
    unit's functionality type will be deleted.
    Function has no effect on the current unit.
NOTE: returns 0, if the function is a valid activation function,
      error code otherwise.
*/
int     krui_setUnitActFunc( UnitNo, unitActFuncName )
int     UnitNo;
char    *unitActFuncName;
{
  struct Unit   *unit_ptr;
  FunctionPtr   act_func_ptr, act_deriv_func_ptr;


  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */
  if ( !krf_funcSearch( unitActFuncName, ACT_FUNC, &act_func_ptr ))
    return( KRERR_ACTFUNC );    /*  act func isn't defined  */
  /*  set the derivation function of the activation function  */ 
  if ( !krf_funcSearch( unitActFuncName, ACT_DERIV_FUNC, &act_deriv_func_ptr ))
    return( KRERR_DERIV_FUNC );    /*  Derivation function of the activation function doesn't exist  */

  unit_ptr->act_func = (ActFuncPtr) act_func_ptr;
  unit_ptr->act_deriv_func = (ActDerivFuncPtr) act_deriv_func_ptr;
  unit_ptr->Ftype_entry = NULL;

  return( KRERR_NO_ERROR );
}


/*  Returns the functionality type name of the unit.
    Function has no effect on the current unit.
    Returns NULL if unit has no FType.
*/
char    *krui_getUnitFTypeName( UnitNo )
int     UnitNo;
{
  struct FtypeUnitStruct    *Ftype_entry;
  struct Unit   *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( NULL );    /*  invalid unit #  */

  if ( (Ftype_entry = unit_ptr->Ftype_entry) == NULL)
    return( NULL );

  return( Ftype_entry->Ftype_symbol->Entry.symbol );
}





/*  Returns the activation value of the unit.
    Function has no effect on the current unit.
*/
FlintType   krui_getUnitActivation( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( (FlintType) 0 );    /*  invalid unit #  */

  return( unit_ptr->act );
}



/*  Sets the activation value of the unit.
    Function has no effect on the current unit.
*/
void    krui_setUnitActivation( UnitNo, unit_activation )
int     UnitNo;
FlintTypeParam  unit_activation;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->act = (FlintType) unit_activation;
}



/*  Returns the initial activation value of the unit.
    Function has no effect on the current unit.
*/
FlintType   krui_getUnitInitialActivation( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( (FlintType) 0 );    /*  invalid unit #  */

  return( unit_ptr->i_act );
}


/*  Sets the initial activation value of the unit.
    Function has no effect on the current unit.
*/
void    krui_setUnitInitialActivation( UnitNo, unit_i_activation )
int     UnitNo;
FlintTypeParam  unit_i_activation;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->i_act = (FlintType) unit_i_activation;
}



/*  Returns the output value of the unit.
    Function has no effect on the current unit.
*/
FlintType   krui_getUnitOutput( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( (FlintType) 0);    /*  invalid unit #  */

  return( unit_ptr->Out.output );
}


/*  Sets the output value of the unit.
    Function has no effect on the current unit.
*/
krui_err    krui_setUnitOutput( unit_no, unit_output )
int     unit_no;
FlintTypeParam  unit_output;
{
  struct  Unit      *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  unit_ptr->Out.output = (FlintType) unit_output;
  return( KRERR_NO_ERROR );
}



/*  Returns the bias value of the unit.
    Function has no effect on the current unit.
*/
FlintType   krui_getUnitBias( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( (FlintType) 0);    /*  invalid unit #  */

  return( unit_ptr->bias );
}



/*  Sets the bias value of the unit.
    Function has no effect on the current unit.
*/
void    krui_setUnitBias( UnitNo, unit_bias )
int     UnitNo;
FlintTypeParam  unit_bias;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->bias = (FlintType) unit_bias;
}



/*  Returns the subnet number of the unit.
    Function has no effect on the current unit.
NOTE: The range of the subnet no. is -32736...+32735
*/
int     krui_getUnitSubnetNo( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( 0 );    /*  invalid unit #  */

  return( (int) unit_ptr->subnet_no );
}

/*  Sets the subnet number of the unit.
    Function has no effect on the current unit.
NOTE: The range of the subnet no. is -32736...+32735
*/
void    krui_setUnitSubnetNo( UnitNo, subnet_no)
int     UnitNo,
        subnet_no;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->subnet_no = (short) subnet_no;
}


/*  Returns the layer number of the unit.
    Function has no effect on the current unit.
NOTE: The range of the layer no. is -32736...+32735
*/
unsigned short	krui_getUnitLayerNo( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( 0 );    /*  invalid unit #  */

  return( (unsigned short) unit_ptr->layer_no );
}

/*  Sets the layer number of the unit.
    Function has no effect on the current unit.
NOTE: The range of the layer no. is -32736...+32735
*/
void    krui_setUnitLayerNo( UnitNo, layer_no)
int     UnitNo;
int     layer_no;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->layer_no = (unsigned short) layer_no;
}



/*  Returns the position of the unit.
    Function has no effect on the current unit.
*/
void    krui_getUnitPosition( UnitNo, position )
int     UnitNo;
struct  PosType *position;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  position->x = unit_ptr->unit_pos.x;
  position->y = unit_ptr->unit_pos.y;
#ifdef KERNEL3D
  position->z = unit_ptr->unit_pos.z;
#endif
}

/*  Sets the position of the unit.
    Function has no effect on the current unit.
*/
void    krui_setUnitPosition( UnitNo, position )
int     UnitNo;
struct PosType  *position;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->unit_pos.x = position->x;
  unit_ptr->unit_pos.y = position->y;
#ifdef KERNEL3D
  unit_ptr->unit_pos.z = position->z;
#endif
}


/*  Returns the unit no. at the given position and subnet number or 0, if
    no unit exists at this position.
    Function has no effect on the current unit.
NOTE: This function is slow because it uses linear search to
      find the unit with the given position.

REMARK: getUnitNoAtPosition is for downward compatibility only.
        Do not use this function in future applications.
*/
int   krui_getUnitNoAtPosition( position, subnet_no )
struct PosType  *position;
int  subnet_no;
{
  register int       i;
  register short     x, y, net_no;
  struct Unit     *unit_ptr;


  x = position->x;
  y = position->y;

  net_no = (short) subnet_no;

  for (i = 1, unit_ptr = unit_array + 1;
       i <= MaxUnitNo;
       i++, unit_ptr++)
    {
    if ( ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE ) &&
         (unit_ptr->subnet_no == net_no) &&
         (unit_ptr->unit_pos.x == x) && (unit_ptr->unit_pos.y == y) )

      return( i );
    }

  return( 0 );
}


/*  Returns the unit no. near the given position and the specified
    subnet no or 0, if no unit exists at this position.
    Function has no effect on the current unit.
               
    NOTE: This function is slow because it uses linear search to
          find the unit with the given position.
*/
int  krui_getUnitNoNearPosition( position, subnet_no, range, gridWidth )
struct PosType  *position;
int   subnet_no, range, gridWidth;
{
  register int       i, devit, width;
  register short     x, y, net_no;
  struct Unit        *unit_ptr;


  x = position->x;
  y = position->y;
  net_no = (short) subnet_no;
  devit = range;
  width = gridWidth;

  for (i = 1, unit_ptr = unit_array + 1; i <= MaxUnitNo;
       i++, unit_ptr++)
    {
    if ( ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE ) &&
         (unit_ptr->subnet_no == net_no) &&
         (abs( unit_ptr->unit_pos.x - x) * width <= devit) && 
         (abs( unit_ptr->unit_pos.y - y) * width <= devit) )
      return( i );
    }

  return( 0 );
}



/*  Returns the topologic type, i.e. input, output, hidden
    Function has no effect on the current unit.
*/
int  krui_getUnitTType( UnitNo )
int  UnitNo;
{
  FlagWord  flags;
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );    /*  invalid unit #  */


  flags = unit_ptr->flags & UFLAG_TTYP_PAT;

  return( kr_getTType( flags ) );
}


/*  Sets the topologic type of the unit.
    Function has no effect on the current unit.
    Returns error code if topologic type or unit number is invalid.
*/
krui_err  krui_setUnitTType( unit_no, UnitTType )
int  unit_no,
     UnitTType;
{
  FlagWord	flags, flg;
  int		intflags;
  struct  Unit	*unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  flags = unit_ptr->flags & ~UFLAG_TTYP_PAT;

  if ( (intflags = kr_setTType( UnitTType )) == -1)
    return( KRERR_TTYPE );  /*  Topologic type invalid  */

  flg = (FlagWord)  intflags;
  flags |= flg;

  NetModified = TRUE;
  unit_ptr->flags = flags;

  return( KRERR_NO_ERROR );
}


/*  Freezes the output of a unit, i.e. the unit is disabled.
    Function has no effect on the current unit.
*/
void  krui_freezeUnit( UnitNo )
int  UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->flags &= (~UFLAG_ENABLED);
}


/*  Enables a previosly freezed unit.
    Function has no effect on the current unit.
*/
void    krui_unfreezeUnit( UnitNo )
int     UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return;    /*  invalid unit #  */

  unit_ptr->flags |= UFLAG_ENABLED;
}


/*  Returns true, if unit is frozen
*/
bool  krui_isUnitFrozen( UnitNo )
int   UnitNo;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( FALSE );    /*  invalid unit #  */

  return( (unit_ptr->flags & UFLAG_ENABLED) == 0 );
}



/*  Returns the input type of the unit:
     NO_INPUTS    : if the unit has not inputs (at least not now)
     SITES        : if the unit has one or more sites (and no direct input links !)
     DIRECT_LINKS : if the unit has direct input links (and no sites !)

    Function has no effect on the current unit.
*/
int  krui_getUnitInputType( UnitNo )
int  UnitNo;
{
  FlagWord    flags;
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( UnitNo ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */


  flags = unit_ptr->flags & UFLAG_INPUT_PAT;

  switch (flags)
    {
    case UFLAG_NO_INP :  return( NO_INPUTS    );
    case UFLAG_SITES  :  return( SITES        );
    case UFLAG_DLINKS :  return( DIRECT_LINKS );

    default :  return( NO_INPUTS );
    }
}


/*  *************************************   */

/*  creates a unit with default properties:
    1. default activation and output functions
    2. default activation and bias
    3. default position, subnet and layer number
    4. no functionality type
    5. no sites
    6. no inputs and outputs
    7. no unit_name

    Returns an (negative) error code, if memory allocation fails or
    invalid functions occured. Returns (positive) unit number otherwise.
    Function has no effect on the current unit.

NOTE: See file "kr_def.h" for more details about default presettings.
*/
int  krui_createDefaultUnit()
{
  int  unit_no;

  if ((unit_no = krm_getUnit() ) == 0)
    return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

  NetModified = TRUE;
  kr_makeDefaultUnit( unit_array + unit_no );

  return( unit_no );
}


/*  Creates a user defined unit.
    Returns an (negative) error code, if memory allocation fails or
    invalid functions occured. Returns (positive) unit number otherwise.
    Function has no effect on the current unit.

    Unit has additional default settings:
    1. default position, subnet and layer number
    2. no functionality type
    3. no sites
    4. no inputs and outputs

*/
int  krui_createUnit( unit_name, out_func_name, act_func_name,
		      i_act, bias)
char  *unit_name, *out_func_name, *act_func_name;
FlintTypeParam   i_act, bias;
{
  int   unit_no;
  char  *str_ptr;
  struct Unit   *unit_ptr;
  FunctionPtr   out_func_ptr, act_func_ptr, act_deriv_func_ptr;


  if (!kr_symbolCheck( unit_name ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if ( !krf_funcSearch( out_func_name, OUT_FUNC, &out_func_ptr ) )
    return( KRERR_OUTFUNC );  /*  Invalid unit output function  */
  if ( !krf_funcSearch( act_func_name, ACT_FUNC, &act_func_ptr ) )
    return( KRERR_ACTFUNC );  /*  Invalid unit activation function  */
  /*  set the derivation function of the activation function  */ 
  if ( !krf_funcSearch( act_func_name, ACT_DERIV_FUNC, &act_deriv_func_ptr ))
    return( KRERR_DERIV_FUNC );    /*  Derivation function of the activation function doesn't exist  */

  if ( (str_ptr = krm_NTableInsertSymbol( unit_name, UNIT_SYM ) ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */
  if ((unit_no = krm_getUnit() ) == 0)
    return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

  unit_ptr = unit_array + unit_no;
  kr_makeDefaultUnit( unit_ptr );

  unit_ptr->act       = (FlintType) i_act;
  unit_ptr->i_act     = (FlintType) i_act;
  unit_ptr->bias      = (FlintType) bias;
  unit_ptr->out_func  = (OutFuncPtr) out_func_ptr;
  unit_ptr->act_func  = (ActFuncPtr) act_func_ptr;
  unit_ptr->act_deriv_func = (ActDerivFuncPtr) act_deriv_func_ptr;
  unit_ptr->unit_name = str_ptr;

  NetModified = TRUE;
  return( unit_no );
}


/*  Deletes a unit from network. Removes all links to other
    units. Returns error code if unit doesn't exist.
*/
krui_err    krui_deleteUnit( unit_no )
int     unit_no;
{
  struct  Unit      *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  NetModified = TRUE;

  /*    delete inputs   */
  kr_deleteAllInputs( unit_ptr );

  /*    delete output links */
  kr_deleteAllOutputLinks( unit_ptr );

  /*    check references to the unit symbol   */
  krm_NTableCheckSymbol( unit_ptr->unit_name, UNIT_SYM );

  /*    delete Unit */
  krm_releaseUnit( unit_no );
  return( KRERR_NO_ERROR );
}


/*  creates a unit with FType properties, but:
    1. no inputs and outputs
    2. default position, subnet and layer

    Returns the unit number or (negative) error code if memory allocation
    fails or functionality type isn't defined.
    Function has no effect on the current unit.
*/
int   krui_createFTypeUnit( Ftype_symbol )
char  *Ftype_symbol;
{
  int     unit_no;
  struct  Unit      *unit_ptr;
  struct  FtypeUnitStruct   *ftype_ptr;


  if (!kr_symbolCheck( Ftype_symbol ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if ( (ftype_ptr = krm_FtypeSymbolSearch( Ftype_symbol ) ) == NULL)
    return( KRERR_FTYPE_SYMBOL );   /*  Ftype name isn't defined    */

  if ((unit_no = krm_getUnit() ) == 0)
    return( KRERR_INSUFFICIENT_MEM );   /*  memory alloc failed */

  NetModified = TRUE;
  unit_ptr = unit_array + unit_no;

  if ( kr_makeFtypeUnit( unit_ptr, ftype_ptr ) == NULL)
    {
    krm_releaseUnit( unit_no );
    return( KRERR_INSUFFICIENT_MEM );   /*  memory alloc failed */
    }

  return( unit_no );
}



/*  Changes the properties of unit <UnitNo> to FType properties.
    Changes: activation/output function and site name/functions.

    Returns an error code if
     - FType name doesn't exist or
     - unit doesn't exist or
     - memory allocation fails
     otherwise 0.

    Function has no effect on the current unit.
*/
int   krui_setUnitFType( unit_no, Ftype_symbol )
int   unit_no;
char  *Ftype_symbol;
{
  struct  Unit      *unit_ptr;
  struct  FtypeUnitStruct   *ftype_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if (!kr_symbolCheck( Ftype_symbol ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if ( (ftype_ptr = krm_FtypeSymbolSearch( Ftype_symbol ) ) == NULL)
    return( KRERR_FTYPE_SYMBOL );   /*  Ftype name isn't defined    */

  NetModified = TRUE;
  kr_changeFtypeUnit( unit_ptr, ftype_ptr );
  return( KRERR_NO_ERROR );
}



/*  Copy a given unit, according to the copy mode
        1. copy unit (with it sites, if available) and input/output links
        2. copy unit (with it sites, if available) and input links
        3. copy unit (with it sites, if available) and output links
        4. copy unit (with it sites, if available) but no input/output links

    Returns the unit number of the new unit or error message < 0 , if errors occured.
    Function has no effect on the current unit.

NOTE: Copying of output links is slow.
      If return code < 0, an error occured.
*/
int   krui_copyUnit( unit_no, copy_mode)
int   unit_no,
      copy_mode;
{
  FlagWord   flags;
  int   new_unit_no;
  struct Unit   *new_unit_ptr,
                *source_unit_ptr;


  if ( (source_unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if ( (new_unit_no = krm_getUnit() ) == 0 )
    return( KRERR_INSUFFICIENT_MEM );   /*  memory alloc failed */

  new_unit_ptr = unit_array + new_unit_no;

  /*    copy unit   */
  kr_copyUnit( source_unit_ptr, new_unit_ptr );

  flags = source_unit_ptr->flags;

  switch (copy_mode)
    {
    case INPUTS_AND_OUTPUTS :
    /*  copy unit (with it sites, if available) and input/output links  */
      if ( kr_copyOutputLinks( source_unit_ptr, new_unit_ptr) != 0)
	{
	/*    delete Unit */
	krm_releaseUnit( new_unit_no );
        return( KRERR_CRITICAL_MALLOC );   /*  mem alloc failed during critical op */
	}

    case ONLY_INPUTS :
    /*  copy unit (with it sites, if available) and input links */
      if (source_unit_ptr->sites == NULL)
        break;
      if ( (flags & UFLAG_INPUT_PAT) == UFLAG_SITES)
        {   /*  unit has sites  */
        if ( (new_unit_ptr->sites = kr_copySiteListAndLinks( source_unit_ptr->sites ) ) == NULL)
	  {
	  /*	delete Unit */
	  krm_releaseUnit( new_unit_no );
          return( KRERR_CRITICAL_MALLOC ); /*  mem alloc failed during critical op */
	  }
        }
      else
        {   /*  unit don't has sites    */
        if ( (new_unit_ptr->sites = (struct Site *) kr_copyLinkList( (struct Link *) source_unit_ptr->sites) ) == NULL)
	  {
	  /*	delete Unit */
	  krm_releaseUnit( new_unit_no );
          return( KRERR_CRITICAL_MALLOC ); /*  mem alloc failed during critical op */
	  }
        }

      break;


    case ONLY_OUTPUTS :
    /*  copy unit (with it sites, if available) and output links    */
      if ( kr_copyOutputLinks( source_unit_ptr, new_unit_ptr) != 0)
	{
	/*    delete Unit */
	krm_releaseUnit( new_unit_no );
        return( KRERR_CRITICAL_MALLOC );   /*  mem alloc failed during critical op */
	}

    case ONLY_UNIT :
    /*  copy unit (with it sites, if available) but no input/output links   */
      if ( (flags & UFLAG_INPUT_PAT) == UFLAG_SITES)
        {   /*  unit has sites  */
        if ( (new_unit_ptr->sites = kr_copySiteList( source_unit_ptr->sites ) ) == NULL)
	  {
	  /*	delete Unit */
	  krm_releaseUnit( new_unit_no );
          return( KRERR_CRITICAL_MALLOC ); /*  mem alloc failed during critical op */
	  }
        }

      break;

    default:
      /*    delete Unit */
      krm_releaseUnit( new_unit_no );
      return( KRERR_COPYMODE );    /*  invalid copy mode   */
    }


  /*  Successful copy   */
  new_unit_ptr->flags = source_unit_ptr->flags;     /*  copy flags  */

  NetModified = TRUE;
  return( new_unit_no );
}



/*#################################################

GROUP: Functions for manipulation of the Unit-Functionality-List
REMEMBER: The Unit-Functionality-List stores:
            1. unit activation and output functions
            2. if sites:  3. all site functions
                          4. all site names


#################################################*/

/*  Initializes the first FType entry.
    Returns true, if an entry is available
*/
bool  krui_setFirstFTypeEntry()
{
  UICurrentFtypeEntry = krm_getFtypeFirstEntry();
  UICurrentFtypeSite = NULL;
  return( UICurrentFtypeEntry != NULL );
}

/*  Initializes the next FType entry.
    Returns true, if an entry is available
*/
bool  krui_setNextFTypeEntry()
{
  struct  FtypeUnitStruct   *ftype_ptr;


  UICurrentFtypeSite = NULL;
  if ( (ftype_ptr = krm_getFtypeNextEntry() ) != NULL)
    {
    UICurrentFtypeEntry = ftype_ptr;
    return( TRUE );
    }

  return( FALSE );
}


/*  Initializes the FType entry with the given name.
    Returns true, if an entry with this name is available.
*/
bool  krui_setFTypeEntry( Ftype_symbol )
char  *Ftype_symbol;
{
  struct  FtypeUnitStruct   *ftype_ptr;


  if ( (ftype_ptr = krm_FtypeSymbolSearch( Ftype_symbol ) ) != NULL)
    {
    UICurrentFtypeEntry = ftype_ptr;
    return( TRUE );
    }

  return( FALSE );
}


/*  Returns the name of the current FType entry.
NOTE: The FType name is definite and will never be
      a NULL pointer.
*/
char       *krui_getFTypeName()
{
  if (UICurrentFtypeEntry == NULL)
    return( NULL );

  return( UICurrentFtypeEntry->Ftype_symbol->Entry.symbol );
}


/*  Sets the name of the current FType entry.
    Returns error code if memory allocation fails or Ftype name
    isn't definite, 0 otherwise.
NOTE: The new FType name have to be definite and must not be
      a NULL pointer.
*/
krui_err  krui_setFTypeName( Ftype_symbol )
char  *Ftype_symbol;
{
  struct NameTable    *NTable_ptr;


  if (UICurrentFtypeEntry == NULL)
    return( KRERR_FTYPE_ENTRY );   /*  Current Ftype entry isn't defined   */

  if (Ftype_symbol == NULL)
    return( KRERR_FTYPE_NAME );    /*  Ftype name isn't definite   */

  if (!kr_symbolCheck( Ftype_symbol ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if (krm_NTableSymbolSearch( Ftype_symbol, FTYPE_UNIT_SYM ) != NULL)
    return( KRERR_FTYPE_NAME );    /*  Ftype name isn't definite   */

  if ( (NTable_ptr = krm_NTableCreateEntry( Ftype_symbol, FTYPE_UNIT_SYM ) ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

  krm_NTableReleaseEntry( UICurrentFtypeEntry->Ftype_symbol );

  UICurrentFtypeEntry->Ftype_symbol = NTable_ptr;

  return( KRERR_NO_ERROR );
}


/*  Returns the name of the activation function in the current
    FType entry.
*/
char  *krui_getFTypeActFuncName()
{
  if (UICurrentFtypeEntry == NULL)
    return( NULL );

  return( krf_getFuncName( (FunctionPtr) UICurrentFtypeEntry->act_func ) );
}


/*  Sets the activation function in the current FType entry
    returns 0, if the function is a valid activation function,
    error code otherwise.
    All units (in the existing network) of the current Ftype changes
    their activation function.
*/
krui_err  krui_setFTypeActFunc( act_func_name )
char   *act_func_name;
{
  FunctionPtr  act_func_ptr, act_deriv_func_ptr;


  if (UICurrentFtypeEntry == NULL)
    return( KRERR_FTYPE_ENTRY );   /*  Current Ftype entry isn't defined   */

  if ( !krf_funcSearch( act_func_name, ACT_FUNC, &act_func_ptr ) )
    return( KRERR_ACTFUNC );    /*  act func isn't defined  */
  /*  set the derivation function of the activation function  */ 
  if ( !krf_funcSearch( act_func_name, ACT_DERIV_FUNC, &act_deriv_func_ptr ))
    return( KRERR_DERIV_FUNC );    /*  Derivation function of the activation function doesn't exist  */

  UICurrentFtypeEntry->act_func = (ActFuncPtr) act_func_ptr;
  UICurrentFtypeEntry->act_deriv_func = (ActDerivFuncPtr) act_deriv_func_ptr;

  kr_changeFtypeUnits( UICurrentFtypeEntry );
  return( KRERR_NO_ERROR );
}


/*  Returns the name of the output function in the current
    FType entry.
*/
char  *krui_getFTypeOutFuncName()
{
  if (UICurrentFtypeEntry == NULL)
    return( NULL );

  return( krf_getFuncName( (FunctionPtr) UICurrentFtypeEntry->out_func ) );
}




/*  Sets the output function in the current FType entry
    Returns 0, if the function is a valid output function,
    error code otherwise.
    All units (in the existing network) of the current Ftype changes
    their output function.
*/
krui_err    krui_setFTypeOutFunc( out_func_name )
char   *out_func_name;
{
  FunctionPtr	 out_func_ptr;


  if (UICurrentFtypeEntry == NULL)
    return( KRERR_FTYPE_ENTRY );   /*  Current Ftype entry isn't defined   */

  if ( !krf_funcSearch( out_func_name, OUT_FUNC, &out_func_ptr ) )
    return( KRERR_OUTFUNC );    /*  out func isn't defined  */

  UICurrentFtypeEntry->out_func = (OutFuncPtr) out_func_ptr;

  kr_changeFtypeUnits( UICurrentFtypeEntry );
  return( KRERR_NO_ERROR );
}



/*  FType sites
*/

/*  Initializes the first site of the current FType.
    Returns FALSE, if no sites are available in the current
    FType entry.
*/
bool  krui_setFirstFTypeSite()
{
  if (UICurrentFtypeEntry == NULL)
    {
    UICurrentFtypeSite = NULL;
    return( FALSE );
    }

  if (UICurrentFtypeEntry->sites == NULL)
    {
    UICurrentFtypeSite = NULL;
    return( FALSE );
    }
  else
    {
    UICurrentFtypeSite = UICurrentFtypeEntry->sites;
    return( TRUE );
    }
}


/*  Initializes the next FType site.
    Returns FALSE, if no more sites are available in the current
    FType entry.
*/
bool  krui_setNextFTypeSite()
{
  if (UICurrentFtypeSite == NULL)
    return( FALSE );

  if (UICurrentFtypeSite->next == NULL)
    return( FALSE );

  UICurrentFtypeSite = UICurrentFtypeSite->next;
  return( TRUE );
}


/*  Returns the name of the current FType site (in the current
    Ftype entry).
*/
char  *krui_getFTypeSiteName()
{
  if (UICurrentFtypeSite == NULL)
    return( NULL );

  return( ((UICurrentFtypeSite->site_table)->Entry.site_name)->Entry.symbol );
}



/*  Sets the name and function of the current FType site (in the
    current FType entry).
    All sites (in the existing network) of the current Ftype and the
    same (old) name changes their names and site functions.

    Returns an error code if
      - current Ftype site isn't defined
      - site name does not exist in the site name table
      0 otherwise.
*/
krui_err  krui_setFTypeSiteName( FType_site_name )
char  *FType_site_name;
{
  struct  SiteTable     *stbl_ptr;


  if (UICurrentFtypeSite == NULL)
    return( KRERR_FTYPE_SITE );   /*  current Ftype site isn't defined */

  if (!kr_symbolCheck( FType_site_name ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if ( (stbl_ptr = krm_STableSymbolSearch( FType_site_name )) == NULL)
    return( KRERR_UNDEF_SITE_NAME );   /*   site name isn't defined */

  kr_changeFtypeSites( UICurrentFtypeEntry, UICurrentFtypeSite->site_table, stbl_ptr);
  UICurrentFtypeSite->site_table = stbl_ptr;

  return( KRERR_NO_ERROR );
}



/*  Create a new functionality type, needs a definite FType symbol,
    the unit output and activation function
    and the number of sites provided for this unit FType.
    An additional array with N elements of pointers to site names
    is required for the definition of the sites.

    Returns error code if:
     - memory allocation fails
     - FType name isn't definite (symbol is used for another FType
       or symbol is a NULL pointer)
     - one or more site names are undefined

     0 otherwise.

NOTE: The number of Ftype entries and the number of sites per Ftype is
      only limited by the size of system memory.
*/
krui_err  krui_createFTypeEntry( Ftype_symbol, act_func_name, out_func_name,
				 no_of_sites, array_of_site_names )
char  *Ftype_symbol,
      *out_func_name,
      *act_func_name;
int   no_of_sites;
char  * *array_of_site_names;
{
  FunctionPtr  func_ptr;
  ActFuncPtr   act_func;
  OutFuncPtr   out_func;
  ActDerivFuncPtr act_deriv_func;
  struct  FtypeUnitStruct   *Ftype_entry;
  struct  SiteTable         *STable_entry;
  int       i;
  char      *Ftype_site_name;


  if (Ftype_symbol == NULL)
    return( KRERR_FTYPE_NAME );    /*  Ftype name isn't definite   */

  if (!kr_symbolCheck( Ftype_symbol ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */
  if (krm_NTableSymbolSearch( Ftype_symbol, FTYPE_UNIT_SYM ) != NULL)
    return( KRERR_FTYPE_NAME );    /*  Ftype name isn't definite   */
  if ( !krf_funcSearch( out_func_name, OUT_FUNC, &func_ptr ) )
    return( KRERR_OUTFUNC );    /*  out func isn't defined  */
  out_func = (OutFuncPtr) func_ptr;

  if ( !krf_funcSearch( act_func_name, ACT_FUNC, &func_ptr ) )
    return( KRERR_ACTFUNC );    /*  act func isn't defined  */
  act_func = (ActFuncPtr) func_ptr;

  /*  set the derivation function of the activation function  */ 
  if ( !krf_funcSearch( act_func_name, ACT_DERIV_FUNC, &func_ptr ))
    return( KRERR_DERIV_FUNC );    /*  Derivation function of the activation function doesn't exist  */
  act_deriv_func = (ActDerivFuncPtr) func_ptr;

  if ( (Ftype_entry = krm_FtypeCreateEntry( Ftype_symbol, out_func, act_func, act_deriv_func ) ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );    /*  memory alloc failed */

  /*  create sites at the Ftype  */
  for (i = 0; i < no_of_sites; i++)
    {
    if ( (Ftype_site_name = array_of_site_names[ i ]) == NULL)
      return( KRERR_UNDEF_SITE_NAME );   /*   site name isn't defined because it is a NULL pointer  */

    if ( (STable_entry = krm_STableSymbolSearch( Ftype_site_name )) == NULL)
      {  /*  site isn't defined  */
      krm_releaseFtypeEntry( Ftype_entry );
      return( KRERR_UNDEF_SITE_NAME );   /*   site name isn't defined */
      }

    if (krm_FtypeAddSite( Ftype_entry , STable_entry ) == NULL)
      {  /*  memory alloc failed */
      krm_releaseFtypeEntry( Ftype_entry );
      return( KRERR_INSUFFICIENT_MEM );      /*  memory alloc failed */
      }

    }

  return( KRERR_NO_ERROR );
}


/*  Deletes the specified FType entry. If there exists units in the
    network with this FType, all these units will lost their FType
    but the functionality of the units will not be changed.
    Returns error code if FType symbol dosn't exist, 0 otherwise.
*/
krui_err    krui_deleteFTypeEntry( Ftype_symbol )
char   *Ftype_symbol;
{
  struct  FtypeUnitStruct   *ftype_ptr;


  if ( (ftype_ptr = krm_FtypeSymbolSearch( Ftype_symbol ) ) == NULL)
    return( KRERR_FTYPE_SYMBOL );   /*  FType symbol isn't defined  */

  kr_deleteUnitsFtype( ftype_ptr );
  krm_releaseFtypeEntry( ftype_ptr );
  return( KRERR_NO_ERROR );
}



/*#################################################

GROUP: Functions for reading of the function table

#################################################*/

/*  Returns the number of available functions
*/
int  krui_getNoOfFunctions()
{
  return( krf_getNoOfFuncs() );
}


/*  Returns the name of the function and the function type (Output,
    Activation, Site).
*/
void  krui_getFuncInfo( func_no, func_name, func_type )
int     func_no;
char    * *func_name;
int     *func_type;
{
  struct FuncTable   *f_ptr;


  if ( (func_no < 1) || (func_no > krf_getNoOfFuncs() ))
    {
    *func_name = NULL;
    *func_type = 0;
    }
  else
    {
    f_ptr = krf_getFuncTableEntry( func_no );
    *func_name = f_ptr->func_name;
    *func_type = f_ptr->func_type;
    }
}


/*  Returns true if the given function name and type exists.
*/
bool  krui_isFunction( func_name, func_type )
char    *func_name;
int     func_type;
{
  FunctionPtr     dummy_func_ptr;

  return( krf_funcSearch( func_name, func_type, &dummy_func_ptr ) );
}


/*  Returns the no. of input and output parameters of the given
    function (only relevant for learning, update and initialisation
    functions).

    Returns TRUE if the given function exists, FALSE otherwise.
*/
bool  krui_getFuncParamInfo( func_name, func_type,
                             no_of_input_params,
                             no_of_output_params )
char *func_name;
int  func_type;
int *no_of_input_params, *no_of_output_params;
{
  struct FuncTable   *f_ptr;


  if ((f_ptr = krf_getFuncTblPtr( func_name, func_type )) == NULL)
    return( FALSE );

  *no_of_input_params  = (int) f_ptr->no_of_input_parameters;
  *no_of_output_params = (int) f_ptr->no_of_output_parameters;

  return( TRUE );
}



/*#################################################

GROUP: Site Table Functions

#################################################*/


/*  Returns the first site name/function pair in the site table.
    Returns FALSE and NULL, if not available.
*/
bool   krui_getFirstSiteTableEntry( site_name, site_func )
char   * *site_name,
       * *site_func;
{
  struct  SiteTable   *s_ptr;


  if ( (s_ptr = krm_getSTableFirstEntry() ) == NULL)
    {
    *site_name = NULL;
    *site_func = NULL;
    return( FALSE );
    }
  else
    {
    *site_name = (s_ptr->Entry.site_name)->Entry.symbol;
    *site_func = krf_getFuncName( (FunctionPtr) s_ptr->site_func );
    return( TRUE );
    }
}



/*  Returns the next site name/function pair in the site table.
    Returns FALSE and NULL, if not available.
*/
bool  krui_getNextSiteTableEntry( site_name, site_func )
char  * *site_name,
      * *site_func;
{
  struct  SiteTable   *s_ptr;


  if ( (s_ptr = krm_getSTableNextEntry() ) == NULL)
    {
    *site_name = NULL;
    *site_func = NULL;
    return( FALSE );
    }
  else
    {
    *site_name = (s_ptr->Entry.site_name)->Entry.symbol;
    *site_func = krf_getFuncName( (FunctionPtr) s_ptr->site_func );
    return( TRUE );
    }
}



/*  Returns the name of the site function that is associated with
    the site name.
    If the site name do not exist, function returns NULL.
*/
char  *krui_getSiteTableFuncName( site_name )
char    *site_name;
{
  struct SiteTable  *s_ptr;


  if ( (s_ptr = krm_STableSymbolSearch( site_name )) == NULL)
    return( NULL );

  return( krf_getFuncName( (FunctionPtr) s_ptr->site_func ) );
}



/*  Creates a new site name and associate this name with a site
    function.
    Returns error code if:
     - site name already exists or
     - site function is invalid or
     - memory allocation has failed
    0 otherwise.
*/
krui_err  krui_createSiteTableEntry( site_name, site_func )
char    *site_name,
        *site_func;
{
  FunctionPtr   func_ptr;
  SiteFuncPtr	site_func_ptr;

  if ( !krf_funcSearch( site_func, SITE_FUNC, &func_ptr ) )
    return( KRERR_SITEFUNC );    /*  site function isn't defined */
  site_func_ptr = (SiteFuncPtr) func_ptr;

  if (!kr_symbolCheck( site_name ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if (krm_STableSymbolSearch( site_name ) != NULL)
    return( KRERR_REDEF_SITE_NAME );   /*  symbol is already in the site table */

  if ( krm_STableCreateEntry( site_name, site_func_ptr ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );    /*  memory allocation failed    */

  return( KRERR_NO_ERROR );
}


/*  Changes the site function of a previously defined site name.
    Returns error code if <old_site_name> or <new_site_func>
    isn't defined, 0 otherwise.

NOTE: All sites in the network with the name <old_site_name>
      changes their names and functions.
*/
krui_err  krui_changeSiteTableEntry( old_site_name, new_site_name, new_site_func )
char  *old_site_name,
      *new_site_name,
      *new_site_func;
{
  FunctionPtr   func_ptr;
  SiteFuncPtr	    site_func_ptr;
  struct SiteTable  *stbl_ptr1,
                    *stbl_ptr2;


  if (!kr_symbolCheck( new_site_name ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if ( !krf_funcSearch( new_site_func, SITE_FUNC, &func_ptr ) )
    return( KRERR_SITEFUNC );    /*  site function isn't defined */
  site_func_ptr = (SiteFuncPtr) func_ptr;

  if ( (stbl_ptr1 = krm_STableSymbolSearch( old_site_name )) == NULL)
    return( KRERR_UNDEF_SITE_NAME );   /*  old site name isn't defined */

  stbl_ptr2 = krm_STableSymbolSearch( new_site_name );
  if ( (stbl_ptr2 != NULL) && (stbl_ptr2 != stbl_ptr1) )
    return( KRERR_REDEF_SITE_NAME );   /*  new symbol is already in the site table
                        ( and new_site_name and old_site_name are not identical) */

  if (krm_STableChangeEntry( stbl_ptr1 , new_site_name , site_func_ptr ) == NULL)
    return( KRERR_INSUFFICIENT_MEM );  /*  memory alloc failed  */

  return( KRERR_NO_ERROR );
}


/*  Removes the current site name entry from the site table.

    Returns an error code if
     - there exists sites with the given name in the network or
     - <site_name> isn't defined
     0 otherwise.
*/
krui_err  krui_deleteSiteTableEntry( site_name )
char   *site_name;
{
  struct SiteTable    *st_ptr;


  if ((st_ptr = krm_STableSymbolSearch( site_name )) == NULL)
    return( KRERR_UNDEF_SITE_NAME );   /*  Site name isn't defined */

  if (kr_searchNetSite( st_ptr ) != 0)
    return( KRERR_INUSE_SITE );   /*  Site is in use and must not be deleted  */

  krm_STableRemoveEntry( st_ptr );
  return( KRERR_NO_ERROR );
}


/*#################################################

GROUP: Site Functions

#################################################*/




/*  Initializes the first site at the current unit.
    Returns false if no site available or if no sites permitted at this unit.
*/
bool  krui_setFirstSite()
{
  UICurrentPredSitePtr = NULL;
  if (UICurrentUnitPtr == NULL)
    return( FALSE );

  if ( (UICurrentUnitPtr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
    {   /*  Unit has sites  */
    UICurrentSitePtr = UICurrentUnitPtr->sites;
    return( TRUE );
    }
  else
    {
    UICurrentSitePtr = NULL;
    return( FALSE );
    }
}


/*  Initializes the next site at the current unit.
    Returns false if no more sites available.
*/
bool  krui_setNextSite()
{
  if (UICurrentUnitPtr == NULL)
    return( FALSE );

  if ( (UICurrentSitePtr == NULL) ||
       (UICurrentSitePtr->next == NULL) )  return( FALSE );

  UICurrentPredSitePtr = UICurrentSitePtr;
  UICurrentSitePtr     = UICurrentSitePtr->next;
  return( TRUE );
}


/*  Initializes the given site at the current unit.
    Returns error code if
     - unit dosn't exist
     - site name doesn't exist
     - unit don't has sites
     - unit don't has a site with this name
     0 otherwise.
*/

krui_err  krui_setSite( site_name )
char    *site_name;
{
  struct  SiteTable     *stbl_ptr;


  if (UICurrentUnitPtr == NULL)
    return( KRERR_UNIT_NO );  /*  Invalid unit number  */

  if ( (UICurrentUnitPtr->flags & UFLAG_INPUT_PAT) != UFLAG_SITES )
    return( KRERR_NO_SITES );  /*  Current unit doesn't have sites  */

  if ( (stbl_ptr = krm_STableSymbolSearch( site_name )) == NULL)
    return( KRERR_UNDEF_SITE_NAME );   /*   site name isn't defined */

  kr_searchUnitSiteWithPred( UICurrentUnitPtr, stbl_ptr,
                             &UICurrentSitePtr, &UICurrentPredSitePtr );

  if (UICurrentSitePtr == NULL)
    return( KRERR_NO_SUCH_SITE );  /*  Current unit doesn't have a site with this name  */

  return( KRERR_NO_ERROR );
}




/*  Returns the actual value of the current site
*/
FlintType   krui_getSiteValue()
{
  if (UICurrentSitePtr != NULL)
    return( (*UICurrentSitePtr->site_table->site_func) (UICurrentSitePtr) );
  else
    return( (FlintType) 0);

}


/*  returns the name of the current unit/site,
    NULL if not available.
*/
char  *krui_getSiteName()
{
  if (UICurrentSitePtr == NULL)  return( NULL );

  return( ((UICurrentSitePtr->site_table)->Entry.site_name)->Entry.symbol );
}

/*  Sets the name/function of the current unit/site.
    Current Unit will loose the functionality type.
    Returns error code if site name isn't defined.
*/
krui_err  krui_setSiteName( site_name )
char  *site_name;
{
  struct  SiteTable     *stbl_ptr;


  if (UICurrentSitePtr == NULL)
    return( KRERR_FTYPE_SITE );   /*  current Ftype site isn't defined */

  if (!kr_symbolCheck( site_name ))
    return( KRERR_SYMBOL );  /*  Symbol pattern invalid (must match [A-Za-z]^[|, ]*)  */

  if ( (stbl_ptr = krm_STableSymbolSearch( site_name )) == NULL)
    return( KRERR_UNDEF_SITE_NAME );   /*   site name isn't defined */

  UICurrentUnitPtr->Ftype_entry = NULL;
  UICurrentSitePtr->site_table  = stbl_ptr;
  return( KRERR_NO_ERROR );
}




/*  Returns the name of the current unit/site function
*/
char  *krui_getSiteFuncName()
{
  if (UICurrentSitePtr == NULL)  return( NULL );

  return( krf_getFuncName( (FunctionPtr) UICurrentSitePtr->site_table->site_func ) );
}




/*  Adds a site at the current unit. If the unit has already sites, this new
    site will be inserted above all other sites, i.e. the new created site will
    be the first site at this unit.
    If the unit has direct input links, i.e the unit has input links
    but no sites, the creation of sites is not permitted (krui_addSite will
    return an error code).
    If there exists already a site with the given name, the creation of the
    new site is prohibited and krui_addSite returns an error code.

    krui_addSite has no effect on the current site. To change the current site
    to this new site, call krui_setFirstSite().
    The unit's FType will be deleted.

    Returns error code if
      - memory allocation fails or
      - unit has direct input links or
      - site name isn't defined or
      - site with the given name exists already at this unit
      0 otherwise.

NOTE: The number of sites per unit is nearly unlimited (2^32).
*/
krui_err  krui_addSite( site_name )
char  *site_name;
{
  FlagWord    flags;
  struct Site       *site_ptr,
                    *site_ptr1;
  struct SiteTable  *stbl_ptr;


  if ( (stbl_ptr = krm_STableSymbolSearch( site_name )) == NULL)
    return( KRERR_UNDEF_SITE_NAME );   /*  site name isn't defined */

  flags = UICurrentUnitPtr->flags & UFLAG_INPUT_PAT;

  switch (flags)
    {
    case UFLAG_NO_INP :
        /*  Unit has no inputs  */
	if ( (site_ptr = kr_createDefaultSite() ) == NULL)
          return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

        UICurrentUnitPtr->sites = site_ptr;     /*  Connect site    */
        UICurrentUnitPtr->flags |= UFLAG_SITES; /*  Set site flag   */

        UICurrentUnitPtr->Ftype_entry = NULL;   /*  Delete Ftype    */

        break;

    case UFLAG_SITES :
        /*  Unit has already sites  */
        if (kr_searchUnitSite( UICurrentUnitPtr, stbl_ptr ) != NULL)
          return( KRERR_DUPLICATED_SITE ); /*  there exists already a site with this name at
                            this unit   */

        if ( (site_ptr = kr_createDefaultSite() ) == NULL)
          return( KRERR_INSUFFICIENT_MEM );  /*  memory alloc fails  */

        site_ptr1 = UICurrentUnitPtr->sites;
        UICurrentUnitPtr->sites = site_ptr;     /*  Connect site    */
        site_ptr->next = site_ptr1;

        UICurrentUnitPtr->Ftype_entry = NULL;   /*  Delete Ftype    */

        break;

    case UFLAG_DLINKS :
        /*  Unit has direct input links and can't have sites */
        return( KRERR_CREATE_SITE );  /*  Creation of sites isn't permitted because unit has direct input links  */
    default : return( KRERR_CREATE_SITE );
    }

  site_ptr->site_table = stbl_ptr;

  NetModified = TRUE;
  return( KRERR_NO_ERROR );
}


/*  Removes the current site at the current unit and removes all
    links from predecessor units to this site.
    krui_setFirstSite (krui_setNextSite) must be called at least once
    before using this function.
    The current site will be set to the next available site, if no more
    sites available, krui_deleteSite returns FALSE otherwise TRUE.
    The unit's FType will be set to 0, i.e. the unit's functionality
    type will be deleted.

NOTE: To delete all sites at a unit:
        if ( krui_setFirstSite() )
          while ( krui_deleteSite() ) { }
*/
bool  krui_deleteSite()
{
  struct Site   *next_site_ptr;
  struct Unit   *unit_ptr;


  NetModified = TRUE;
  unit_ptr = UICurrentUnitPtr;

  if ( ((unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES) &&
       (unit_ptr->sites != NULL) &&
       (UICurrentSitePtr != NULL)   )
    {   /*  Unit has sites  */

    next_site_ptr = UICurrentSitePtr->next;

    krm_releaseAllLinks( UICurrentSitePtr->links );  /*  Remove links    */
    krm_releaseSite( UICurrentSitePtr );             /*  Remove site     */

    if (UICurrentPredSitePtr == NULL)
      {     /*  This site is the first site at the current unit */

      if (next_site_ptr == NULL)
        {   /*  Unit has only this site */
        unit_ptr->sites = NULL;                 /*  Clear site pointer  */
        unit_ptr->flags &= (~UFLAG_INPUT_PAT);  /*  Clear input flags   */

        UICurrentSitePtr     = NULL;            /*  No more sites available  */
        UICurrentPredSitePtr = NULL;
        }
      else
        {   /*  It is the first site at the unit but not the only one   */
        unit_ptr->sites  = next_site_ptr;       /*  Connect the other sites */
        UICurrentSitePtr = next_site_ptr;       /*  Current site ptr points to the next site    */
        }
      }
    else
      {     /*  This is not the first site at the unit  */
      UICurrentPredSitePtr->next = next_site_ptr;   /*  Connect the previous site
                                                        with the next site  */
      UICurrentSitePtr = next_site_ptr; /*  Current site ptr points to the next site    */
      }

    unit_ptr->Ftype_entry = NULL;   /*  Delete unit's Ftype */
    }

  return( UICurrentSitePtr != NULL );   /*  Returns true if more sites available    */
}



/*#################################################

GROUP: Link Functions

#################################################*/

/*  Returns the no. of first predecessor unit of the current unit/site
    and the connection strenght.
    Returns 0 if no predecessor unit available, i.e. if the current unit
    and/or site has no inputs.

NOTE: If a predecessor unit exists, the current link is set to the link
      between the two units.
*/
int  krui_getFirstPredUnit( strength )
FlintType  *strength;
{
  struct Link   *link_ptr;

  if (UICurrentUnitPtr == NULL)
    return( 0 );

  UICurrentPredLinkPtr = NULL;

  if ( (UICurrentUnitPtr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
    /*  Unit has sites  */
    {
    if (UICurrentSitePtr == NULL)
      return( 0 );

    link_ptr = UICurrentSitePtr->links;
    }
  else
    /*  Unit has no sites   */
    link_ptr = (struct Link *) UICurrentUnitPtr->sites;


  if (link_ptr == NULL)
    {     /*  No inputs   */
    UICurrentLinkPtr = NULL;

    return( 0 );
    }
  else
    {     /*  Inputs available    */
    UICurrentLinkPtr = link_ptr;
    *strength = link_ptr->weight;

    UICurrentLinkUnitNo = UICurrentLinkPtr->to - unit_array;
    return( UICurrentLinkUnitNo );    /*  Return Unit number  */
    }
}



/*  Returns the no. of the next predecessor unit of the current
    unit/site and the connection strenght.
    Returns 0 if no more predecessor units exists.

NOTE: If another predecessor unit exists, the current link is set to the link
      between the two units.
*/
int  krui_getNextPredUnit( strength )
FlintType   *strength;
{
  if (UICurrentUnitPtr == NULL)
    return( 0 );

  if ( (UICurrentLinkPtr == NULL) ||
       (UICurrentLinkPtr->next == NULL) )  return( 0 );

  UICurrentPredLinkPtr = UICurrentLinkPtr;
  UICurrentLinkPtr     = UICurrentLinkPtr->next;
  *strength = UICurrentLinkPtr->weight;

  UICurrentLinkUnitNo = UICurrentLinkPtr->to - unit_array;
  return( UICurrentLinkUnitNo );    /*  Return Unit number  */
}




/*  Returns the no. of the current predecessor unit (of the current
    unit/site) and the connection strenght.
    Returns 0 if no predecessor unit available, i.e. if the current unit
    and/or site has no inputs.
*/
int  krui_getCurrentPredUnit( strength )
FlintType   *strength;
{
  if (UICurrentUnitPtr == NULL)
    return( 0 );

  if (UICurrentLinkPtr == NULL)  return( 0 );

  *strength = UICurrentLinkPtr->weight;
  return( UICurrentLinkUnitNo );    /*  Return Unit number  */
}



/*  Returns the no. of the first successor unit of the unit <unit_no>
    and the connection strenght.
    Returns (negative) error code if unit doesn't exist.
    Returns 0 if no successor unit available, i.e. if the current unit
    has no output connection.

IMPORTANT: If a successor unit exists, the current unit and site will be
           set to this successor unit and the attached site.

NOTE: This function is slow (Units are backward chained only).

REMARK: getFirstSuccUnit is for downward compatibility only.
        Do not use this function in future applications.
*/
int  krui_getFirstSuccUnit( source_unit_no, strength )
int  source_unit_no;
FlintType   *strength;
{
  register struct Link   *link_ptr,
                         *pred_link_ptr;
  register struct Unit   *source_unit_ptr;
  register struct Site   *site_ptr,
                         *pred_site_ptr;
  register struct Unit   *unit_ptr;
  register FlagWord    flags;
  int   i;


  if ( (source_unit_ptr = kr_getUnitPtr( source_unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  UISourceUnitPtrFS = source_unit_ptr;

  for (i = MinUnitNo, unit_ptr = unit_array + MinUnitNo; i <= MaxUnitNo; i++, unit_ptr++)
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {     /*  unit is in use  */
      if ( (flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
        {       /*  unit has sites  */

        for (site_ptr = unit_ptr->sites, pred_site_ptr = NULL; site_ptr != NULL; pred_site_ptr = site_ptr, site_ptr = site_ptr->next)
          for (link_ptr = site_ptr->links, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
            if (link_ptr->to == source_unit_ptr)
              {
              UICurrentUnitPtr   = unit_ptr;
              UICurrentUnitNo  = i;
              UICurrentSuccUnitPtrFS = unit_ptr;
              UICurrentPredSitePtr   = pred_site_ptr;
              UICurrentPredSitePtrFS = pred_site_ptr;
              UICurrentSitePtr   = site_ptr;
              UICurrentSitePtrFS = site_ptr;
              UICurrentPredLinkPtr = pred_link_ptr;
              UICurrentLinkPtr     = link_ptr;

              *strength = link_ptr->weight;
              UICurrentLinkUnitNo = i;

              return( i );      /*  Und tschss..   */
              }
        }
      else
        {       /*  unit has no sites   */
        if ( (flags & UFLAG_INPUT_PAT) == UFLAG_DLINKS )
          {     /*  unit has direct links   */

          for (link_ptr = (struct Link *) unit_ptr->sites, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
            if (link_ptr->to == source_unit_ptr)
              {
              UICurrentUnitPtr   = unit_ptr;
              UICurrentUnitNo  = i;
              UICurrentSuccUnitPtrFS = unit_ptr;
              UICurrentPredSitePtr   = NULL;
              UICurrentPredSitePtrFS = NULL;
              UICurrentSitePtr   = NULL;
              UICurrentSitePtrFS = NULL;
              UICurrentPredLinkPtr = pred_link_ptr;
              UICurrentLinkPtr     = link_ptr;

              *strength = link_ptr->weight;
              UICurrentLinkUnitNo = i;

              return( i );      /*  Und tschss..   */
              }
          }
        }
      }
    }

  UICurrentLinkPtr = NULL;
  UISourceUnitPtrFS= NULL;

  return( 0 );
}


/*  Returns the no. of the next successor unit and the connection strenght.

IMPORTANT: If a successor unit exists, the current unit and site will be
           set to this successor unit and the attached site.

NOTE: This function is extremly ineffective  (Units are backward chained only)

REMARK: getNextSuccUnit is for downward compatibility only.
        In future versions, this function may not longer be
        supported !
*/
int  krui_getNextSuccUnit( strength )
FlintType  *strength;
{
  register struct Link   *link_ptr,
                         *pred_link_ptr;
  register struct Unit   *source_unit_ptr;
  register struct Site   *site_ptr,
                         *pred_site_ptr;
  register struct Unit   *unit_ptr;
  register FlagWord    flags;
  int   i;


  source_unit_ptr = UISourceUnitPtrFS;
  if (source_unit_ptr == NULL)  return( 0 );

  if( (UICurrentSuccUnitPtrFS->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
    {   /*  Continue search at last unit and next site  */
    i = UICurrentLinkUnitNo;
    unit_ptr = UICurrentSuccUnitPtrFS;

    for (site_ptr = UICurrentSitePtrFS->next, pred_site_ptr = UICurrentSitePtrFS; site_ptr != NULL; pred_site_ptr = site_ptr, site_ptr = site_ptr->next)
      for (link_ptr = site_ptr->links, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
        if (link_ptr->to == source_unit_ptr)
          {
          UICurrentUnitPtr   = unit_ptr;
          UICurrentUnitNo  = i;
          UICurrentPredSitePtr   = pred_site_ptr;
          UICurrentPredSitePtrFS = pred_site_ptr;
          UICurrentSitePtr   = site_ptr;
          UICurrentSitePtrFS = site_ptr;
          UICurrentPredLinkPtr = pred_link_ptr;
          UICurrentLinkPtr     = link_ptr;

          *strength = link_ptr->weight;

          return( i );      /*  Und tschss..   */
          }
    }

  /*  Start the unit search at next unit (unit has only direct links) */

  for (i = UICurrentLinkUnitNo + 1, unit_ptr = UICurrentSuccUnitPtrFS + 1; i <= MaxUnitNo; i++, unit_ptr++)
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {     /*  unit is in use  */

      if ( (flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
        {       /*  unit has sites  */
        for (site_ptr = unit_ptr->sites, pred_site_ptr = NULL; site_ptr != NULL; pred_site_ptr = site_ptr, site_ptr = site_ptr->next)
          for (link_ptr = site_ptr->links, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
            if (link_ptr->to == source_unit_ptr)
              {
              UICurrentUnitPtr   = unit_ptr;
              UICurrentUnitNo  = i;
              UICurrentSuccUnitPtrFS = unit_ptr;
              UICurrentPredSitePtr   = pred_site_ptr;
              UICurrentPredSitePtrFS = pred_site_ptr;
              UICurrentSitePtr   = site_ptr;
              UICurrentSitePtrFS = site_ptr;
              UICurrentPredLinkPtr = pred_link_ptr;
              UICurrentLinkPtr     = link_ptr;

              *strength = link_ptr->weight;
              UICurrentLinkUnitNo = i;

              return( i );      /*  Und tschss..   */
              }
        }
      else
        {       /*  unit has no sites   */
        if ( (flags & UFLAG_INPUT_PAT) == UFLAG_DLINKS )
          {     /*  unit has direct links   */

          for (link_ptr = (struct Link *) unit_ptr->sites, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
            if (link_ptr->to == source_unit_ptr)
              {
              UICurrentUnitPtr = unit_ptr;
              UICurrentUnitNo  = i;
              UICurrentSuccUnitPtrFS = unit_ptr;
              UICurrentPredSitePtr   = NULL;
              UICurrentPredSitePtrFS = NULL;
              UICurrentSitePtr   = NULL;
              UICurrentSitePtrFS = NULL;
              UICurrentPredLinkPtr = pred_link_ptr;
              UICurrentLinkPtr     = link_ptr;

              *strength = link_ptr->weight;
              UICurrentLinkUnitNo = i;

              return( i );      /*  Und tschss..   */
              }
          }
        }
      }
    }

  UICurrentLinkPtr = NULL;
  UISourceUnitPtrFS= NULL;
  return( 0 );
}





/*  Returns the no. of the current successor unit and the connection strenght.

IMPORTANT: If the successor unit exists, the current unit and site will be
           set to this successor unit and the attached site.

REMARK: getCurrentSuccUnit is for downward compatibility only.
        In future versions, this function may be not longer
        supported !
*/
int         krui_getCurrentSuccUnit( strength )
FlintType  *strength;
{
  if (UISourceUnitPtrFS == NULL)  return( 0 );

  UICurrentUnitPtr     = UICurrentSuccUnitPtrFS;
  UICurrentPredSitePtr = UICurrentPredSitePtrFS;
  UICurrentSitePtr     = UICurrentSitePtrFS;

  *strength = UICurrentLinkPtr->weight;

  return( UICurrentLinkUnitNo );
}


/*  True if there exists a connection between source unit <source_unit_no>
    and target unit <target_unit_no>, otherwise false. If there exist a
    connection between these units, krui_areConnected returns the connection
    strength also.
    Returns FALSE if unit doesn't exist.

IMPORTANT: If there exist a connection, the current unit and site will be
           set to the target unit and the attached site.

NOTE: This function is slow (Units are backward chained only).

REMARK: areConnected is for downward compatibility only.
        Do not use this function in future applications.
*/
bool    krui_areConnected( source_unit_no, target_unit_no, weight)
int     source_unit_no, target_unit_no;
FlintType   *weight;
{
  register struct Link   *link_ptr,
                         *pred_link_ptr;
  register struct Unit   *source_unit_ptr;
  register struct Site   *site_ptr,
                         *pred_site_ptr;
  register struct Unit   *unit_ptr;


  if ( (source_unit_ptr = kr_getUnitPtr( source_unit_no ) ) == NULL)
    return( FALSE );    /*  invalid unit #  */

  if ( (unit_ptr = kr_getUnitPtr( target_unit_no ) ) == NULL)
    return( FALSE );    /*  invalid unit #  */


  if ( (unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_SITES )
    {       /*  unit has sites  */

    for (site_ptr = unit_ptr->sites, pred_site_ptr = NULL; site_ptr != NULL; pred_site_ptr = site_ptr, site_ptr = site_ptr->next)
      for (link_ptr = site_ptr->links, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
        if (link_ptr->to == source_unit_ptr)
          {
          UICurrentUnitPtr   = unit_ptr;
          UICurrentUnitNo  = target_unit_no;
          UICurrentSuccUnitPtrFS = unit_ptr;
          UICurrentPredSitePtr   = pred_site_ptr;
          UICurrentPredSitePtrFS = pred_site_ptr;
          UICurrentSitePtr   = site_ptr;
          UICurrentSitePtrFS = site_ptr;
          UICurrentPredLinkPtr = pred_link_ptr;
          UICurrentLinkPtr     = link_ptr;

          *weight = link_ptr->weight;
          UICurrentLinkUnitNo = target_unit_no;

          return( TRUE );      /*  Und tschss..   */
          }
    }
  else
    {       /*  unit has no sites   */
    if ( (unit_ptr->flags & UFLAG_INPUT_PAT) == UFLAG_DLINKS )
      {     /*  unit has direct links   */

      for (link_ptr = (struct Link *) unit_ptr->sites, pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
        if (link_ptr->to == source_unit_ptr)
          {
          UICurrentUnitPtr   = unit_ptr;
          UICurrentUnitNo  = target_unit_no;
          UICurrentSuccUnitPtrFS = unit_ptr;
          UICurrentPredSitePtr   = NULL;
          UICurrentPredSitePtrFS = NULL;
          UICurrentSitePtr   = NULL;
          UICurrentSitePtrFS = NULL;
          UICurrentPredLinkPtr = pred_link_ptr;
          UICurrentLinkPtr     = link_ptr;

          *weight = link_ptr->weight;
          UICurrentLinkUnitNo = target_unit_no;

          return( TRUE );      /*  Und tschss..   */
          }
      }
    }

  UICurrentLinkPtr = NULL;
  UISourceUnitPtrFS= NULL;

  return( FALSE );
}


/*  True if there exists a connection between source unit <source_unit_no>
    and the current unit/site, otherwise false.

NOTE: If there exists a connection between the two units, the current link is set
      to the link between the two units. (alter the link weight with krui_setLinkWeight)
*/
bool        krui_isConnected( source_unit_no )
int  source_unit_no;
{
  FlagWord    flags;
  struct Unit   *source_unit_ptr;
  struct Link   *link_ptr,
                *pred_link_ptr;


  source_unit_ptr = unit_array + source_unit_no;

  if ( (source_unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
    /*  Source unit is in use    */
    {
    flags = UICurrentUnitPtr->flags & UFLAG_INPUT_PAT;

    if (flags != UFLAG_NO_INP)
      /*  Unit has inputs  */
      {

      if ( flags == UFLAG_SITES )
        /*  Unit has sites  */
        link_ptr = UICurrentSitePtr->links;
      else
        link_ptr = (struct Link *) UICurrentUnitPtr->sites;

      for (pred_link_ptr = NULL; link_ptr != NULL; pred_link_ptr = link_ptr, link_ptr = link_ptr->next)
        if (link_ptr->to == source_unit_ptr)
          {
          UICurrentPredLinkPtr = pred_link_ptr;
          UICurrentLinkPtr     = link_ptr;

          return( TRUE );
          }
      }
    }

  /*    There isn't a connection    */
  UICurrentLinkPtr = NULL;
  return( FALSE );
}



/*  Returns the link weight of the current link.
*/
FlintType  krui_getLinkWeight()
{
  if (UICurrentLinkPtr != NULL)
    return( UICurrentLinkPtr->weight );
  else
    return( (FlintType) 0);
}



/*  Sets the link weight of the current link
*/
void  krui_setLinkWeight( strength )
FlintTypeParam   strength;
{
  if (UICurrentLinkPtr != NULL)
    UICurrentLinkPtr->weight = (FlintType) strength;
}



/*  Creates a link between source unit and the current unit/site.
    Returns an error code:
     - if memory allocation fails
     - if source unit doesn't exist or
     - if there exists already a connection between current unit/site and
       the source unit
    0 otherwise.
    krui_createLink DO NOT set the current link.

NOTE: If you want to create a link and its unknown if there exists already a
      connection between the two units, use krui_createLink and test the return
      code, instead of the sequence krui_isConnected and krui_createLink.
*/
krui_err  krui_createLink( source_unit_no, strength )
int   source_unit_no;
FlintTypeParam  strength;
{
  register struct Link	 *link_ptr;
  register struct Unit	 *source_unit_ptr;
  struct Link	*first_link_ptr;
  FlagWord    flags;


  source_unit_ptr = unit_array + source_unit_no;

  if ( (source_unit_ptr->flags & UFLAG_IN_USE) != UFLAG_IN_USE)
    /*	Source unit isn't in use  */
    return( KRERR_UNIT_NO );  /*  Invalid unit number  */

  flags = UICurrentUnitPtr->flags & UFLAG_INPUT_PAT;

  switch (flags)
    {
    case  UFLAG_NO_INP:    /*  Unit had no inputs  */
      if ( (link_ptr = krm_getLink() ) == NULL)
	return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

      link_ptr->to     = source_unit_ptr;
      link_ptr->weight = (FlintType) strength;
      link_ptr->next   = NULL;

      UICurrentUnitPtr->sites = (struct Site *) link_ptr;
      UICurrentUnitPtr->flags |= UFLAG_DLINKS;

      break;

    case  UFLAG_SITES:	  /*  Unit has sites  */
      first_link_ptr = UICurrentSitePtr->links;

      for (link_ptr = first_link_ptr; link_ptr != NULL; link_ptr = link_ptr->next)
	if (link_ptr->to == source_unit_ptr)
	  /*  There exists already a connection   */
	  return( KRERR_ALREADY_CONNECTED );

      if ( (link_ptr = krm_getLink() ) == NULL)
	return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

      link_ptr->to     = source_unit_ptr;
      link_ptr->weight = strength;
      link_ptr->next   = first_link_ptr;

      UICurrentSitePtr->links = link_ptr;

      break;

    case  UFLAG_DLINKS:    /*  Unit has direct links   */
      first_link_ptr = (struct Link *) UICurrentUnitPtr->sites;

      for (link_ptr = first_link_ptr; link_ptr != NULL; link_ptr = link_ptr->next)
	if (link_ptr->to == source_unit_ptr)
	  /*  There exists already a connection   */
	  return( KRERR_ALREADY_CONNECTED );

      if ( (link_ptr = krm_getLink() ) == NULL)
	return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

      link_ptr->to     = source_unit_ptr;
      link_ptr->weight = strength;
      link_ptr->next   = first_link_ptr;

      UICurrentUnitPtr->sites = (struct Site *) link_ptr;

      break;

    default:
      break;
  }

  NetModified = TRUE;
  return( KRERR_NO_ERROR );
}





/*  Deletes the current link.

NOTE: To delete a link between the current unit/site and the source unit
      <source_unit_no>, call krui_isConnected( source_unit_no ) and
      krui_deleteLink().
*/
void  krui_deleteLink()
{
  struct Link   *next_link_ptr;


  if ( (UICurrentLinkPtr != NULL) &&
       ((UICurrentUnitPtr->flags & UFLAG_INPUT_PAT) != UFLAG_NO_INP) )
    {   /*  Unit has inputs */

    next_link_ptr = UICurrentLinkPtr->next;
    krm_releaseLink( UICurrentLinkPtr );
    UICurrentLinkPtr = next_link_ptr;

    if (UICurrentPredLinkPtr != NULL)
      /*  Current link is not the first link at the unit  */
      UICurrentPredLinkPtr->next = next_link_ptr;     /*  connect next link   */
    else
      {     /*  Current link is the first link at the unit/site */
      if (UICurrentSitePtr == NULL)
        /*  Unit has no sites   */
        {   /*  Current link is the first link at the unit  */
        UICurrentUnitPtr->sites = (struct Site *) next_link_ptr;

        if (next_link_ptr == NULL)
          /*  The unit has no inputs now  */
          UICurrentUnitPtr->flags &= (~UFLAG_INPUT_PAT);
        }
      else
        /*  Unit has sites  */
        UICurrentSitePtr->links = next_link_ptr;

      }
    }

  NetModified = TRUE;
}




/*  Deletes all input links at current unit/site.
*/
void        krui_deleteAllInputLinks()
{
  if (UICurrentUnitPtr == NULL)
    return;  /*  no current unit defined  */

  if (UICurrentSitePtr != NULL)
    {  /*  Current unit has sites  */
    krm_releaseAllLinks( UICurrentSitePtr->links );
    UICurrentSitePtr->links = NULL;  /*  Site has no inputs now   */
    }
  else
    /*  Current unit don't has sites  */
    if ( (UICurrentUnitPtr->sites != NULL) )
      {  /*  Unit has input links */
      krm_releaseAllLinks( (struct Link *) UICurrentUnitPtr->sites );
      UICurrentUnitPtr->sites = NULL;  /*  Unit don't has inputs now   */
      UICurrentUnitPtr->flags &= (~UFLAG_INPUT_PAT);
      }

  UICurrentLinkPtr = NULL;
  NetModified = TRUE;
}



/*  Deletes all output links at current unit.
    NOTE: This function is slow.
*/
void  krui_deleteAllOutputLinks()
{
  kr_deleteAllOutputLinks( UICurrentUnitPtr );

  NetModified = TRUE;
  UICurrentLinkPtr = NULL;
/*
  UICurrentSitePtr = NULL;
*/
}


/*  Add uniform distributed random values to connection weights.
    <minus> must be less then <plus>.
*/
void  krui_jogWeights( minus, plus)
FlintTypeParam   minus, plus;
{
  if (minus >= plus)
    return;  /*  wrong bounds    */

  kr_jogWeights( minus, plus );
}



/*#################################################

GROUP: Functions for network propagation

#################################################*/

/*  Updates a single unit. Returns error code if unit doesn't exist,
    0 otherwise.

    NOTE: Updates also frozen Units.
*/
krui_err    krui_updateSingleUnit( unit_no )
int     unit_no;
{
  register struct Unit   *unit_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no )) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if (unit_ptr->out_func == NULL)
    /*  Identity Function   */
    unit_ptr->Out.output = unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);
  else
    {
    unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);
    unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
    }

  return( KRERR_NO_ERROR );
}


/*  Returns the current update function. The default update function is
    SerialOrder() (see also kr_def.h).
*/
char   *krui_getUpdateFunc()
{
  return( krf_getFuncName( (FunctionPtr) kr_getUpdateFunc() ) );
}


/*  Changes the current update function.  Returns error code if update
    function is invalid.
*/
krui_err   krui_setUpdateFunc( update_func )
char  *update_func;
{
  FunctionPtr func_ptr;


  if ( !krf_funcSearch( update_func, UPDATE_FUNC, &func_ptr ) )
    return( KRERR_UPDATE_FUNC );  /*  Invalid update function  */

  kr_setUpdateFunc( (UpdateFuncPtr) func_ptr );
  return( KRERR_NO_ERROR );
}


/*  Updates the network according to update function:
NOTE: The network should be feedforward in topological mode,
      otherwise function will return a warning message.

      See also krui_setSeedNo for initializing the pseudo
      random generator.

Returns error code if an error occured, 0 othrwise.
*/
krui_err    krui_updateNet( parameterArray, NoOfParams  )
float   parameterArray[];
int  NoOfParams;
{
  int  ret_code;


  if (NoOfUnits <= 0)
    return( KRERR_NO_UNITS );  /*  No Units defined  */

  ret_code = kr_updateNet( parameterArray, NoOfParams );
  
  return( ret_code );
}


/*  Calculates the network error whith the given pattern. 
    Uses the current update function to propagate the network.

    parameterArray contains the parameters of the update function.
    NoOfParams contains the number of parameters of the update function.

    parameterInArray[0] contains the max. devitation.
    Set NoOfInParameters to 1.

    parameterOutArray[0] contains the error of the network/pattern,
    parameterOutArray[0] contains the number of output units with a higher error value than 
    the given max. devitation.
    NoOfOutParams will be set to 2.

NOTE: Patterns must be loaded before calling this function.

Returns error code if an error occured, 0 otherwise.
*/
krui_err    krui_testNet( pattern_no, 
                          parameterArray, NoOfParams,
                          parameterInArray, NoOfInParams, 
                          parameterOutArray, NoOfOutParams )
int  pattern_no;
float   parameterArray[], parameterInArray[], * *parameterOutArray;
int  NoOfParams, NoOfInParams, *NoOfOutParams;
{
  int  ret_code;


  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );  /*  No Units defined    */
  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );  /*  no patterns defined  */
  if (!PatternsValid)
    return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */

  if ((pattern_no < 1) || (pattern_no > NoOfPatternPairs) )
    return( KRERR_PATTERN_NO );  /*  Invalid pattern number  */

  kr_showPattern( pattern_no - 1, OUTPUT_NOTHING );

  ret_code = kr_updateNet( parameterArray, NoOfParams );
  if (ret_code != 0)  return( ret_code );

  ret_code = kr_checkUnitsOutput( pattern_no - 1,
				  parameterInArray, NoOfInParams,
				  parameterOutArray, NoOfOutParams );

  return( ret_code );
}




/*#################################################

GROUP: Initialisation Functions

#################################################*/


/*  Returns the current initialisation function. The default initialisation function is
    RandomizeWeights (see also kr_def.h).
*/
char   *krui_getInitialisationFunc()
{
  return( krf_getFuncName( (FunctionPtr) kr_getInitialisationFunc() ) );
}


/*  Changes the current learning function.  Returns error code if learning
    function is invalid.
*/
krui_err   krui_setInitialisationFunc( initialisation_func )
char  *initialisation_func;
{
  FunctionPtr   func_ptr;


  if ( !krf_funcSearch( initialisation_func, INIT_FUNC, &func_ptr ) )
    return( KRERR_INIT_FUNC );  /*  Invalid initialisation function  */

  kr_setInitialisationFunc( (InitFuncPtr) func_ptr );
  return( KRERR_NO_ERROR );
}

/*  Initializes the network
*/
krui_err   krui_initializeNet( parameterArray, NoOfParams )
float   parameterArray[];
int  NoOfParams;
{
  return( kr_initializeNet( parameterArray, NoOfParams ) );
}




/*#################################################

GROUP: Learning Functions

#################################################*/


/*  Returns the current learning function. The default learning function is
    Backpropagation (see also kr_def.h).
*/
char   *krui_getLearnFunc()
{
  return( krf_getFuncName( (FunctionPtr) kr_getLearnFunc() ) );
}


/*  Changes the current learning function.  Returns error code if learning
    function is invalid.
*/
krui_err   krui_setLearnFunc( learning_func )
char  *learning_func;
{
  FunctionPtr   func_ptr;


  if ( !krf_funcSearch( learning_func, LEARN_FUNC, &func_ptr ) )
    return( KRERR_LEARNING_FUNC );  /*  Invalid learning function  */

  kr_setLearnFunc( (LearnFuncPtr) func_ptr );
  return( KRERR_NO_ERROR );
}



/*  Learn all pattern pairs using current learning method.
    parameterInArray contains the learning
    parameter(s). NoOfInParams stores the number of learning parameters.
    parameterOutArray returns the results from the learning function.
    this array is a static array defined in the learning function.
    *NoOfOutParams points to a integer value that contains the number
    of output parameters from the current learning function.
    Returns an error code if memory allocation has failed or if
    the parameters are invalid.

    Returns error code of the learning function.
REMEMBER:  The backpropagation learning function takes the learning
           parameter from parameterInArray[ 0 ]. parameterOutArray[ 0 ]
           returns the current net error.

NOTE: Patterns must be loaded before calling this function.
*/
krui_err   krui_learnAllPatterns( parameterInArray, NoOfInParams, parameterOutArray, NoOfOutParams )
/*  REMEMBER:  parameterOutArray[ 0 ] returns the current net error
               parameterInArray[ 0 ] contains the 1st learning parameter
*/
float   parameterInArray[], * *parameterOutArray;
int  NoOfInParams, *NoOfOutParams;
{
  int  ret_code;


  if (NoOfPatternPairs == 0)  return( KRERR_NO_PATTERNS );  /*  no patterns defined  */
  if (!PatternsValid)  return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */

  /*  learn all patterns  */
  ret_code = kr_learnNet( 0, NoOfPatternPairs - 1, 
                          parameterInArray, NoOfInParams, 
                          parameterOutArray, NoOfOutParams );
  
  return( ret_code );
}



/*  Same as krui_learnAllPatterns( ... ) but learns only the current
    pattern pair.

NOTE: Patterns must be loaded before calling this function.
*/
krui_err   krui_learnSinglePattern( pattern_no, parameterInArray, NoOfInParams, parameterOutArray, NoOfOutParams)
int     pattern_no;
float   parameterInArray[],
        * *parameterOutArray;
int     NoOfInParams,
        *NoOfOutParams;
{
  int  ret_code;


  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );      /*  no patterns defined  */
  if (!PatternsValid)
    return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */
  if ((pattern_no < 1) && (pattern_no > NoOfPatternPairs) )
    return( KRERR_PATTERN_NO );       /*  Invalid pattern number  */

  ret_code = kr_learnNet( pattern_no - 1, pattern_no - 1, 
                          parameterInArray, NoOfInParams, 
                          parameterOutArray, NoOfOutParams );

  return( ret_code );
}



/*#################################################

GROUP: Pattern Management

#################################################*/

/*  Sets the current pattern.
    Returns a error code if pattern number is invalid.

NOTE: Patterns must be loaded before calling this function.
*/
krui_err  krui_setPatternNo( pattern_no )
int  pattern_no;
{
  if (kr_setCurrentPattern( pattern_no  ) == 0)
    return( KRERR_NO_ERROR );
  else 
    return( KRERR_PATTERN_NO );  /*  Invalid pattern number  */
}


/*  delete the current pattern pair
*/
krui_err  krui_deletePattern()
{
  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );  /*  No Units defined    */
  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );  /*  no patterns defined  */
  if (!PatternsValid)
    return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */

  switch (kr_deletePattern())
    {
    case -1:  return( KRERR_CRITICAL_MALLOC );   /*  Memory allocation failed during critical operation  */
    case -2:  return( KRERR_PATTERN_NO );  /*  Invalid pattern number  */
    }

  return( KRERR_NO_ERROR );
}

/*  modify the current pattern pair
*/
krui_err  krui_modifyPattern()
{
  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );  /*  No Units defined    */
  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );  /*  no patterns defined  */
  if (!PatternsValid)
    return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */

  if (kr_modifyPattern() != 0)
    return( KRERR_PATTERN_NO );  /*  Invalid pattern number  */
  else
    return( KRERR_NO_ERROR );
}


/*  According to the mode krui_showPattern stores the current
    Pattern into the units activation (and/or output) values.
    The modes are:
      - OUTPUT_NOTHING
         store input pattern into input units activations
      - OUTPUT_ACT
         store input pattern into input units activations and
         store output pattern into output units activations
      - OUTPUT_OUT
         store input pattern into input units activations,
         store output pattern into output units activations and
         update output units output

NOTE: See include file glob_typ.h for mode constants.
*/
krui_err    krui_showPattern( mode )
int  mode;
{
  int  pattern_no;


  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );  /*  No Units defined    */
  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );  /*  no patterns defined  */
  if (!PatternsValid)
    return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */

  pattern_no = kr_getCurrentPattern();
  if ((pattern_no < 1) || (pattern_no > NoOfPatternPairs) )
    return( KRERR_PATTERN_NO );  /*  Invalid pattern number  */

  kr_showPattern( pattern_no - 1, mode );

  return( KRERR_NO_ERROR );
}




/*  Creates a new pattern pair.
    A pattern pair can be created by modifying the activation
    value of the input/output units.
    Returns error code if memory is insufficent or no. of
    input/output units is incompatible, 0 otherwise.

NOTE: krui_newPattern switches pattern shuffeling off.
      For shuffeling the new pattern pairs call
        krui_newPattern(...)
        krui_shufflePattern( TRUE )
*/
krui_err    krui_newPattern()
{
  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );  /*  No Units defined    */

  if (NoOfPatternPairs > 0)
    {
    if (NoOfInputUnits != kr_getNoOfUnits( INPUT ) )
      return( KRERR_CHANGED_I_UNITS );   /*  No. of input units have changed   */
    if (NoOfOutputUnits != kr_getNoOfUnits( OUTPUT ) )
      return( KRERR_CHANGED_O_UNITS );   /*  No. of output units have changed  */

    PatternsValid = TRUE;  /*  Patterns are valid now  */
    }
  else
    {
    NoOfInputUnits = kr_getNoOfUnits( INPUT );
    NoOfOutputUnits = kr_getNoOfUnits( OUTPUT );
    if (NoOfInputUnits == 0)
      return( KRERR_NO_INPUT_UNITS );  /*  No input units defined  */
    if (NoOfOutputUnits == 0)
      return( KRERR_NO_OUTPUT_UNITS );  /*  No output units defined  */

    PatternsValid = TRUE;  /*  Patterns are valid now  */
    }

  if (kr_createPattern() != 0)
    return( KRERR_INSUFFICIENT_MEM );  /*  Insufficient memory  */

  return( KRERR_NO_ERROR );
}


/*  Release previously defined patterns from memory.
    Call krui_releasePatterns() if you want to create totally
    new patterns with krui_newPattern().
*/
void        krui_deleteAllPatterns()
{
  krm_releasePatternArrays();  /*  Free pattern arrays	*/
  krm_releasePatternNumbers();	 /*  Free pattern permutation array  */
  PatternsValid = FALSE;   /*  Patterns are deleted and not valid  */
}



/*  Shuffle pattern pairs by using pseudo random generator.
    Returns error code if memory allocation fails.
    Shuffeling of patterns is used by krui_learnAllPatterns(...).
    krui_shufflePatterns( TRUE ) switches shuffeling of patterns
    on, krui_shufflePatterns( FALSE ) switches shuffeling of
    patterns off.
    The default presetting is krui_shufflePatterns( FALSE ).

NOTE:  See also krui_setSeedNo( seed )
*/
krui_err    krui_shufflePatterns( on_or_off )
bool    on_or_off;
{
  if (!on_or_off)
    {
    /*	Free pattern permutation array	*/
    krm_releasePatternNumbers();
    return( KRERR_NO_ERROR );
    }

  if (NoOfPatternPairs == 0)
    return( KRERR_NO_PATTERNS );  /*  no patterns defined  */
  if (!PatternsValid)
    return( KRERR_INCORE_PATTERNS );  /*  (In-Core) patters not valid  */

  if (on_or_off)
    {  /*  shuffle pattern numbers  */
    if (kr_shufflePatterns() != 0)
      return( KRERR_INSUFFICIENT_MEM );   /*  Insufficient memory  */
    }

 return( KRERR_NO_ERROR );
}




/*#################################################

GROUP: I/O Functions

#################################################*/

/*  Save a network.
    If netname is a NULL pointer, the net will get the name "UNTITLED"
    Returns error code if an error occured, or 0 otherwise.
*/
krui_err  krui_saveNet( filename, netname)
char    *filename,
        *netname;
{
  return( krio_saveNet( filename, netname) );
}


/*  Load a network file and create a network.
    Returns the name of the net or "UNTITLED" if unknown.
    Returns error code if an error occured during
    loading/memory allocation, or 0 otherwise.
*/
krui_err  krui_loadNet( filename, netname )
char    *filename,
        **netname;
{
  int   err;
  char	*netfile_version;  /*  isn't used  */
              

  if ((err = krio_loadNet( filename, netname, &netfile_version )) != KRERR_NO_ERROR)
    return( err );

  kr_updateUnitOutputs();
  if (NoOfPatternPairs > 0)
    /*	In-Core patterns exist	*/
    if ( (NoOfInputUnits != kr_getNoOfUnits( INPUT )) ||
	 (NoOfOutputUnits != kr_getNoOfUnits( OUTPUT )) )
      {
      PatternsValid = FALSE;
      return( KRERR_INCORE_PATTERNS );	/*  In-Core patterns incompatible with current network	*/
      }

  return( KRERR_NO_ERROR );
}


/*  Load file containing pattern pairs.
    Returns error code if an error occured during
    loading/memory allocation, or 0 otherwise.

NOTE: krui_loadPatterns tries to store all patterns into RAM.
      If the available memory is insufficient, at least one pattern
      pair will be buffered during disk access.
*/
krui_err    krui_loadPatterns( filename )
char   *filename;
{
  return( krio_loadPatterns( filename ) );
}

/*  Save all pattern pairs.
    Returns error code if an error occured during file I/O,
    0 otherwise.
*/
krui_err    krui_savePatterns( filename)
char  *filename;
{
  return( krio_savePatterns( filename ) );
}




/*#################################################

GROUP: Functions for memory management

#################################################*/

/*  Allocates a given number of units, additional units
    may allocated by calling this function again.
    This function is called automatically if the user
    construct more units than have been allocated before, but
    it is recommended to use this function if a large amount
    of units is needed (the UNIX System can manage system resources
    much better, if the amount of memory used for the network is
    approximately known before the network is created).

    Returns error code if memory allocation fails, 0 otherwise.

NOTE: If krui_create_unit has been called before using this
      function, at least <UNIT_BLOCK> numbers of units have been
      allocated.
      See "kr_def.h" for more information about memory allocation
      block sizes.
*/
int         krui_allocateUnits( no_of_units )
int     no_of_units;
{
  return( krm_allocUnits( no_of_units ) );
}



/*  Delete network, names and unit types.
    Frees all memory used for the network.

NOTE: If krui_loadNet is called more then once,  krui_deleteNet will
      be called by krui_loadNet, because the kernel have to free the
      memory used for the old network.
      It is recommended (but not neccessary) to call krui_deleteNet
      before terminating the SNNS-Kernel.
*/
void        krui_deleteNet()
{
  krm_releaseMem();
}


/*#################################################

GROUP: Functions for reading/searching the symbol table

#################################################*/

/*  Returns the first symbol/symbol type entry in the
    symbol table. Returns true if this entry is available,
    false otherwise.
*/
bool    krui_getFirstSymbolTableEntry( symbol_name, symbol_type )
char * *symbol_name;
int  *symbol_type;
{
  struct NameTable      *n_tbl;


  n_tbl = krm_getNTableFirstEntry();

  while ( (n_tbl != NULL) && ((int) n_tbl->sym_type == UNUSED_SYM) )
    n_tbl = krm_getNTableNextEntry();

  if (n_tbl == NULL)
    {
    *symbol_name = NULL;
    *symbol_type = 0;
    return( FALSE );
    }

  *symbol_name = n_tbl->Entry.symbol;
  *symbol_type = (int) n_tbl->sym_type;
  return( TRUE );
}



/*  Returns the next symbol/symbol type entry in the
    symbol table. Returns true if another entry is available,
    false otherwise.
*/
bool    krui_getNextSymbolTableEntry( symbol_name, symbol_type )
char * *symbol_name;
int  *symbol_type;
{
  struct NameTable      *n_tbl;


  n_tbl = krm_getNTableNextEntry();

  while ( (n_tbl != NULL) && ((int) n_tbl->sym_type == UNUSED_SYM) )
    n_tbl = krm_getNTableNextEntry();

  if (n_tbl == NULL)
    {
    *symbol_name = NULL;
    *symbol_type = 0;
    return( FALSE );
    }

  *symbol_name = n_tbl->Entry.symbol;
  *symbol_type = (int) n_tbl->sym_type;
  return( TRUE );
}





/*  Searches the symbol table for a given symbol and
    symbol type (unit name symbol, site name symbol,
    functionality unit name symbol)
    Returns true, if the symbol exists.
*/
bool    krui_symbolSearch( symbol, symbol_type)
char    *symbol;
int     symbol_type;
{
  return( krm_NTableSymbolSearch( symbol, symbol_type ) != NULL );
}






/*#################################################

GROUP: Miscellanous

#################################################*/


/*  Returns the current Version of the SNNS-Kernel.
*/
char    *krui_getVersion()
{
  static char	snns_version[128];


  strcpy( snns_version, SNNS_VERSION );
  strcat( snns_version, KERNEL_PATCH );
  return( snns_version );
}


/*  Returns miscellanous information about the current network.
*/
void    krui_getNetInfo( no_of_sites, no_of_links, no_of_STable_entries,
                         no_of_FTable_entries )
int     *no_of_sites,
        *no_of_links,
        *no_of_STable_entries,
        *no_of_FTable_entries;
{
  int   array_size,
        info_array[ 10 ];


  krm_getMemoryManagerInfo( &array_size, info_array );

  *no_of_sites          = info_array[ 0 ];
  *no_of_links          = info_array[ 1 ];
  *no_of_STable_entries = info_array[ 2 ];
  *no_of_FTable_entries = info_array[ 3 ];
}


/*  Returns the number of ALLOCATED (not the number of USED) bytes
    per entry.
*/
void  krui_getMemoryManagerInfo( unit_bytes, site_bytes, link_bytes,
                                 NTable_bytes, STable_bytes, FTable_bytes )
int     *unit_bytes,
        *site_bytes,
        *link_bytes,
        *NTable_bytes,
        *STable_bytes,
        *FTable_bytes;
{
  int   array_size,
        info_array[ 10 ];


  krm_getMemoryManagerInfo( &array_size, info_array );

  *unit_bytes   = info_array[ 4 ] * UNIT_SIZE;
  *site_bytes   = info_array[ 5 ] * SITE_SIZE;
  *link_bytes   = info_array[ 6 ] * LINK_SIZE;
  *NTable_bytes = info_array[ 7 ] * NTABLE_SIZE;
  *STable_bytes = info_array[ 8 ] * STABLE_SIZE;
  *FTable_bytes = info_array[ 9 ] * FTYPE_UNIT_SIZE;
}


/*  Returns Information about the unit default settings.
    For more information about default settings see krui_createDefaultUnit() and
    krui_createFTypeUnit( .. ).
*/

void    krui_getUnitDefaults( act, bias, st, subnet_no, layer_no,
                              act_func, out_func )
FlintType   *act, *bias;
int         *st;
int         *subnet_no, *layer_no;
char        * *act_func, * *out_func;
{
  int	ttflags;
  ActFuncPtr  act_func_ptr;
  OutFuncPtr  out_func_ptr;


  kr_getUnitDefaults( act, bias, &ttflags, subnet_no, layer_no,
		      &act_func_ptr, &out_func_ptr );

  *st	     = kr_getTType( ttflags );
  *act_func  = krf_getFuncName( (FunctionPtr) act_func_ptr );
  *out_func  = krf_getFuncName( (FunctionPtr) out_func_ptr );
}


/*  Changes the unit default settings.
    For more information about default settings see krui_createDefaultUnit() and
    krui_createFTypeUnit( .. ).

    Returns error code if
     - activation/output function is invalid
     - Topologic type is invalid
    0 otherwise.
*/

krui_err    krui_setUnitDefaults( act, bias, st, subnet_no, layer_no,
                                  act_func, out_func )
FlintTypeParam  act, bias;
int         st;
int         subnet_no, layer_no;
char        *act_func, *out_func;
{
  int	       ttflags;
  FunctionPtr  func_ptr;
  OutFuncPtr   out_func_ptr;
  ActFuncPtr   act_func_ptr;
  ActDerivFuncPtr act_deriv_func_ptr;


  if ( !krf_funcSearch( out_func, OUT_FUNC, &func_ptr ) )
    return( KRERR_OUTFUNC );  /*  Invalid unit output function  */
  out_func_ptr = (OutFuncPtr) func_ptr;

  if ( !krf_funcSearch( act_func, ACT_FUNC, &func_ptr ) )
    return( KRERR_ACTFUNC );  /*  Invalid unit activation function  */
  act_func_ptr = (ActFuncPtr) func_ptr;

  if ( !krf_funcSearch( act_func, ACT_DERIV_FUNC, &func_ptr ))
    return( KRERR_DERIV_FUNC );    /*  Derivation function of the activation function doesn't exist  */
  act_deriv_func_ptr = (ActDerivFuncPtr) func_ptr;


  if ( (ttflags = kr_setTType( st )) == -1)
    return( KRERR_TTYPE );  /*  Topologic type invalid  */

  kr_setUnitDefaults( act, bias, ttflags, subnet_no, layer_no,
		      act_func_ptr, act_deriv_func_ptr, out_func_ptr );

  return( KRERR_NO_ERROR );
}



/*  Reset the network by changeing the unit activation
    to the initial activation value.
*/
void  krui_resetNet()
{
  register int   i;
  register struct Unit   *unit_ptr;


  if ( (unit_array == NULL) || (NoOfUnits == 0) )
    return;

  for (i = MinUnitNo, unit_ptr = unit_array + MinUnitNo; i <= MaxUnitNo; i++, unit_ptr++)
    if ( (unit_ptr->flags & UFLAG_INITIALIZED) == UFLAG_INITIALIZED)
      /*  unit is in use  */
      unit_ptr->act = unit_ptr->i_act;

}



/*  Initialize the pseudo random generator.
    0 as argument reinitializes the random generator.
*/
void  krui_setSeedNo( seed )
long  seed;
{
  if (seed != 0)
    srand48( seed );   /*  initialize random generator with seed value	*/
  else
    srand48( (long) time( (long *) 0) );  /*  initialize random generator random value	*/
}


/*  Returns the no. of input units
*/
int  krui_getNoOfInputUnits()
{
  return( kr_getNoOfUnits( INPUT ) );
}

/*  returns the no. of output units
*/
int  krui_getNoOfOutputUnits()
{
  return( kr_getNoOfUnits( OUTPUT ) );
}

/*  returns the no. of units of the specified topologic type
    (i.e. Input, Hidden, Output or Special units)
*/
int  krui_getNoOfTTypeUnits( UnitTType )
int  UnitTType;
{
  return( kr_getNoOfUnits( UnitTType ) );
}



/*  Returns the no. of pattern pairs or 0 if no patterns have
    been loaded.
*/
int         krui_getNoOfPatterns()
{
  return( NoOfPatternPairs );
}



#ifdef KERNEL3D

/*  returns the base address of the XY-Translation table
*/
krui_err  krui_getXYTransTable( xy_trans_tbl_ptr )
struct TransTable  * *xy_trans_tbl_ptr;
{
  *xy_trans_tbl_ptr = krio_getXYTransTablePtr();

  return( KRERR_NO_ERROR );
}

/*  Returns the 3D transform center of the specified unit and center number
*/
krui_err  krui_getUnitCenters( unit_no, center_no, unit_center )
int  unit_no, center_no;
struct PositionVector  * *unit_center;
{
  struct Unit   *unit_ptr;

  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if ((center_no < 0) || (center_no >= NO_OF_UNIT_CENTER_POS))
    return( KRERR_PARAMETERS );

  *unit_center = unit_ptr->unit_center_pos + center_no;
  return( KRERR_NO_ERROR );
}


/*  Sets the 3D transform center of the specified unit and center number
*/
krui_err  krui_setUnitCenters( unit_no, center_no, unit_center )
int  unit_no, center_no;
struct PositionVector  *unit_center;
{ 
  struct Unit   *unit_ptr;
  struct PositionVector    *unit_posvec_ptr;


  if ( (unit_ptr = kr_getUnitPtr( unit_no ) ) == NULL)
    return( KRERR_UNIT_NO );    /*  invalid unit #  */

  if ((center_no < 0) || (center_no >= NO_OF_UNIT_CENTER_POS))
    return( KRERR_PARAMETERS );

  unit_posvec_ptr = unit_ptr->unit_center_pos + center_no;

  memcpy( (char *) unit_posvec_ptr, (char *) unit_center, sizeof( struct PositionVector ) );

  return( KRERR_NO_ERROR );
}


#else


krui_err  krui_getXYTransTable( dummy )
int dummy;
{
  return( KRERR_NOT_3D );
}


krui_err  krui_getUnitCenters( unit_no, center_no, dummy )
int  unit_no, center_no, dummy;
{ 
  return( KRERR_NOT_3D );
}


krui_err  krui_setUnitCenters( unit_no, center_no, dummy )
int  unit_no, center_no, dummy;
{ 
  return( KRERR_NOT_3D );
}

#endif



/*  Returns an error message, depending on the error code.
    If a function returns an error code use krui_error to
    get the message.
*/
char *      krui_error( error_code )
int     error_code;
{
  static char	*err_message[] = {
      "Invalid error code",
      "Insufficient memory",
      "Invalid unit number",
      "Invalid unit output function",
      "Invalid unit activation function",
      "Invalid site function",
      "Creation of sites isn't permitted because unit has direct input links",
      "Creation of a link isn't permitted because there exists already a link between these units",
      "Memory allocation failed during critical operation. Have lost some pointers, but consistency of the network is guaranteed",
      "Ftype name isn't definite",
/*10*/"Current Ftype entry isn't defined",
      "Invalid copy mode",
      "Current unit doesn't have sites",
      "Can't update unit because unit is frozen",
      "Redefinition of site name isn't permitted (site name already exists)",
      "Site name isn't defined",
      "Invalid Function: Not a 3D-Kernel",
      "This unit has already a site with this name",
      "Can't delete site table entry because site is in use",
      "Current Ftype site isn't defined",
/*20*/"Ftype symbol isn't defined",
      "Physical I/O error",
      "Creation of output file failed (line length limit exceeded)",
      "The depth of the network doesn't fit to the learning function",
      "No Units defined",
      "Unexpected EOF",
      "Line length exceeded",
      "Incompatible file format",
      "Can't open file",
      "Syntax error at line",
/*30*/"Memory allocation error 1",
      "Topologic type invalid",
      "Symbol pattern invalid (must match [A-Za-z]^[|, ]*)",
      "Current unit doesn't have a site with this name",
      "No hidden units defined",
      "cycle(s):",
      "dead unit(s):",
      "Pattern file contains not the same no. of input units as the network",
      "Pattern file contains not the same no. of output units as the network",
      "No. of input units have changed",
/*40*/"No. of output units have changed",
      "No input units defined",
      "No output units defined",
      "No patterns defined",
      "In-Core patterns incompatible with current network (remove In-Core patterns before loading a new network)",
      "Invalid pattern number",
      "Invalid learning function",
      "Invalid parameters",
      "Invalid update function",
      "Invalid initialisation function",
/*50*/"Derivation function of the activation function doesn't exist",
      "input unit(s) with input connections to other units:",
      "output unit(s) with output connections to other units:",
      "Invalid topological sorting mode",
      "Learning function doesn't support sites"

    };	/*  54 error messages */

  static char	*ext_messages[] = {
      "No Errors",
      "SNNS-Kernel Error: ",
      "SNNS-Kernel Error: Topologic sort has found "
      };



  int  NoOfMessages = (sizeof err_message) / (sizeof err_message[0]);
  static char  mesg[512], aux[512];
  char	 *unit_name;


  if (error_code == 0)	return( ext_messages[0] );  /*	No errors  */

  error_code = abs( error_code );
  if ( error_code >= NoOfMessages )  error_code = 0;  /*  invalid error code  */

  switch (-error_code)
    {
    case KRERR_CYCLES:
      strcpy( mesg, ext_messages[2] );

      if ((unit_name = krui_getUnitName( topo_err_unit )) == NULL)
	sprintf( aux, "%d cycle(s) in the network. (unit #%d produces the first cycle)",
		 topo_NoOfCycles, topo_err_unit );
      else
	sprintf( aux, "%d cycle(s) in the network. (unit #%d (%s) produces the first cycle)",
		 topo_NoOfCycles, topo_err_unit, unit_name );

      strcat( mesg, aux );
      return( mesg );

    case KRERR_DEAD_UNITS:
      strcpy( mesg, ext_messages[2] );

      if ((unit_name = krui_getUnitName( topo_err_unit )) == NULL)
	sprintf( aux, "%d dead unit(s) in the network. (unit #%d is the first dead unit)",
		 topo_NoOfDeadUnits, topo_err_unit );
      else
	sprintf( aux, "%d dead unit(s) in the network. (unit #%d (%s) is the first dead unit)",
		 topo_NoOfDeadUnits, topo_err_unit, unit_name );

      strcat( mesg, aux );
      return( mesg );

    case KRERR_I_UNITS_CONNECT:
      strcpy( mesg, ext_messages[2] );
      strcpy( mesg, "input units with input connections: " );

      if ((unit_name = krui_getUnitName( topo_err_unit )) == NULL)
	sprintf( aux, "unit #%d is one input unit with inputs",
		 topo_err_unit );
      else
	sprintf( aux, "unit #%d (%s) is one input unit with inputs",
		 topo_err_unit, unit_name );

      strcat( mesg, aux );
      return( mesg );

    case KRERR_O_UNITS_CONNECT:
      strcpy( mesg, ext_messages[2] );
      strcpy( mesg, "output units with output connections: " );

      if ((unit_name = krui_getUnitName( topo_err_unit )) == NULL)
	sprintf( aux, "unit #%d is one output unit outputs",
		 topo_err_unit );
      else
	sprintf( aux, "unit #%d (%s) is one output unit with outputs",
		 topo_err_unit, unit_name );

      strcat( mesg, aux );
      return( mesg );


    case KRERR_FILE_OPEN:
      lineno = 0;

    default:
      if (lineno != 0)
	{  /*  file I/O error  */
	strcpy( mesg, ext_messages[1] );
	sprintf( aux, "Loading file failed at line %d : ", lineno );
	strcat( mesg, aux );
	strcat( mesg, err_message[ error_code ] );
	lineno = 0;
	return( mesg );
      }
      else
	{  /*  standard error  */
	strcpy( mesg, ext_messages[1] );
	strcat( mesg, err_message[ error_code ] );
	return( mesg );
      }
    }
}
