/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel transfer functions
NOTES  : 
AUTHOR : Niels Mache
DATE   : 25.06.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <math.h>
#include <string.h>

#include "kr_typ.h"	    /*	Kernel types and constants  */
#include "kr_def.h"	    /*	Default values	*/
#include "func_mac.h"	    /*	Transfer function macros  */


/*#################################################

GROUP: Unit Output Functions

#################################################*/

/*  Linear Output Function
    This function isn't used now, because the identity output function is
    the NULL pointer.
*/
FlintType  OUTP_Identity( activation )
register FlintType  activation;
{
  return( activation );
}


/*  Clipping [0,1] function
*/
FlintType  OUT_Clip_01( activation )
register FlintType  activation;
{
  if (activation < 0.0)  return( (FlintType) 0.0 );
  if (activation > 1.0)  return( (FlintType) 1.0 );
  return( activation );
}


/*  Clipping [-1,1] function
*/
FlintType  OUT_Clip_11( activation )
register FlintType  activation;
{
  if (activation < -1.0)  return( (FlintType) -1.0 );
  if (activation > 1.0)  return( (FlintType) 1.0 );
  return( activation );
}

/*  Threshold 0.5 Output Function
*/
FlintType  OUT_Threshold05( activation )
register FlintType  activation;
{
  if (activation < 0.5)  return( (FlintType) 0.0 );
  return( (FlintType) 1.0 );
}


/*#################################################

GROUP: Unit Activation Functions

#################################################*/


/*  Linear Activation Function
*/
FlintType   ACT_Linear( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  return( sum );
}

/*  Brain-State-in-a-Box Function
*/
FlintType   ACT_BSBFunction( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  return( sum * GET_UNIT_BIAS( unit_ptr ));
}

/*  Minimum Function (Unit's output and weight)
*/
FlintType   ACT_MinOutPlusWeight( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  min1, min2;


  min1 = 0.0;

  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    {
    min1 = GET_OUTPUT + GET_WEIGHT;
    while (GET_NEXT_LINK)
      if ((min2 = GET_OUTPUT + GET_WEIGHT) < min1)
	min1 = min2;
    }
  else
    if (GET_FIRST_SITE( unit_ptr ))
      {
      min1 = GET_SITE_VALUE;
      while (GET_NEXT_SITE)
       if ((min2 = GET_SITE_VALUE) < min1)
	    min1 = min2;
      }

  return( min1 );
}


/*  Hyperbolic Tangent Function
*/
FlintType   ACT_TanHFunction( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  return( tanh(-sum - GET_UNIT_BIAS( unit_ptr )));
}



/*  Sigmoid Function
*/
FlintType   ACT_Logistic( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  return( (FlintType) (1.0 / (1.0 + exp( -sum - GET_UNIT_BIAS( unit_ptr )))) );
}


/*  Perceptron Function
*/
FlintType   ACT_Perceptron( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  if (sum > 0.0)
    return( sum );

  return( (FlintType) 0.0 );
}

/*  Signum Function
*/
FlintType   ACT_Signum( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  if (sum > 0.0)
    return( (FlintType) 1.0 );

  return( (FlintType) -1.0 );
}


/*  Signum0 Function
*/
FlintType   ACT_Signum0( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  if (sum > 0.0)  return( (FlintType) 1.0 );
  if (sum < 0.0)  return( (FlintType) -1.0 );
  return( (FlintType) 0.0 );
}


/*  Step Function
*/
FlintType   ACT_StepFunction( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  if (sum > 0.0)  return( (FlintType) 1.0 );
  return( (FlintType) 0.0 );
}


/*  Bi-Directional Associative Memory
*/
FlintType   ACT_BAMFunction( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum =  0.0;
  if (GET_FIRST_UNIT_LINK( unit_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);
  else
    if (GET_FIRST_SITE( unit_ptr ))
      do
	sum += GET_SITE_VALUE;
      while (GET_NEXT_SITE);

  if (sum > 0.0)  return( (FlintType) 1.0 );
  if (sum < 0.0)  return( (FlintType) -1.0 );
  return( GET_UNIT_ACT( unit_ptr ) );
}


/*  demonstation function: this function act like the Logistic function,
    but the site with the name "Inhibit" will skipped.
*/
FlintType   ACT_LogisticI( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  register FlintType  sum;


  sum = 0.0;
  if (GET_FIRST_SITE( unit_ptr ))
    /*	Do not calculate the 'Inhibit' site */
    do
      if (strcmp( "Inhibit", GET_SITE_NAME ))
        sum += GET_SITE_VALUE;
    while (GET_NEXT_SITE);
  else
    if (GET_FIRST_UNIT_LINK( unit_ptr ))
      do
        sum += GET_WEIGHTED_OUTPUT;
      while (GET_NEXT_LINK);

  return( (FlintType) (1.0 / (1.0 + exp( -sum - GET_UNIT_BIAS( unit_ptr )))) );
}


/*#################################################

GROUP: Derivation Functions of the Activation Functions

#################################################*/

/*  Sigmoid Derivation Function
*/
FlintType   ACT_DERIV_Logistic( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS
  
  return( GET_UNIT_ACT( unit_ptr ) * (1.0 - GET_UNIT_ACT( unit_ptr )) );
}


/*  Identity Derivation Function
*/
FlintType   ACT_DERIV_Identity( unit_ptr )
UNIT_PTR    unit_ptr;
{
  return( (FlintType) 1.0 );
}

/*  Brain-State-in-a-Box Derivation Function
*/
FlintType   ACT_DERIV_BSBFunction( unit_ptr )
UNIT_PTR    unit_ptr;
{
  ACT_FUNC_DEFS

  return( GET_UNIT_BIAS( unit_ptr ));
}


/*  Dummy function for the derivation functions. Returns always the value 1.0.
    This function is used for activation functions that can't have a derivation
    function.

    NOTE: All activation functions have to provide a derivation function.
*/
FlintType   ACT_DERIV_Dummy( unit_ptr )
UNIT_PTR    unit_ptr;
{
  return( (FlintType) 1.0 );
}


/*#################################################

GROUP: Site functions

#################################################*/

/*  Linear Site Function
*/
FlintType  SITE_WeightedSum( site_ptr )
SITE_PTR    site_ptr;
{
  SITE_FUNC_DEFS
  register FlintType  sum;


  sum = 0.0;
  if (GET_FIRST_SITE_LINK( site_ptr ))
    do
      sum += GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);

  return( sum );
}


/*  Product of all predecessor outputs and input link weights
*/
FlintType  SITE_Product( site_ptr )
SITE_PTR    site_ptr;
{
  SITE_FUNC_DEFS
  register FlintType  prod;


  if (GET_FIRST_SITE_LINK( site_ptr ))
    {
    prod = 1.0;
    do
      prod *= GET_WEIGHTED_OUTPUT;
    while (GET_NEXT_LINK);

    return( prod );
    }
  else
    return( (FlintType) 0.0 );
}


/*  Like SITE_Product() but no weighting of the unit's output
*/
FlintType  SITE_ProductA( site_ptr )
SITE_PTR    site_ptr;
{
  SITE_FUNC_DEFS
  register FlintType  prod;


  if (GET_FIRST_SITE_LINK( site_ptr ))
    {
    prod = 1.0;
    do
      prod *= GET_OUTPUT;
    while (GET_NEXT_LINK);

/*  Future Application (in SNNS-Kernel V1.4 the sites don't have weights).
    So the return value is only the product.
*/
    return( GET_SITE_WEIGHT * prod );
    }
  else
    return( (FlintType) 0.0 );
}



/*  Get the highest weighted output
*/
FlintType  SITE_Max( site_ptr )
SITE_PTR    site_ptr;
{
  SITE_FUNC_DEFS
  register FlintType  max, out;


  if (GET_FIRST_SITE_LINK( site_ptr ))
    {
    max = GET_WEIGHTED_OUTPUT;

    while (GET_NEXT_LINK)
      {
      out = GET_WEIGHTED_OUTPUT;
      if (max < out)  max = out;
      }

    return( max );
    }
  else
    return( (FlintType) 0.0 );
}


/*  Get the lowest weighted output
*/
FlintType  SITE_Min( site_ptr )
SITE_PTR    site_ptr;
{
  SITE_FUNC_DEFS
  register FlintType  min, out;


  if (GET_FIRST_SITE_LINK( site_ptr ))
    {
    min = GET_WEIGHTED_OUTPUT;

    while (GET_NEXT_LINK)
      {
      out = GET_WEIGHTED_OUTPUT;
      if (min > out)  min = out;
      }

    return( min );
    }
  else
    return( (FlintType) 0.0 );
}
