/*****************************************************************************
  FILE     : ui_color.c
  SHORTNAME: ui_color.c

  PURPOSE  : handling colors
  FUNCTIONS: 
  NOTES    :

  AUTHOR   : Tilman Sommer, Universitaet Stuttgart
  DATE     : 18.2.1991

             (c) 1991 by SNNS-Group, Institut fuer Informatik (IfI)
             Universitaet Stuttgart
******************************************************************************/

#include <stdio.h>
#include <X11/Xlib.h>

#include "ui.h"
#include "ui_textP.h"


#define UI_MAXCOLORSTEPS 15

/* UI_MAXCOLORSTEPS segments for positive values,
   UI_MAXCOLORSTEPS segments for negative values,
               1 segment  for zero,
	       1 color    for background color 

   total of 2 * UI_MAXCOLORSTEPS + 2 colors to allocate
   
*/

extern Display       *ui_display;
extern int            ui_screen;

extern Widget         ui_toplevel;

unsigned long ui_back_col_blue = 65535;
unsigned long ui_back_col_green = 65535;
unsigned long ui_back_col_red = 0; 

unsigned long   ui_col_rangePixels[2 * UI_MAXCOLORSTEPS + 1];

/* UI_MAXCOLORSTEPS segments for positive values,
   UI_MAXCOLORSTEPS segments for negative values,
               1 segment  for zero
*/


Bool    ui_col_monochromeMode = TRUE;
Bool    ui_col_colorDisplay;

XColor  ui_col_max,ui_col_min,ui_col_mid,ui_col_back,ui_col_text,ui_col_select;

Visual  *ui_col_visual;

int     ui_col_steps = UI_MAXCOLORSTEPS;

unsigned long ui_cells[64];
unsigned long ui_planes;

/* all palettes (10): text, back, selection 
   with each display only the palette number is stored! The pixels
   values for the associated colors are held in the array: */ 

unsigned long ui_col_mainCells[10][3]; 


/*
 * Visual structure; contains information about colormapping possible.


	XExtData *ext_data;	 hook for extension to hang data 
	VisualID visualid;	 visual id of this visual 
if defined(__cplusplus) || defined(c_plusplus)
	int c_class;		 C++ class of screen (monochrome, etc.)
else
	int class;		 class of screen (monochrome, etc.)
endif
	unsigned long red_mask, green_mask, blue_mask;	mask values
	int bits_per_rgb;	 log base 2 of distinct color values
	int map_entries;	 color map entries 

*/



/*****************************************************************************
  FUNCTION : ui_col_setColor

  PURPOSE  :
  NOTES    :

  UPDATE   : 18.2.1991
******************************************************************************/

void
ui_col_setColor(color, value, intensity, print)
     XColor *color;
     int    value;
     int    intensity;
     Bool   print;
{
    switch (value) {
      case UI_COL_SEA: 
	color->blue  = ui_back_col_blue;
	color->red   = ui_back_col_red;
	color->green = ui_back_col_green;
	if (print) ui_tw_printMessage("seablue\n");
	break;
      case UI_COL_BLUE: 
	color->blue  = 65535;
	color->red   = 0;
	color->green = 0;
	if (print) ui_tw_printMessage("blue\n");
	break;
      case UI_COL_YELLOW: 
	color->blue  = 0;
	color->red   = 65535;
	color->green = 65535;
	if (print) ui_tw_printMessage("yellow\n");
	break;
      case UI_COL_BLACK: 
	color->blue  = 0;
	color->red   = 0;
	color->green = 0;
	if (print) ui_tw_printMessage("black\n");
	break;
      case UI_COL_WHITE: 
	color->blue  = 65535;
	color->red   = 65535;
	color->green = 65535;
	if (print) ui_tw_printMessage("white\n");
	break;
      case UI_COL_RED: 
	color->blue  = 0;
	color->red   = 65535;
	color->green = 0;
	if (print) ui_tw_printMessage("red\n");
	break;
      case UI_COL_GREEN: 
	color->blue  = 0;
	color->red   = 0;
	color->green = 65535;
	if (print) ui_tw_printMessage("green\n");
	break;
    }

    switch (intensity) {
      case UI_COL_DARK: 
	color->blue  = color->blue DIV 2;
	color->red   = color->red DIV 2;
	color->green = color->green DIV 2;
	break;
      case UI_COL_BRIGHT: 
	  if (color->blue < 32768)
	      color->blue  += (65535 - color->blue ) DIV 3;
	  if (color->red  < 32768)
	      color->red   += (65535 - color->red  ) DIV 3;
	  if (color->green < 32768)
	      color->green += (65535 - color->green) DIV 3;
	break;
    }
}



/*****************************************************************************
  FUNCTION : ui_col_setMainColors

  PURPOSE  :
  NOTES    :

  UPDATE   : 18.2.1991
******************************************************************************/

void
ui_col_setMainColors(text,back,sel)
     int text, back,sel;
{
    ui_col_setColor(&ui_col_back, back, UI_COL_NORM, FALSE);
    ui_col_setColor(&ui_col_text, text, UI_COL_NORM, FALSE);
    ui_col_setColor(&ui_col_select, sel, UI_COL_BRIGHT, FALSE);
}



/*****************************************************************************
  FUNCTION : ui_col_setRangeColors

  PURPOSE  :
  NOTES    :

  UPDATE   : 18.2.1991
******************************************************************************/

void
ui_col_setRangeColors(min,mid,max)
     int min, mid, max;
{
    ui_tw_printMessage("positive values: ");
    ui_col_setColor(&ui_col_max , max , UI_COL_NORM, TRUE);
    ui_tw_printMessage("negative values: ");
    ui_col_setColor(&ui_col_min , min , UI_COL_NORM, TRUE);
    ui_tw_printMessage("zero     value : ");
    switch (mid) {
      case UI_COL_BLACK:
	ui_col_setColor(&ui_col_mid , mid , UI_COL_BRIGHT, TRUE);
	break;
      case UI_COL_BLUE:
	ui_col_setColor(&ui_col_mid , mid , UI_COL_DARK, TRUE);
	break;
      default:
	ui_col_setColor(&ui_col_mid , mid , UI_COL_NORM, TRUE);
    }
}




/*****************************************************************************
  FUNCTION : ui_col_colorSegment

  PURPOSE  : the 'distance' beetwen the colors mid and end is divided into
             ui_col_steps segments. This functions calculates the color of
	     a given segment beetwin start and end
  NOTES    :

  UPDATE   : 18.2.1991
******************************************************************************/

void
ui_col_colorSegment(start, end, segment, color)
     XColor start, end;
     int    segment;
     XColor *color; /* result */
{
    /* use a little bit linear algebra */

    color->red = 
	(segment ) * (end.red   - start.red  ) DIV ui_col_steps 
	    + start.red;
    color->green = 
	(segment  ) * (end.green - start.green) DIV ui_col_steps 
	    + start.green;
    color->blue = 
	(segment  ) * (end.blue  - start.blue ) DIV ui_col_steps 
	    + start.blue;
}






/*****************************************************************************
  FUNCTION : ui_col_setPalette

  PURPOSE  :
  NOTES    :

  UPDATE   : 18.2.1991
******************************************************************************/

void
ui_col_setPalette(no)
     int no;
{
    if (ui_col_colorDisplay AND 
	((no == UI_PALETTE_MONO) OR (no == UI_PALETTE_MONO_INVERSE))) {
	ui_col_setColor(&ui_col_select, UI_COL_YELLOW, UI_COL_BRIGHT, FALSE);
    }
    
    if (ui_col_colorDisplay)
	if (NOT XAllocColor(ui_display, 
			    DefaultColormap(ui_display, ui_screen),
			    &ui_col_select))
	    ui_tw_printMessage("Can't allocate selection color!\n");
    

    switch (no) {
      case UI_PALETTE_MONO: /* monochrome */
	ui_col_text.pixel     = BlackPixel(ui_display, ui_screen);
	ui_col_back.pixel     = WhitePixel(ui_display, ui_screen);
	break;
      case UI_PALETTE_MONO_INVERSE: /* monochrome */
	ui_col_text.pixel     = WhitePixel(ui_display, ui_screen);
	ui_col_back.pixel     = BlackPixel(ui_display, ui_screen);
	break;
      case UI_PALETTE_RED_BLUE_GREEN_WHITE_WHITE_YELLOW:
	ui_col_setMainColors(UI_COL_WHITE, UI_COL_SEA, UI_COL_YELLOW);
	break;
      case UI_PALETTE_RED_BLUE_GREEN_WHITE_BLUE_YELLOW:
	ui_col_setMainColors(UI_COL_WHITE, UI_COL_BLUE, UI_COL_YELLOW);
	break;
      case UI_PALETTE_RED_BLUE_GREEN_WHITE_BLACK_YELLOW:
	ui_col_setMainColors(UI_COL_WHITE, UI_COL_BLACK, UI_COL_YELLOW);
	break;
      case UI_PALETTE_RED_BLUE_GREEN_BLACK_WHITE_YELLOW:
	ui_col_setMainColors(UI_COL_BLACK, UI_COL_WHITE, UI_COL_YELLOW);
	break;
      case UI_PALETTE_RED_BLUE_GREEN_WHITE_YELLOW_WHITE:
	ui_col_setMainColors(UI_COL_WHITE, UI_COL_YELLOW, UI_COL_WHITE);
	break;
    }


    if ((no != UI_PALETTE_MONO) AND (no != UI_PALETTE_MONO_INVERSE)) {
	int i;

	if (NOT XAllocColor(ui_display, 
			    DefaultColormap(ui_display, ui_screen),
			    &ui_col_text))
	    ui_tw_printMessage("Can't allocate text color!\n");
	
	if (NOT XAllocColor(ui_display, 
			    DefaultColormap(ui_display, ui_screen),
			    &ui_col_back))
	    ui_tw_printMessage("Can't allocate background color!\n");
	
    }	    

    /* store in table */

    ui_col_mainCells[no][0] = ui_col_text.pixel;
    ui_col_mainCells[no][1] = ui_col_back.pixel;
    ui_col_mainCells[no][2] = ui_col_select.pixel;
    
}




/*****************************************************************************
  FUNCTION : ui_col_init

  PURPOSE  : init color module
  NOTES    : called from ui_init()

  UPDATE   : 18.2.1991
******************************************************************************/


void
ui_col_init()
{

    char buf[80], vc[30];
    Display *dpy  = XtDisplay(ui_toplevel);
    Colormap cmap = DefaultColormap(dpy, DefaultScreen(dpy));
    int  i;


    ui_col_visual = DefaultVisual(dpy, DefaultScreen(dpy));

    ui_col_colorDisplay = (ui_col_visual->class != StaticGray);

#ifdef DEBUG
    switch (ui_col_visual->class) {
      case StaticGray: 
	sprintf(vc,"StaticGray");
	break;
      case GrayScale: 
	sprintf(vc,"GrayScale");
	break;
      case StaticColor: 
	sprintf(vc,"StaticColor");
	break;
      case PseudoColor: 
	sprintf(vc,"PseudoColor");
	break;
      case TrueColor: 
	sprintf(vc,"TrueColor");
	break;
      case DirectColor: 
	sprintf(vc,"DirectColor");
	break;
    }
    sprintf(buf,"\nVisual Class: %s    bits per rgb: %d    map entries: %d\n",
	    vc, ui_col_visual->bits_per_rgb,
	    ui_col_visual->map_entries); 
    ui_tw_printMessage(buf);
#endif

    /* initialize all palettes */

    for (i=0; i<=3; i++)
	ui_col_setPalette(i); 

    if (! ui_col_colorDisplay) {
	ui_tw_printMessage("XGUI running in monochrome mode!\n");
	ui_col_monochromeMode = TRUE;
    } else {
	ui_tw_printMessage("Color/Grayscale screen detected!\n");
	ui_tw_printMessage("XGUI running in color mode!\n");
	ui_col_monochromeMode = FALSE;

	/* initialize the array of all colors used to draw units/links */

	ui_col_setRangeColors(UI_COL_RED, UI_COL_BLUE, UI_COL_GREEN);

	for (i=0; i <= 2 * ui_col_steps; i++) {
	    XColor color;
	    
	    if (i<ui_col_steps) /* range: [min,0) */
		ui_col_colorSegment(ui_col_mid, ui_col_min, 
				    ui_col_steps - i, &color);
	    else /* range: [0,max] */ {
		if (i == ui_col_steps)
		    color = ui_col_mid;
		else
		    ui_col_colorSegment(ui_col_mid, ui_col_max,
					i - ui_col_steps, &color);
	    }
	    if (NOT XAllocColor(ui_display, 
				DefaultColormap(ui_display, ui_screen),
				&color))
		ui_tw_printMessage("Can't allocate color!\n");
	    ui_col_rangePixels[i] = color.pixel;
#ifdef DEBUG
	    printf("%2d  %5d  %5d  %5d\n",
		   i, color.red, color.green, color.blue);
#endif
	}
	
    } 
}

/* end of file */
