/*****************************************************************************
  FILE     : ui_file.c
  SHORTNAME: file.c

  PURPOSE  : popups a window with all filenames. The user may alter the names
             and may request a save or load operation with the correspondent
	     kind of file (NET, GUI, PAT, CFG, TXT).
  FUNCTIONS: 
  NOTES    : is called only during initialisation

  AUTHOR   : Tilman Sommer
  DATE     : 22.5.1990

  COPYRIGHT: (c) 1990 by Tilman Sommer and the SNNS-Group
             
******************************************************************************/


#include "ui.h"


#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Cardinals.h>

#include "ui_file.h"
#include "ui_fileP.h"
#include "ui_xWidgets.h"


Widget    ui_filePanel; 

/* widgets of all input fields */

Widget    ui_fileNET, ui_filePAT, ui_fileCFG, ui_fileTXT, ui_path;


/* strings storing the data of all input fields */

extern char   ui_filenameNET[];
extern char   ui_filenamePAT[];
extern char   ui_filenameCFG[];
extern char   ui_filenameTXT[];
extern char   ui_pathname[];

Bool      ui_fileIsCreated = FALSE; /* file panel created yes or no */



/*****************************************************************************
  FUNCTION :                 UI_XCREATE FILE PANEL

  PURPOSE  : create the file panel
  RETURNS  : void
  NOTES    : the wigdet will be created in relative position to eachother
*****************************************************************************/


void
ui_xCreateFilePanel(parent)
     Widget parent;      /* the parent widget of the new form widget */
{

    /* tabs for data fields and small buttons */


    Widget      netLabel, patLabel, cfgLabel, txtLabel, pathLabel;
    Widget      button, sbutton;
    
    int  fontWidth = 8;

    int  titelWidth  = 10 * fontWidth;
    int  nameWidth   = 30 * fontWidth;  



/***************************************************************************/

    ui_filePanel = 
	XtCreateManagedWidget("fPanel", formWidgetClass, parent, NULL, ZERO);


/*****************************  HEAD LINE  *********************************/

    pathLabel =
	ui_xCreateLabelItem("directory",ui_filePanel, titelWidth,  NULL, NULL);
    ui_xSetResize(pathLabel, TRUE);

    netLabel =
	ui_xCreateLabelItem("net",      ui_filePanel, titelWidth,  NULL, pathLabel);
    ui_xSetResize(netLabel, TRUE);

    patLabel =
	ui_xCreateLabelItem("pattern",  ui_filePanel, titelWidth,  NULL, netLabel);
    ui_xSetResize(patLabel, TRUE);

    cfgLabel =
	ui_xCreateLabelItem("config",   ui_filePanel, titelWidth,  NULL, patLabel);
    ui_xSetResize(cfgLabel, TRUE);
 
    txtLabel =
	ui_xCreateLabelItem("logFile",  ui_filePanel, titelWidth,  NULL, cfgLabel);
    ui_xSetResize(txtLabel, TRUE);

/*****************************  NAMES  *************************************/

    ui_path =
	ui_xCreateDialogItem("path", ui_filePanel, ui_pathname, 0 /* free */, 
			     pathLabel, NULL);

    ui_fileNET =
	ui_xCreateDialogItem("net", ui_filePanel, "", nameWidth, 
			     netLabel, pathLabel);
    ui_xSetResize(ui_fileNET, TRUE);
    ui_xSetString(ui_fileNET, ui_filenameNET);

    ui_filePAT =
	ui_xCreateDialogItem("pat", ui_filePanel, "", nameWidth,  
			     patLabel, netLabel);
    ui_xSetResize(ui_filePAT, TRUE);
    ui_xSetString(ui_filePAT, ui_filenamePAT);

    ui_fileCFG = 
	ui_xCreateDialogItem("cfg", ui_filePanel, "", nameWidth,  
			     cfgLabel, patLabel);
    ui_xSetResize(ui_fileCFG, TRUE);
    ui_xSetString(ui_fileCFG, ui_filenameCFG);
	      
    ui_fileTXT = 
	ui_xCreateDialogItem("txt", ui_filePanel, "", nameWidth,  
			     txtLabel, cfgLabel);
    ui_xSetResize(ui_fileTXT, TRUE);
    ui_xSetString(ui_fileTXT, ui_filenameTXT);

/*****************************  BUTTONS  ***********************************/

    button =
	ui_xCreateButtonItem("save", ui_filePanel, ui_fileNET , pathLabel);

    XtAddCallback(button, XtNcallback, ui_file_saveNet, NULL); 
    button =
	ui_xCreateButtonItem("load", ui_filePanel, button    , pathLabel);
    XtAddCallback(button, XtNcallback, ui_file_loadNet, NULL); 

    button =
	ui_xCreateButtonItem("save", ui_filePanel, ui_filePAT , netLabel);
    XtAddCallback(button, XtNcallback, ui_file_savePatterns, NULL); 

    button =
	ui_xCreateButtonItem("load", ui_filePanel, button    , netLabel);
    XtAddCallback(button, XtNcallback, ui_file_loadPatterns, NULL); 

    sbutton =
	ui_xCreateButtonItem("save", ui_filePanel, ui_fileCFG, patLabel);
    XtAddCallback(sbutton, XtNcallback, ui_file_saveConfiguration, NULL); 

    button =
	ui_xCreateButtonItem("load", ui_filePanel, sbutton    , patLabel);
    XtAddCallback(button, XtNcallback, ui_file_loadConfiguration, NULL); 

    button =
	ui_xCreateButtonItem("load", ui_filePanel, sbutton    , cfgLabel);
    XtAddCallback(button, XtNcallback, ui_file_loadText, NULL); 


/***************************************************************************/

    XawFormDoLayout(ui_filePanel, True);
}








/* end of file */
/* lines: 176*/
